#ifndef _ALLOCATING_BIPOINTER_H
#define _ALLOCATING_BIPOINTER_H

#include <stdio.h>

#include <stdlib.h>
#include <time.h>

#include <string>
#include <iostream>
#include <sstream>
#include <fstream>
#include <cmath>
#include <cassert>

#include <stdexcept>

#include <vector>
#include <list>
#include <map>

using namespace std;

#include "exception.h"
#include "tester.h"
#include "bipointer.h"

using namespace VDrift;

template<typename T>
class ALLOCATING_BIPOINTER
{
private:
	BIPOINTER <T> biptr;
	T * local_storage;
	void ClearLocalStorage() {if (local_storage != NULL) delete local_storage;local_storage = NULL;}
	void ClearBipointer() {biptr.Invalidate();}

public:
	ALLOCATING_BIPOINTER() : local_storage(NULL) {}
	~ALLOCATING_BIPOINTER() {Invalidate();}
	
	ALLOCATING_BIPOINTER(const ALLOCATING_BIPOINTER <T> & other) : local_storage(NULL) {Set(other);}
	ALLOCATING_BIPOINTER & operator=(const ALLOCATING_BIPOINTER <T> & other) {Set(other);return *this;}

	const TESTER Test()
	{
		TESTER mytest;
		
		int points;
		
		points = 0;
		if (!PtrValid()) points++;
		AllocateLocalStorage();
		if (local_storage != NULL) points++;
		if (!biptr.PtrValid()) points++;
		if (PtrValid()) points++;
		T testptr;
		SetPtr(testptr);
		if (local_storage == NULL) points++;
		if (biptr.PtrValid()) points++;
		if (PtrValid()) points++;
		Invalidate();
		if (local_storage == NULL) points++;
		if (!biptr.PtrValid()) points++;
		if (!PtrValid()) points++;
		SetPtr(testptr);
		if (local_storage == NULL) points++;
		if (biptr.PtrValid()) points++;
		if (PtrValid()) points++;
		
		mytest.SubTestComplete("Pointer/bipointer allocation/deallocation", points, 13);
		
		return mytest;
	}
	
	void Set(const ALLOCATING_BIPOINTER <T> & other)
	{
		//cout << "ALLOCATING_BIPOINTER::Set()" << endl;
		Invalidate();
		if (other.biptr.PtrValid())
			biptr = other.biptr;
		else if (other.local_storage != NULL)
		{
			AllocateLocalStorage();
			*local_storage = *(other.GetPtr());
		}
	}
	
	const bool PtrValid() const {return (local_storage != NULL || biptr.PtrValid());}
	
	void Invalidate() {ClearLocalStorage();ClearBipointer();}
	
	void SetPtr(T & objref)
	{
		ClearLocalStorage();
		biptr.SetPtr(objref);
	}
	
	T * const GetPtr() const
	{
		if (biptr.PtrValid() && local_storage != NULL)
		{
			stringstream err;
			err << __FILE__ << ", line " << __LINE__ << ":  ALLOCATING_BIPOINTER has two valid pointers" << endl;
			throw EXCEPTION(err.str());
			//throw EXCEPTION(__FILE__, __LINE__, "ALLOCATING_BIPOINTER has two valid pointers");
		}
		else if (!biptr.PtrValid() && local_storage == NULL)
		{
			stringstream err;
			err << __FILE__ << ", line " << __LINE__ << ":  ALLOCATING_BIPOINTER has no valid pointers" << endl;
			throw EXCEPTION(err.str());
			//throw EXCEPTION(__FILE__, __LINE__, "ALLOCATING_BIPOINTER has no valid pointers");
		}
		else if (biptr.PtrValid())
			return biptr.GetPtr();
		else if (local_storage != NULL)
			return local_storage;
		else
		{
			throw EXCEPTION(__FILE__, __LINE__, "ALLOCATING_BIPOINTER has odd specification");
		}
	}
	
	void AllocateLocalStorage()
	{
		ClearBipointer();
		ClearLocalStorage();
		local_storage = new T;
	}
};

#endif
