#ifndef _BIPOINTER_H
#define _BIPOINTER_H

#include <stdio.h>

#include <stdlib.h>
#include <time.h>

#include <string>
#include <iostream>
#include <sstream>
#include <fstream>
#include <cmath>
#include <cassert>

#include <stdexcept>

#include <vector>
#include <list>
#include <map>

using namespace std;

#include "exception.h"

using namespace VDrift;

//#define OPTIMIZE

class BIPOINTER_TARGET;

class BIPOINTER_COMMON
{
public:
	virtual void InvalidatePointerTo(BIPOINTER_TARGET * targ) {}
	virtual ~BIPOINTER_COMMON() {}
};

class BIPOINTER_TARGET
{
private:
	void BIPOINTER_notify(); //notifies registrants that we have been destroyed
	list <BIPOINTER_COMMON *> registrants;
public:
	virtual ~BIPOINTER_TARGET() {BIPOINTER_notify();}
	void BIPOINTER_register(BIPOINTER_COMMON * newptr);
	void BIPOINTER_deregister(BIPOINTER_COMMON * oldptr);
};

template<typename T>
class BIPOINTER : public BIPOINTER_COMMON
{
private:
	BIPOINTER_TARGET * ptr;

public:
	BIPOINTER() : ptr(NULL) {}
	virtual ~BIPOINTER() {if (ptr != NULL) ptr->BIPOINTER_deregister(this);}
	
	BIPOINTER(const BIPOINTER <T> & other) {Set(other);}
	BIPOINTER <T> & operator=(const BIPOINTER <T> & other) {Set(other);return *this;}
	
	void Set(const BIPOINTER <T> & other)
	{
		//cout << "Setting this " << this << " ptr to " << other.GetPtr() << endl;
		
		if (other.PtrValid())
		{
			SetPtr(*other.GetPtr());
		}
		else
			ptr = NULL;
	}
	
	const bool PtrValid() const {return ptr != NULL;}
	
	void SetPtr(T & objref)
	{
		//cout << "Setptr this " << this << " ptr " << ptr << " to " << &objref << endl;
		
		//if (ptr == (BIPOINTER_TARGET *)(&objref))
		if (ptr == &objref)
			return;
		
		if (ptr != NULL)
		{
			ptr->BIPOINTER_deregister(this);
		}
		
		//ptr = (BIPOINTER_TARGET *)(&objref);
		ptr = &objref;
		objref.BIPOINTER_register(this);
	}
	
	T * const GetPtr() const
	{
		#ifndef OPTIMIZE
		if (!PtrValid())
		{
			stringstream err;
			err << __FILE__ << ", line " << __LINE__ << ":  Asked for NULL pointer: " << this << endl;
			throw EXCEPTION(err.str());
		}
		#endif
		return (T*) ptr;
	}
	
	T & GetRef() const
	{
		#ifndef OPTIMIZE
		if (!PtrValid())
		{
			stringstream err;
			err << __FILE__ << ", line " << __LINE__ << ":  Asked for NULL reference" << endl;
			throw EXCEPTION(err.str());
			//throw EXCEPTION(__FILE__, __LINE__, "Asked for NULL reference");
		}
		#endif
		return *((T*) ptr);
	}
	
	virtual void InvalidatePointerTo(BIPOINTER_TARGET * targ)
	{
		if (ptr == targ)
			ptr = NULL;
		else
		{
			stringstream err;
			err << __FILE__ << ", line " << __LINE__ << ":  Detected pointer reference corruption" << endl;
			throw EXCEPTION(err.str());
		}
	}
	
	void Invalidate()
	{
		if (ptr != NULL)
		{
			ptr->BIPOINTER_deregister(this);
		}
		
		ptr = NULL;
	}
};

#endif
