/***************************************************************************
 *            font.h
 *
 *  Sun Sep 26 13:33:51 2004
 *  Copyright  2004  Joe Venzon
 *  joe@venzon.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifndef _FONT_H
#define _FONT_H

#include "exception.h"
#include "settings.h"
#include "textures.h"
#include "bipointer.h"
#include "scenegraph.h"

class FONTCOLOR
{
public:
	float r,g,b,a;
	const bool operator<(const FONTCOLOR& other) const;
	const bool operator==(const FONTCOLOR& other) const {return (memcmp(this,&other,sizeof(other))==0);}
};

class CHARACTERINFO
{
public:
	char character;
	int minx;
	int maxx;
	int miny;
	int maxy;
};

class FONT
{
private:
	bool loaded;
	//map < FONTCOLOR, BIPOINTER <DRAWABLE> > drawable_map;
	list < BIPOINTER<DRAWABLE> > drawable_list;
	//BIPOINTER <VERTEXARRAY> varray;
	string fontname;
	CHARACTERINFO charinfo[96];
	float averagecenter;
	int fontsize, totalx, totaly, tw, th;
	TEXTUREINFO tinfo;

	DRAWABLE & GetDrawable(const FONTCOLOR & color);

	SCENENODE * sceneparentnode;
	
public:
	FONT() : loaded(false),sceneparentnode(NULL) {}
	
	const bool Load(const string & newfontname, SCENENODE & parentnode);
	const string & GetFontName() const {return fontname;}
	const float GetAverageCenter() const {return averagecenter;}
	const int GetFontSize() const {return fontsize;}
	const float Print(const float px, const float py, const string & str, const float scale, const float r, const float g, const float b, const float a, const bool draw);
	const float DrawChar(const char curchar, const float scale, const bool draw, const float px, const float py, VERTEXARRAY & v);
	void ClearAllText();
};

class FONTMANAGER
{
private:
	list <FONT> fontlist;
	BIPOINTER <SCENENODE> fontrootnode;

	const string GetBackwardCompatibilityFontName(const int set) const;
	const float GetBackwardCompatibilityScale(const int size) const;
	
public:
	FONTMANAGER() {}
	
	FONT * const GetFont(const string & name);
		
	void ClearFonts() {fontlist.clear();}
	
	const bool LoadFont(const string & newfontname);
	const float Print(const string & name, const float px, const float py, const string & str, const float scale, const float r, const float g, const float b, const float a);
	//float Print(string & name, float px, float py, string & str, float scale, float r, float g, float b, float a, RENDERRECT & dest);
	const float PrintWidth(const string & name, const string & str, const float scale);  //returns a width in percent of screen
	const float GetAverageVerticalPosition(const string & name);
	const int GetFontSize(const string & name);
	//const float GetNormalizedFontHeight(const string & name, const float scale, RENDERRECT & dest);
	void ClearAllText();
	
	//old interface, provided for backwards compatibility
	const float GetFontHeight(const string & name, const float scale);
	void Print( float px, float py, const char *string, int set, int size, float opacity ) {Print( px, py, string, set, size, 1,1,1, opacity);}
	void Print( float px, float py, const char *string, int set, int size, float r, float g, float b ) {Print(px, py, string, set, size, r, g, b, 1);}
	void Print( float px, float py, const char *string, int set, int size, float r, float g, float b, float trans );
	const float Width(const char * str, int set, int size);
	float Height(const char * str, int set, int size);
	void Load();
	
	SCENENODE * GetFontSceneNode() {if (fontrootnode.PtrValid()) return fontrootnode.GetPtr(); else return NULL;}
};


#endif
