#ifndef _GAME_H
#define _GAME_H

#include <string>
#include <list>

using namespace std;

#define AVERAGEFRAMES 3

#include "font.h"
#include "utility.h"
#include "textures.h"
#include "camera.h"
#include "keyman.h"
#include "mouse.h"
#include "messageq.h"
#include "particles.h"
#include "settings.h"
#include "sound.h"
#include "timer.h"
#include "controls.h"
#include "vamosworld.h"
#include "replay.h"
#include "logging.h"
#include "logo.h"
#include "gamestate.h"
#include "objects.h"
#include "multiplay.h"
#include "net.h"
#include "configfile.h"
#include "bezier.h"
#include "track.h"
#include "textures.h"
#include "exception.h"
#include "cardinfo.h"
#include "singleton.h"
#include "graphics.h"
#include "autox_track.h"
#include "physics.h"

using namespace VDrift;

#include <vamos/body/Gl_Car.h>
#include <vamos/geometry/Texture_Image.h>
#include <vamos/geometry/Three_Vector.h>
#include <vamos/body/Car.h>
#include <vamos/body/Fuel_Tank.h>
#include <vamos/body/Wheel.h>
#include <vamos/geometry/Conversions.h>
#include <vamos/world/World.h>

#include "gui/gui.h"

class GAME
{
private:
	SDL_Surface *surface;

	GLint T0;
	GLint Frames;
	GLfloat fps;
	
	bool showfps;
	int frameno;
	float timefactor;
	bool benchmode; // should we run a benchmark test and then quit?
	float pfps;
	float lfps[AVERAGEFRAMES];
	int lfpspos;
	bool lfpsfull;
	double execution_time;
	bool pauserequest;
	bool unpauserequest;
	bool esc_pressed;
	float view_dist;
	float field_of_view;
	string cur_tex_size;
	int cur_screen_w;
	int cur_screen_h;
	int cur_screen_bpp;
	bool cur_fullscreen;
	string data_dir;
	bool verbose_output;
	bool done;
	LOGGER& log;
	
	void handleKeyPress( SDL_keysym *keysym );
	void handleKeyRelease( SDL_keysym *keysym );
	void glSetup();
	
	TEXTURE_HANDLE loadscreen[2];
	int loadanim;
	const int loadanimcount;
	
	float frame_time;
	int collision_frameskip;

public:
	GAME() : timefactor(1.0),benchmode(false),lfpspos(0),lfpsfull(false),
		execution_time(0),pauserequest(false),unpauserequest(false),
		esc_pressed(false),view_dist(10000.0),field_of_view(45.0),
		cur_screen_w(0),cur_screen_h(0),cur_screen_bpp(0),
		cur_fullscreen(false),verbose_output(true),done(false),log(VDrift::LOGGING::GetLogger("VDrift")),
		loadanim(0),loadanimcount(2),
		frame_time(0.004), collision_frameskip(5) {}
			
	SETTINGS settings;
	KEYMAN keyman;
	GAMECONTROLS gamecontrols;
	GRAPHICS graphics;
	TEXTURES textures;
	CAMERA cam;
	FONTMANAGER font;
	UTILITY utility;
	MOUSE mouse;
	MESSAGEQ mq1;
	PARTICLE particle;
	SOUND sound;
	TIMER timer;
	REPLAY replay;
	LOGO logo;
	GAMESTATE state;
	OBJECTS objects;
	NET net;
	MULTIPLAY multiplay;
	VGUI::Gui gui;
	TRACK track;
	CARDINFO gfxcard;
	VAMOSWORLD world;
	AUTOX_TRACK autox_track;
	PHYSICS physics;
	
	bool Start(list <string> & args);
	void End(bool success);
	void WriteDebuggingData(string data) {cerr << data << endl;}
	float FrameTime() {return frame_time;}
	int GetCollisionFrameskip() {return collision_frameskip;}
	void Test();
	
	int Screenshot(char *filename);
	void ResetWorld(bool fullreset);
	int resizeWindow(int width, int height);
	void LoadingScreen(string loadtext);
	void ChangeDisplay(int w, int h, int bpp, bool fullscreen, bool reloadtextures);
	bool ReloadDisplay(bool test);
	bool UnloadWorld();
	void Update();
	void snap_screenshot();
	void MainPause();
	void MainUnpause();
	void SelectCar(string cfile, bool trim);
	bool LoadWorld();
	void InitGameData();
	void InitGL();
	//void ApplyScaleBias(float scale, float bias);
	int drawGLScene();
	int drawGLScene_reflected( QUATERNION & direction, GLenum cube_target, VERTEX position, int sizex, int sizey, GLuint texture_id, bool reflection_effects );
	//void RenderQuad();
	int drawGLScene_reflected( QUATERNION & direction, GLenum cube_target, VERTEX position, int sizex, int sizey, int faceid );
	int ScreenW() const {return cur_screen_w;}
	int ScreenH() const {return cur_screen_h;}
	void Quit();
};

#endif
