/***************************************************************************
 *            graphics.h
 *
 *  Wed May 09 18:48:05 2007
 *  Copyright  2007  Joe Venzon
 *  joe@venzon.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
  
#ifndef _GRAPHICS_H

#include "pch.h"
#include "scenegraph.h"
#include "shader.h"

#define NUM_CASCADED_SHADOW_MAP 2

using namespace std;

class FBTEXTURE_GL
{
private:
	GLuint fbtexture;
	GLuint renderbuffer_depth;
	GLuint framebuffer_object;
	bool loaded;
	bool inited;
	int sizew, sizeh;
	bool rectangle;
	bool depth;

public:
	FBTEXTURE_GL() : loaded(false),inited(false),sizew(0),sizeh(0),rectangle(false),depth(false) {}
	~FBTEXTURE_GL() {DeInit();}
	//void Init() {Init(false);}
	void Init(int sizex, int sizey, bool rect, bool newdepth);
	void DeInit();
	void Begin() {Begin(1.0);}
	void Begin(float viewscale);
	void End();
	void Activate();
};

class GRAPHICS_GLSTATEMANAGER
{
private:
	map <int,bool> state;

public:
	void Set(int stateid, bool newval);
};

class GRAPHICS
{
private:
	SCENEGRAPH scene;
	GRAPHICS_GLSTATEMANAGER glstate;
	FBTEXTURE_GL fbtexture;
	//FBTEXTURE_GL fbshadowtexture;
	FBTEXTURE_GL fbblurtexture;
	FBTEXTURE_GL fbdepthtexture[NUM_CASCADED_SHADOW_MAP];
	SHADER_MANAGER shaders;
	//TEXTURE_HANDLE noise;
	TEXTURE_HANDLE static_reflection;
	//TEXTURE_HANDLE static_lightmap;

	int width, height;
	float fov;
	bool shaders_enabled;
	bool blur_enabled;
	//enum {LIGHTMAP_DISABLED, LIGHTMAP_STATIC, LIGHTMAP_DYNAMIC} lightmap_status;
	enum {REFLECTION_DISABLED, REFLECTION_STATIC, REFLECTION_DYNAMIC} reflection_status;

	bool bloom_enabled;
	bool shadow_enabled;
	int shadow_resolution;
	float fps;
	float truefps;
	float prev_modelview[16];
	float prev_modelviewproj[16];

	enum {SHADER_SIMPLE, SHADER_BLURPASS, SHADER_FULL, SHADER_FULL_NOSPECULAR, SHADER_FULL_NOMETALLIC, UNKNOWN, SHADER_DEPTH_GEN, SHADER_SHADOW_PASS} currentshader;
	const TEXTURE_SDLGL * current_diffusemap;
	MATRIX4 current_matrix;
	MATRIX4 light_matrix;
	MATRIX4 last_transform;
	bool last_transform_valid;
	
	QUATERNION light_position;

	void RenderQuad(float scale, float texw, float texh);
	
	//below are used by the DrawList function
	bool FrustumCull(SCENEDRAW & tocull, float lod_far) const;
	void SelectAppropriateShader(SCENEDRAW & forme, int passno);
	void SelectFlags(SCENEDRAW & forme, bool blurring, int passno);
	void SelectTexturing(SCENEDRAW & forme, bool blurring, int passno);
	bool SelectTransformStart(SCENEDRAW & forme, bool blurring);
	void SelectTransformEnd(SCENEDRAW & forme, bool blurring, bool need_pop);

public:
	GRAPHICS() : width(800),height(600),fov(45.0),shaders_enabled(false),blur_enabled(false),reflection_status(REFLECTION_DISABLED),bloom_enabled(false),shadow_enabled(false),shadow_resolution(1024),fps(0),truefps(0),currentshader(UNKNOWN),current_diffusemap(NULL) {}
	
	void RenderQuad(float scale) {RenderQuad(scale,1.0,1.0);}
	void Init();
	void GetProjectionMatrix(float * projout) {glGetFloatv( GL_PROJECTION_MATRIX, projout );}
	void GetModelviewMatrix(float * modl) {glGetFloatv( GL_MODELVIEW_MATRIX, modl );}
	float * GetPrevModelView() {return prev_modelview;}
	float * GetPrevModelViewProj() {return prev_modelviewproj;}
	SCENEGRAPH & GetScene() {return scene;}
	void DrawScene();
	int resizeWindow(int width, int height, float fov);
	bool isExtensionSupported(string extstring);
	void BeginScene();
	void EndScene();
	void ExtractMatrices();
	void ExtractMatricesTo(float * modelviewout, float * modelviewprojout);
	void RecordFPSThisFrame(const float newfps);
	const float GetFPS() const {return fps;}
	const float GetTrueFPS() const {return truefps;}
	float GetW() {return width;}
	float GetH() {return height;}
	inline const QUATERNION & GetLightPosition() {return light_position;}
	void DrawList(list <SCENEDRAW> & drawlist, int passno, float lod_far);
	inline void RenderQuad() {RenderQuad(1.0);}
};

#define _GRAPHICS_H
#endif /* _GRAPHICS_H */
