#ifndef _MODEL_JOE03_H
#define _MODEL_JOE03_H

#include "pch.h"

/*#define dSINGLE
#define dTRIMESH_ENABLED
#include <ode/ode.h>*/

using namespace std;

#include "exception.h"
#include "joepack.h"
#include "3dmath.h"
#include "textures.h"

#define JOE_MAX_TRIANGLES       4096
#define JOE_MAX_VERTICES        2048
#define JOE_MAX_TEXCOORDS       2048
#define JOE_MAX_FACES			32000
#define JOE_MAX_FRAMES          1

#define JOE_VERSION 3

#define JOE_MAX_TEXTURES 1
#define MAX_TEXTURE_UNITS 4

#define TEXTUREMODE_NOTEX 0
#define TEXTUREMODE_TEX 1
#define TEXTUREMODE_REFLECTION 2
#define TEXTUREMODE_ADD 3

#define MODEL_SCALE 1.0

// This holds the header information that is read in at the beginning of the file
struct JOEHeader
{ 
	int magic;                   // This is used to identify the file
	int version;                 // The version number of the file
	int num_faces;	            // The number of faces (polygons)
	int num_frames;               // The number of animation frames
};

// This is used to store the vertices that are read in for the current frame
struct JOEVertex
{
	float vertex[3];
};

// This stores the indices into the vertex and texture coordinate arrays
struct JOEFace
{
	short vertexIndex[3];
	short normalIndex[3];
	short textureIndex[JOE_MAX_TEXTURES*3];
};

// This stores UV coordinates
struct JOETexCoord
{
	float u, v;
};

// This stores the frames vertices after they have been transformed
struct JOEFrame
{
	int num_verts;
	int num_texcoords;
	int num_normals;
	
	JOEFace * faces;
	JOEVertex * verts;
	JOEVertex * normals;
	JOETexCoord * texcoords;
};

// This holds all the information for our model/scene. 
struct JOEObject 
{
	JOEHeader info;
	JOEFrame * frames;
};

// This class handles all of the loading code
class JOEMODEL
{

public:
    JOEMODEL();                             // This inits the data members
	~JOEMODEL();

    bool Load(string strFileName) {return Load(strFileName, NULL);}
	bool Load(string strFileName, JOEPACK * pack) {return Load(strFileName, pack, false);}
	bool Load(string strFileName, JOEPACK * pack, bool adjustnormals);
	bool LoadFromHandle(FILE * f, JOEPACK * pack, bool adjustnormals);
	
	int BinaryRead(void * buffer, unsigned int size, unsigned int count, FILE * f, JOEPACK * pack);
	
	/*void Draw(int frame, float t);
	void Draw(int frame, int nextframe, float t);
	void DrawStatic(); //draw frame 0, optimized for speed

	void NewDraw(int frame, float t, VERTEX lightdir, QUATERNION rotation, int pass);
	void NewDraw(int frame, int nextframe, float t, VERTEX lightdir, QUATERNION rotation, int pass);

	void NoTexture(int tid) {if (tid < MAX_TEXTURE_UNITS) texturemode[tid] = TEXTUREMODE_NOTEX;}
	void Texture(string texturename, int tid);
	void ReflectionTexture(string texturename, int tid);
	void AdditiveTexture(string texturename, int tid);
	
	void TextureID(TEXTURE_HANDLE * texid, int tid) {if (tid < MAX_TEXTURE_UNITS) {texturemode[tid] = TEXTUREMODE_TEX; textureid[tid] = *texid; autounload[tid] = false;}}
	void ReflectionTextureID(TEXTURE_HANDLE * texid, int tid) {if (tid < MAX_TEXTURE_UNITS) {texturemode[tid] = TEXTUREMODE_REFLECTION; textureid[tid] = *texid; autounload[tid] = false;}}
	void AdditiveTextureID(TEXTURE_HANDLE * texid, int tid) {if (tid < MAX_TEXTURE_UNITS) {texturemode[tid] = TEXTUREMODE_ADD; textureid[tid] = *texid; autounload[tid] = false;}}
	
	void SetTU(int tuid, bool tue) {tuenable[tuid] = tue;}*/

	float GetRadius() {return radius + 0.5f;}
	float GetRadiusXZ() {return radiusxz;}
	
	unsigned int GetFaces() {if (!loadedfile) return 0; else return pObject->info.num_faces;}
	
	//bool Collide(VERTEX origin, VERTEX direction, VERTEX &outtri, bool closest);
	
	inline float * GetVert(short index) {return pObject->frames[0].verts[index].vertex;}
	short * GetFace(short index) {return pObject->frames[0].faces[index].vertexIndex;}
	
	inline float * GetNorm(short index) {return pObject->frames[0].normals[index].vertex;}
	short * GetNormIdx(short index) {return pObject->frames[0].faces[index].normalIndex;}
	
	unsigned int GetNumVerts() {if (!loadedfile) return 0; else return pObject->frames[0].num_verts;}
	
	AABB & GetBBOX() {return bbox;}
	
	bool Loaded() const {return loadedfile;}
	
	unsigned int GetDrawList() const {return static_list;}

private:
	JOEObject * pObject;

	bool loadedfile;
	
	AABB bbox;
	
	float radius; //used for frustum culling
	float radiusxz; //used to get an object's vertical profile
	
	GLuint static_list;
	
	void CorrectEndian(struct JOEFace * p, int num);
	void CorrectEndian(struct JOEVertex *p, int num);
	void CorrectEndian(struct JOETexCoord *p, int num);

	void ReadData(FILE *m_FilePointer, JOEPACK * pack, bool adjustnormals); // This reads in the data from the MD2 file and stores it in the member variables

	void CleanUp(FILE *m_FilePointer, JOEPACK * pack); // This frees memory and closes the file
	
	string modelpath;
	
	void Draw(int frame, int nextframe, float t);
};

#endif
