#ifndef _OBJECTS_H

#include "utility.h"
#include "3dmath.h"
#include "camera.h"
#include "model.h"
#include "joepack.h"
#include "bipointer.h"
#include "scenegraph.h"
#include "physics.h"

class OBJECTMODEL
{
public:
	JOEMODEL jmodel;
	string name;
	OBJECTMODEL * next;
	bool fullbright;
	bool skybox;
	bool blend;
};

class OBJECTNODE
{
public:
	VERTEX pos;
	bool driveable;
	bool cancollide;
	//OBJECTNODE * next;
	QUATERNION dir;
	OBJECTMODEL * model;
	string texture; //used to group and draw objects based on texture (for speed)
	float friction1; //the friction coefficient for non-treaded tires
	float friction2; //the friction coefficient for treaded tires
	float bumplength;
	float bumpmag;
	float rolling_resistance_factor;
	float rolling_drag;
	BIPOINTER <DRAWABLE> drawable;
	BIPOINTER <PHYSICSOBJECT> physicsobject;

	//data for collision engine
	PHYSICSOBJECTVERTEX * vertices;
	PHYSICSOBJECTFACE * faces;
	//PHYSICSOBJECTVERTEX * normals; //number of normals is fcount (normals are per face)
	int vcount;
	int fcount;

	OBJECTNODE();
	~OBJECTNODE() {DestroyCollisionData();DestroySceneData();}
	bool operator<(const OBJECTNODE & other) {return texture < other.texture;}

#ifdef WRAPPER
	string GetName() { return model->name; }
#endif
	void LoadModelIntoCollisionData();
	void DestroyCollisionData();
	void RegisterWithCollisionSystem();
	void DestroySceneData();
};

class OBJECTS
{
private:
	float lod_far;
	int display_x, display_y;
	float field_of_view;
	ofstream error_log;
	OBJECTMODEL * model_list;
	void delmodel();

	list <OBJECTNODE> object_list;

	OBJECTMODEL * AddModel(string modelname, string texname, bool mip, bool fullbright, bool skybox, bool blend, JOEPACK * pack);

	string path;

	//OBJECTCOLLISION collision;

	bool verticaltrackingskyboxes;
	
	BIPOINTER <SCENENODE> parentnode;
	
public:
	OBJECTS();
	~OBJECTS();
	void UpdateSettings();
	OBJECTNODE * Add(VERTEX pos, float rotation, string modelname, string texname, bool mip, bool fullbright, bool skybox, bool drv, bool col, bool blend, JOEPACK * pack, float f1, float f2, float bl, float bm, float rr, float rd, SCENENODE * parent);
	void DeleteAll();
	void LoadObjectsFromFolder(string objectpath);
	void SetVerticalSkyboxTracking(bool vtsky) {verticaltrackingskyboxes = vtsky;}

#ifdef WRAPPER
	list <OBJECTNODE>::iterator node_itr;
	OBJECTNODE* get_first_node() { node_itr = object_list.begin(); return &(*node_itr); }
	OBJECTNODE* get_next_node() { node_itr++; return (node_itr==object_list.end())?NULL:&(*node_itr);}
	unsigned int get_no_of_nodes() { return object_list.size(); }
#endif
	const SCENENODE & GetSceneNode() {return parentnode.GetRef();}
};

#define _OBJECTS_H
#endif
