#ifndef _SCENEGRAPH_H
#define _SCENEGRAPH_H

#include <stdio.h>

#include <stdlib.h>
#include <time.h>

#include <string>
#include <iostream>
#include <sstream>
#include <fstream>
#include <cmath>
#include <cassert>

#include <stdexcept>

#include <vector>
#include <list>
#include <map>

using namespace std;

#include "exception.h"
#include "tester.h"
#include "3dmath.h"
#include "textures.h"
#include "bipointer.h"
#include "vertexarray.h"
#include "allocating_bipointer.h"

class SCENENODE;

class DRAWABLE : public BIPOINTER_TARGET
{
friend class GRAPHICS;
friend class DRAWABLE_FILTER;
	
private:
	enum DRAWTYPE
	{
		NONE,
		LIST,
		VARRAY
	} type;
	unsigned int drawlist;
	VERTEXARRAY varray;
	ALLOCATING_BIPOINTER <TEXTURE_HANDLE> diffuse_map;
	ALLOCATING_BIPOINTER <TEXTURE_HANDLE> misc_map1;
	ALLOCATING_BIPOINTER <TEXTURE_HANDLE> additive_map1;
	SCENENODE * parent;
	bool decal;
	float r,g,b,a;
	bool lit;
	bool drawenabled;
	bool is2d;
	bool partial_transparency;
	bool cull;
	bool cull_front;
	float radius; //for bounding sphere
	VERTEX object_center; //used for bounding sphere frustum culling
	int draw_order;
	MATRIX4 last_modelviewproj; //used for blur effect only
	bool blur;
	bool metallic;
	bool skybox;
	bool self_illumination;
	bool issmoke;
	
	void SetDrawOrder(int neworder) {draw_order = neworder;}

public:
	DRAWABLE() : type(NONE),parent(NULL),decal(false),r(1.0),g(1.0),b(1.0),a(1.0),lit(true),drawenabled(true),is2d(false),partial_transparency(false),cull(false),cull_front(false),radius(0.0),draw_order(0),blur(true),metallic(false),skybox(false),self_illumination(false),issmoke(false) {}
	~DRAWABLE();
	/*void Set_list(const QUATERNION & rot, const VERTEX & trans, const unsigned int list) {rotation = rot;translation = trans;drawlist = list;type=LIST;}
	void Set_list(const unsigned int list) {QUATERNION r;VERTEX t;Set_list(r,t,list);}
	void Set_box(float * boxdata) {for (int i = 0; i < 4*3+4*2; i++) box[i] = boxdata[i]; type=BOX;}
	void Set_cal3d(const QUATERNION & rot, const VERTEX & trans, PLAYERMODEL_CAL3D * calmodel) {rotation = rot;translation = trans;cal3dmodel=calmodel;type=CAL3D;}*/

	const bool operator< (const DRAWABLE & other) const;

	void SetDiffuseMap(const TEXTUREINFO & tex) {diffuse_map.AllocateLocalStorage();diffuse_map.GetPtr()->Load(tex);}
	void SetDiffuseMapPtr(TEXTURE_HANDLE & tex) {diffuse_map.SetPtr(tex);}
	const string GetDiffuseMapName() const {return diffuse_map.GetPtr()->GetTexture().GetTextureInfo().GetName();}
	
	void SetMiscMap1(const TEXTUREINFO & tex) {misc_map1.AllocateLocalStorage();misc_map1.GetPtr()->Load(tex);}
	void SetAdditiveMap1(const TEXTUREINFO & tex) {additive_map1.AllocateLocalStorage();additive_map1.GetPtr()->Load(tex);}
	void SetMiscMap1Ptr(TEXTURE_HANDLE & tex) {misc_map1.SetPtr(tex);}
	const string GetMiscMap1Name() const {return misc_map1.GetPtr()->GetTexture().GetTextureInfo().GetName();}
	
	void Set(const DRAWABLE & other) {*this = other;}
	void SetToList(const unsigned int newlist) {drawlist = newlist;type=LIST;}
	VERTEXARRAY & SetToVertexArray() {type=VARRAY;return varray;}
	VERTEXARRAY & GetVertexArray() {if (type != VARRAY) throw EXCEPTION(__FILE__, __LINE__, "Tried to get vertex array from drawable of a different type"); return varray;}
	
	void SetParent(SCENENODE & newparent) {parent = &newparent;}
	SCENENODE * GetParent() {return parent;}
	void SetDecal(bool newdecal) {decal = newdecal;}
	void Set2D(bool new2d) {is2d = new2d;}
	void SetColor(float nr, float ng, float nb, float na) {r=nr;g=ng;b=nb;a=na;}
	void GetColor(float &nr, float &ng, float &nb, float &na) {nr=r;ng=g;nb=b;na=a;}
	void SetLit(bool newlit) {lit = newlit;}
	bool GetLit() const {return lit;}
	void SetDrawEnable(bool newenable) {drawenabled = newenable;}
	void SetPartialTransparency(bool newpt) {partial_transparency = newpt;}
	bool GetPartialTransparency() const {return partial_transparency;}
	void SetCull(bool newcull, bool newcullfront) {cull = newcull; cull_front = newcullfront;}
	void SetRadius(float newrad) {radius = newrad;} //is about object center (0,0,0)
	bool GetDrawEnable() const {return drawenabled;}
	void SetObjectCenter(VERTEX & no) {object_center = no;}
	const VERTEX & GetObjectCenter() const {return object_center;}
	void SetDrawOrderSmoke() {SetDrawOrder(1);}
	void SetDrawOrderFont() {SetDrawOrder(5);}
	void SetDrawOrderCursor() {SetDrawOrder(10);}
	void SetSmoke(bool newsmoke) {issmoke = newsmoke;}
	void SetDrawOrderGUIBackground() {SetDrawOrder(2);}
	void SetDrawOrderGUIForeground() {SetDrawOrder(4);}
	int GetOrder() {return draw_order;}
	void SetBlur(bool newblur) {blur=newblur;}
	void SetMetallic(bool newmetallic) {metallic=newmetallic;}
	void SetSkybox(bool newskybox) {skybox = newskybox;}
	bool GetSkybox() const {return skybox;}
	bool GetDecal() const {return decal;}
	void SetSelfIllumination(bool newsi) {self_illumination=newsi;}
};

class DRAWABLE_FILTER
{
private:
	bool is2d;
	bool is2d_filter;
	bool partial_transparency;
	bool partial_transparency_filter;
	bool blur;
	bool blur_filter;
	bool skybox;
	bool skybox_filter;
public:
	DRAWABLE_FILTER() : is2d(false),is2d_filter(false),partial_transparency(false),partial_transparency_filter(false),blur(false),blur_filter(false),skybox(false),skybox_filter(false) {}
	void SetFilter_is2d(bool filter_enable, bool filter_value) {is2d_filter=filter_enable;is2d=filter_value;}
	void SetFilter_partial_transparency(bool filter_enable, bool filter_value) {partial_transparency_filter=filter_enable;partial_transparency=filter_value;}
	void SetFilter_blur(bool filter_enable, bool filter_value) {blur_filter=filter_enable;blur=filter_value;}
	void SetFilter_skybox(bool filter_enable, bool filter_value) {skybox_filter=filter_enable;skybox=filter_value;}
	bool Matches(const DRAWABLE & drawable) const;
};

class SCENETRANSFORM
{
private:
	QUATERNION rotation;
	VERTEX translation;

public:
	const QUATERNION & GetRotation() const {return rotation;}
	const VERTEX & GetTranslation() const {return translation;}
	void SetRotation(const QUATERNION & rot) {rotation = rot;}
	void SetTranslation(const VERTEX & trans) {translation = trans;}
	bool IsIdentityTransform() const {return (rotation.IsIdentity() && translation.IsZero());}
	void Clear() {rotation.LoadMultIdent();translation.zero();}
};

class SCENEDRAW
{
private:
	const SCENETRANSFORM * transform;
	const DRAWABLE * draw;
	ALLOCATING_BIPOINTER <MATRIX4> mat4;

public:
	SCENEDRAW() : transform(NULL),draw(NULL) {}
	SCENEDRAW(const SCENEDRAW & other) : transform(other.transform),draw(other.draw),mat4(other.mat4) {}
	SCENEDRAW(const SCENETRANSFORM & newtransform) : transform(&newtransform),draw(NULL) {}
	SCENEDRAW(const DRAWABLE & newdraw) : transform(NULL),draw(&newdraw) {}
	SCENEDRAW(const DRAWABLE & newdraw, const MATRIX4 & newmat4) : transform(NULL),draw(&newdraw) {mat4.AllocateLocalStorage();mat4.GetPtr()->Set(newmat4);}
	const bool operator< (const SCENEDRAW & other) const;
	bool IsCollapsed() const {return (mat4.PtrValid() && draw != NULL);}
	bool IsTransform() const {return (transform != NULL);}
	bool IsDraw() const {return (draw != NULL);}
	bool IsEmpty() const {return !IsTransform() && !IsDraw();} //this is used to indicate a pop-transform
	const SCENETRANSFORM * GetTransform() const {return transform;}
	const DRAWABLE * GetDraw() const {return draw;}
	const MATRIX4 * GetMatrix4() const {return mat4.GetPtr();}
	void DebugPrint() const;
	void SetCollapsed(const DRAWABLE & newdraw, const MATRIX4 & newmat4) {draw = &newdraw;mat4.AllocateLocalStorage();mat4.GetPtr()->Set(newmat4);}
};

class SCENENODE : public BIPOINTER_TARGET
{
private:
	list <SCENENODE> childlist;
	list <DRAWABLE> drawlist;
	SCENETRANSFORM transform;

	SCENENODE * parent;

public:
	SCENENODE() : parent(NULL) {}
	~SCENENODE();
	SCENENODE & AddNode();
	DRAWABLE & AddDrawable();
	unsigned int Nodes() const {return childlist.size();}
	unsigned int Drawables() const {return drawlist.size();}
	SCENETRANSFORM & GetTransform() {return transform;}
	void GetDrawList(list <SCENEDRAW> & drawlistoutput, const DRAWABLE_FILTER & filter) const;
	void GetCollapsedDrawList(list <DRAWABLE_FILTER *> & filter_list, map < DRAWABLE_FILTER *, list <SCENEDRAW> > & drawlist_output_map, const MATRIX4 & prev_transform) const;
	void SetParent(SCENENODE & newparent) {parent = &newparent;}
	SCENENODE * GetParent() {return parent;}
	void Clear() {drawlist.clear();childlist.clear();}
	void Delete(SCENENODE * todelete);
	void Delete(DRAWABLE * todelete);
	VERTEX TransformIntoWorldSpace() const {VERTEX zero;return TransformIntoWorldSpace(zero);}
	VERTEX TransformIntoWorldSpace(const VERTEX & localspace) const;
	VERTEX TransformIntoLocalSpace(const VERTEX & worldspace) const;
	MATRIX4 CollapseTransform() const;
	void DebugPrint();
	void SortDrawablesByRenderState();
};

class SCENEGRAPH
{
private:
	SCENENODE rootnode;

public:
	SCENEGRAPH() {}
	~SCENEGRAPH() {}

	const TESTER Test();
	SCENENODE & GetRoot() {return rootnode;}
	//void GetDrawList(list <SCENEDRAW> & drawlist) const;
	void GetDrawList(list <DRAWABLE_FILTER *> & filter_list, map < DRAWABLE_FILTER *, list <SCENEDRAW> > & drawlist_output_map) const;
	void GetCollapsedDrawList(list <SCENEDRAW> & drawlist) const;
	void GetCollapsedDrawList(list <DRAWABLE_FILTER *> & filter_list, map < DRAWABLE_FILTER *, list <SCENEDRAW> > & drawlist_output_map) const;
	void Delete(SCENENODE * todelete);
	void Delete(DRAWABLE * todelete);
};

#endif
