/***************************************************************************
 *            utility.h
 *
 *  Sat Mar 26 08:48:31 2005
 *  Copyright  2005  Joe Venzon
 *  joe@venzon.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
  
#ifndef _UTILITY_H

#include "3dmath.h"
#include "cardinfo.h"
//#include "globals.h"
#include "scenegraph.h"
#include "bipointer.h"

using namespace std;

#define ENDOFFILESTRING "!!!END OF FILE!!!"

#define EPSILON 0.000001
//#define EPSILON 0.00001
#define CROSS(dest,v1,v2) \
          dest[0]=v1[1]*v2[2]-v1[2]*v2[1]; \
          dest[1]=v1[2]*v2[0]-v1[0]*v2[2]; \
          dest[2]=v1[0]*v2[1]-v1[1]*v2[0];
#define DOT(v1,v2) (v1[0]*v2[0]+v1[1]*v2[1]+v1[2]*v2[2])
#define SUB(dest,v1,v2) \
          dest[0]=v1[0]-v2[0]; \
          dest[1]=v1[1]-v2[1]; \
          dest[2]=v1[2]-v2[2]; 
          
#ifdef ENABLE_NLS
	#include <libintl.h>
	#include <locale.h>
	#define _(string) gettext (string.c_str ())
	#define _c(string) gettext (string)
#else
	#define _(string) (string)
	#define _c(string) (string)
#endif

#define LISTITERATE(ttype,target,it) for(list <ttype>::iterator it = target.begin(); it != target.end(); it++)
#define VECTORITERATE(ttype,target,it) for(vector <ttype>::iterator it = target.begin(); it != target.end(); it++)

//#define TEST_CULL

class TEXTURE_HANDLE;
class SCENENODE;

class ELEMENT2D
{
private:
	enum TYPE2D
	{
		NONE,
		BOX,
		BUTTON,
		ELLIPSE,
		QUAD
	} type;
	
	float x1;  //for ellipse, this is the center
	float y1;
	float x2;  //for ellipse, this is the radius
	float y2;
	float x3;  //for quad, this is the 3rd vertex
	float y3;
	float x4;  //for quad, this is the 4th vertex
	float y4;
	int texsize;
	float opacity;
	float sidewidth; // used for button
	float color_r, color_g, color_b; //used for ellipse
	float rotation;
	
	bool draw;
	
	TEXTUREINFO texture;
public:
	ELEMENT2D() : type(NONE), x1(0), y1(0), x2(0), y2(0), x3(0), y3(0), x4(0), y4(0), texsize(128), opacity(1), sidewidth(0), color_r(1), color_g(1),color_b(1),rotation(0.0),draw(false) {}
	const bool operator<(const ELEMENT2D& other) const;
	
	void SetBox(float x1, float y1, float x2, float y2, const TEXTURE_HANDLE * const texid, int texsize, float opacity);
	void SetButton(float x1, float y1, float x2, float y2, float sidewidth, TEXTURE_HANDLE * texid, float opacity);
	void SetEllipse( float center_x, float center_y, float radius_x, float radius_y, float color_r, float color_g, float color_b, float opacity );
	void SetQuad( float x1, float y2, float x2, float y2, float x3, float y3, float x4, float y4, TEXTURE_HANDLE * texid, float opacity);
	void SetRotation(float r) {rotation=r;}
	
	void SetDraw(const bool newdraw) {draw=newdraw;}
	const bool GetDraw() const {return draw;}
	
	void DebugPrint() const;
};

class UTILITY
{
private:
	ofstream error_log;
	GLint nb_multitexture;
	bool initdone;
	string texture_size;
	void initerror();
	
	map < ELEMENT2D, BIPOINTER <DRAWABLE> > draw2d;
	SCENENODE * parentnode;
	
public:
	UTILITY();
	~UTILITY();
	void Init();
	string sGetLine(ifstream &ffrom);
	string sGetParam(ifstream &ffrom);
	int iGetParam(ifstream &ffrom);
	float fGetParam(ifstream &ffrom);
	bool bGetParam(ifstream &ffrom);
	string GetEOFString() {return ENDOFFILESTRING;}
	float GetValue(SDL_Surface * surf, int channel, float x, float y, bool interpolate);
	float GetValue(SDL_Surface * surf, int channel, float x, float y, bool interpolate, bool wrap);
	bool FileExists(string filename);
	int IntersectTriangleD(double orig[3], double dir[3],
                   double vert0[3], double vert1[3], double vert2[3],
                   double *t, double *u, double *v);
	int IntersectTriangleF(float orig[3], float dir[3],
                   float vert0[3], float vert1[3], float vert2[3],
                   float *t, float *u, float *v);
	bool IntersectQuadrilateralF(VERTEX orig, VERTEX dir,
					VERTEX v_00, VERTEX v_10, VERTEX v_11, VERTEX v_01,
					float &t, float &u, float &v);
	int BruteForceTriangleIntersectionF(VERTEX * tri1, VERTEX * tri2, 
					VERTEX & colpt, VERTEX & colseg, VERTEX & coldest, int & whichtri);
	int BruteForceTriangleIntersectionF2(VERTEX * tri1, VERTEX * tri2, 
					VERTEX & colpt, VERTEX & colseg, VERTEX & coldest, int & whichtri,
					float & destdepth, float & origdepth);
	int MollerTriTriIsectWithLine(float * V0, float * V1, float * V2,
					float * U0, float * U1, float * U2, int * coplanar,
					float * isectpt1, float * isectpt2);
	bool FileCopy(const string& srcfile, const string& dstfile) const;
	string GetPathFromFilename(string filename);
	bool isNaN(float val);
	bool DirExists(string folderpath);
	bool GetFolderIndex(string folderpath, list <string> & outputfolderlist);
	double dabs(double val);
	
	void Draw2D(float x1, float y1, float x2, float y2, const TEXTURE_HANDLE * const texid);
	void Draw2D(float x1, float y1, float x2, float y2, const TEXTURE_HANDLE * const texid, int texsize);
	void Draw2D(float x1, float y1, float x2, float y2, const TEXTURE_HANDLE * const texid, int texsize, float opacity);
	void Draw2D(float x1, float y1, float x2, float y2, const TEXTURE_HANDLE * const texid, int texsize, float opacity, bool cursor, bool guibg, bool guifg);
	void Draw2D(float x1, float y1, float x2, float y2, const TEXTURE_HANDLE * const texid, int texsize, float opacity, bool cursor, bool guibg, bool guifg, float rotation);
	//void Draw2DRotation(float x1, float y1, float x2, float y2, const TEXTURE_HANDLE * const texid, float r);
	void DrawButton(float x1, float y1, float x2, float y2, float sidewidth, TEXTURE_HANDLE * texid, float opacity);
	void DrawQuad(float x1, float y1, float x2, float y2, float x3, float y3, float x4, float y4, TEXTURE_HANDLE * texid, float opacity);
	void DrawEllipse( float center_x, float center_y, float radius_x, float radius_y ) { DrawEllipse( center_x, center_y, radius_x, radius_y, 0.0f, 0.0f, 0.0f ); }
	void DrawEllipse( float center_x, float center_y, float radius_x, float radius_y, float color_r, float color_g, float color_b ) { DrawEllipse( center_x, center_y, radius_x, radius_y, color_r, color_g, color_b, 1.0f ); }
	void DrawEllipse( float center_x, float center_y, float radius_x, float radius_y, float color_r, float color_g, float color_b, float opacity );
	void Clear2D();
	void UnDraw2D();
	void SetVertexData2DQuad(float x1, float y1, float x2, float y2, float u1, float v1, float u2, float v2, float * vcorners, float * uvs, int * bfaces) const {SetVertexData2DQuad(x1,y1,x2,y2,u1,v1,u2,v2,vcorners,uvs,bfaces,0);}
	void SetVertexData2DQuad(float x1, float y1, float x2, float y2, float u1, float v1, float u2, float v2, float * vcorners, float * uvs, int * bfaces, int faceoffset) const;
	void SetVertexData2DQuad(float x1, float y1, float x2, float y2, float x3, float y3, float x4, float y4, float u1, float v1, float u2, float v2, float * vcorners, float * uvs, int * bfaces, int faceoffset) const;
};

#define _UTILITY_H
#endif /* _UTILITY_H */
