#include "cardinfo.h"
#include "game.h"
extern GAME game;

// This function returns true if the extension is there.
bool CARDINFO::isExtensionSupported(string extstring)
{
	char * temp = (char *) glGetString(GL_EXTENSIONS);
	if (temp == 0)
	{
		cout << "Error getting extensions.  Continuing anyway, errors may follow!" << endl;
		return true;
	}

	string s = temp;
	string::size_type temppos = s.find(extstring);

	bool hasext = (temppos <= s.length());

	if (!hasext)
	{
		cout << "Extension not supported: " << extstring << endl;
	}

	//cout << s << endl;

	return hasext;
}

void CARDINFO::ResetDB()
{
	db.resize(CARDINFOTYPE::NUM_CARDINFOTYPES);
	dbvalid = false;
}

int CARDINFO::GetCapability(CARDINFOTYPE::TYPE capability)
{
	if (!dbvalid)
		return -1;
	
	if (capability >= 0 && capability < CARDINFOTYPE::NUM_CARDINFOTYPES)
	{
		return db[capability].data;
	}
	
	return -1;
}

void CARDINFO::BuildCardInfo()
{
	stringstream ms;
	stringstream md;
	
	ms << "Card supports:";
	md << "Card does not support:";
	
	GLint maxdrawbuf, auxbuf;
	glGetIntegerv(GL_MAX_DRAW_BUFFERS,&maxdrawbuf);
	ms << " drawbuf" << maxdrawbuf;
	db[CARDINFOTYPE::MAXDRAWBUFFERS].data = maxdrawbuf;
	glGetIntegerv(GL_AUX_BUFFERS,&auxbuf);
	ms << " auxbuf" << auxbuf;
	db[CARDINFOTYPE::AUXBUFFERS].data = auxbuf;
	
	if (isExtensionSupported( "GL_ARB_multisample" ))
	{
		db[CARDINFOTYPE::ANTIALIASING].data = 1;
		ms << " antialiasing";
	}
	else
	{
		db[CARDINFOTYPE::ANTIALIASING].data = 0;
		md << " antialiasing";
	}
	
	if (isExtensionSupported( "GL_EXT_texture_filter_anisotropic" ))
	{
		db[CARDINFOTYPE::ANISOTROPY].data = 1;
		GLint aniso;
		glGetIntegerv( GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT,&aniso );
		db[CARDINFOTYPE::MAXANISOTROPY].data = aniso;
		ms << " anisotropy" << aniso;
	}
	else
	{
		db[CARDINFOTYPE::ANISOTROPY].data = 0;
		db[CARDINFOTYPE::MAXANISOTROPY].data = -1;
		md << " anisotropy";
	}
	
	if (isExtensionSupported( "GL_ARB_texture_cube_map" ))
	{
		db[CARDINFOTYPE::CUBEMAPPING].data = 1;
		ms << " cubemapping";
	}
	else
	{
		db[CARDINFOTYPE::CUBEMAPPING].data = 0;
		md << " cubemapping";
	}
	
	if (isExtensionSupported( "GL_ARB_fragment_program" ) && isExtensionSupported( "GL_ARB_vertex_program" ))
	{
		db[CARDINFOTYPE::FRAGMENTSHADERS].data = 1;
		ms << " shaders";
	}
	else
	{
		db[CARDINFOTYPE::FRAGMENTSHADERS].data = 0;
		md << " shaders";
	}
	
	if (isExtensionSupported( "GL_ARB_multitexture" ))
	{
		db[CARDINFOTYPE::MULTITEXTURE].data = 1;
		GLint tu;
		if (db[CARDINFOTYPE::FRAGMENTSHADERS].data)
		{
			glGetIntegerv( GL_MAX_TEXTURE_IMAGE_UNITS_ARB,&tu );
			GLint tufull;
			glGetIntegerv( GL_MAX_TEXTURE_UNITS_ARB,&tufull );
			ms << " multitexturing" << tu << "(" << tufull << ")";
		}
		else
		{
			glGetIntegerv( GL_MAX_TEXTURE_UNITS_ARB,&tu );
			ms << " multitexturing" << tu;
		}
		
		db[CARDINFOTYPE::TEXTUREUNITS].data = tu;
	}
	else
	{
		db[CARDINFOTYPE::MULTITEXTURE].data = 0;
		db[CARDINFOTYPE::TEXTUREUNITS].data = 1;
		md << " multitexturing";
	}
	
	if (isExtensionSupported( "GL_ARB_texture_rectangle" ))
	{
		db[CARDINFOTYPE::TEXTURERECTANGLE].data = 1;
		ms << " texture_rectangle";
	}
	else
	{
		db[CARDINFOTYPE::TEXTURERECTANGLE].data = 0;
		md << " texture_rectangle";
	}
	
	if (isExtensionSupported( "GL_ARB_depth_texture" ))
	{
		db[CARDINFOTYPE::DEPTHTEXTURE].data = 1;
		ms << " depth_texture";
	}
	else
	{
		db[CARDINFOTYPE::DEPTHTEXTURE].data = 0;
		md << " depth_texture";
	}
	
	if (isExtensionSupported( "GL_ARB_shadow" ))
	{
		db[CARDINFOTYPE::SHADOW].data = 1;
		ms << " shadow";
	}
	else
	{
		db[CARDINFOTYPE::SHADOW].data = 0;
		md << " shadow";
	}
	
	if (isExtensionSupported( "GL_EXT_framebuffer_object" ))
	{
		db[CARDINFOTYPE::FRAMEBUFFER_OBJECTS].data = 1;
		ms << " framebuffer_objects";
	}
	else
	{
		db[CARDINFOTYPE::FRAMEBUFFER_OBJECTS].data = 0;
		md << " framebuffer_objects";
	}
	
	game.WriteDebuggingData(ms.str());
	game.WriteDebuggingData(md.str());
	
	dbvalid = true;
}

void CARDINFO::PrintCardInfo(ostream & out)
{
	/*if (dbvalid)
	{
		out << "Multitexture: ";
		if (db[CARDINFOTYPE::MULTITEXTURE].data == 1)
			out << "Yes" << endl;
		else
			out << "No" << endl;
		out << "Texture units: " << db[CARDINFOTYPE::TEXTUREUNITS].data << endl;
		out << "Antialiasing: ";
		if (db[CARDINFOTYPE::ANTIALIASING].data == 1)
			out << "Yes" << endl;
		else
			out << "No" << endl;
		out << "Anisotropic texture filtering: ";
		if (db[CARDINFOTYPE::ANISOTROPY].data == 1)
		{
			out << "Yes" << endl;
			out << "Maximum Anisotropy: " << db[CARDINFOTYPE::MAXANISOTROPY].data << endl;
		}
		else
			out << "No" << endl;
	}
	else
		cout << "Card info not yet built" << endl;*/
}
