#include "font.h"
#include "common.h"
#include "game.h"
extern GAME game;

const bool FONT::Load(const string & newfontname, SCENENODE & parentnode)
{
	if (loaded)
		return true;
	
	sceneparentnode = &parentnode;
	
	fontname = newfontname;
	
	ifstream f;
	string infopath;
	infopath = game.settings.GetDataDir() + "/" + game.settings.GetSkinPath() + "/fonts/"+newfontname+".txt";
	f.open(infopath.c_str());
	if (!f)
	{
		throw EXCEPTION(__FILE__, __LINE__, "Can't load font info file " + infopath + " for font " + newfontname);
		return false;
	}
	
	string tstr;
	//int tint;
	
	f >> tstr;
	
	if (tstr != "VDriftFontV1")
	{
		throw EXCEPTION(__FILE__, __LINE__, "Can't load font info file " + infopath + " for font " + newfontname);
		return false;
	}
	
	//read header
	int pages;
	f >> pages;
	f >> totalx >> totaly;
	f >> tw >> th;
	f >> fontsize;
	
	float minminy, maxmaxy;
	minminy = 0;
	maxmaxy = 0;
	
	for (int i = 32; i < totalx*totaly; i++)
	{
		int c;
		f >> c;
		charinfo[i-32].character = i;
		f >> charinfo[i-32].miny;
		f >> charinfo[i-32].maxy;
		f >> charinfo[i-32].minx;
		f >> charinfo[i-32].maxx;
		
		if (charinfo[i-32].miny < minminy)
			minminy = charinfo[i-32].miny;
		
		if (charinfo[i-32].maxy > maxmaxy)
			maxmaxy = charinfo[i-32].maxy;
	}
	
	averagecenter = (maxmaxy + minminy)/2.0;
	
	//TODO:  change this to "data" ... etc
	//fonttexture.Load(game.pathmanager.GetPath("data","font","default","texture","font"+newfontname+".png"));
	//fonttexture.Load("font/"+newfontname+".png"));
	//tinfo.SetName(game.settings.GetFullDataPath("font/"+newfontname+".png"));
	tinfo.SetName(game.settings.GetSkinPath() + "/fonts/"+newfontname+".png");
	
	//fonttexture.Load(tinfo);
	//drawable.SetPtr(game.graphics.GetScene().GetRoot().AddDrawable());
	//varray.SetPtr(drawable.GetPtr()->SetToVertexArray());
	//drawable.GetPtr()->SetDiffuseMap(tinfo);
	//drawable.GetPtr()->SetFont(true);
	//drawable.GetPtr()->SetLit(false);
	
	loaded = true;
	
	return true;
}

const bool FONTMANAGER::LoadFont(const string & newfontname)
{
	if (!fontrootnode.PtrValid())
		fontrootnode.SetPtr(game.graphics.GetScene().GetRoot().AddNode());
	
	ITERATE(FONT, fontlist, i)
	{
		if (i->GetFontName() == newfontname)
			return true;
	}
	
	{
		//FONT * newfontptr = fontlist.DB_Add();
		FONT newfontref;
		fontlist.push_back(newfontref);
	}
	FONT & newfontref = fontlist.back();
	return newfontref.Load(newfontname, fontrootnode.GetRef());
	//newfontptr->Load(newfontname);
}

const float FONT::Print(const float px, const float py, const string & str, const float scale, const float r, const float g, const float b, const float a, const bool draw)
{
	float textwidth = 0.0;
	
	/*if (draw)
	{
		fonttexture.Activate();
	}*/
	
	//drawable.GetPtr()->SetColor(r,g,b,a);
	
	FONTCOLOR c;
	c.r = r;
	c.g = g;
	c.b = b;
	c.a = a;
	
	DRAWABLE & d = GetDrawable(c);
	VERTEXARRAY & v = d.GetVertexArray();
	
	for (unsigned int i = 0; i < str.length(); i++)
	{
		//cout << px+textwidth << "," << py << endl;
		textwidth += DrawChar(str[i], scale, draw, px+textwidth, py, v);
		//cout << textwidth << endl;
	}
	
	return textwidth;
}

//float FONTMANAGER::Print(string & name, float px, float py, string & str, float scale, float r, float g, float b, float a)
const float FONTMANAGER::Print(const string & name, const float px, const float py, const string & str, const float scale, const float r, const float g, const float b, const float a)
{
	/*int w = 1;
	int h = 1;*/
	/*RENDERRECT defaultrect;
	defaultrect.Set2D(0, 0, w, h, 0);*/
	
	//setup flags
	{
		//float x = (px*(float)w);
		//float y = ((1.0-py)*(float)h);//-48*scale;
		
		{
			//comment the following back in
			/*game.graphics.DepthTest(false);
			game.graphics.Lighting(false);
			game.graphics.Blend(true);
			game.graphics.BlendOnAlpha();
			game.graphics.TextureClamp();
			game.graphics.SetupOrthoTarget(w, h);
			game.graphics.SetColor(r,g,b,a);*/
		}
	}
	
	float retval = 0;
	//game.graphics.BeginFont(r,g,b,a);
	
	//float retval = Print(name, px, py, str, scale, r, g, b, a, defaultrect);
	retval = GetFont(name)->Print(px, (1.0-py), str, scale, r, g, b, a, true);
	
	//game.graphics.EndFont();
	
	//clear flags
	{
		//comment the following back in
		/*
		game.graphics.ClearOrthoTarget();
		
		game.graphics.DepthTest(true);
		game.graphics.Lighting(true);
		game.graphics.Blend(false);*/
	}
	
	return retval;
}

const float FONTMANAGER::PrintWidth(const string & name, const string & str, const float scale)
{
	/*RENDERRECT nullrect;
	nullrect.Set2D(0,0,1,1,0);*/
	return GetFont(name)->Print(0, 0, str, scale, 1,1,1,1, false);
}

//float FONT::DrawChar(char curchar, float scale, bool draw, float & px, float & py, RENDERRECT & dest)
const float FONT::DrawChar(const char curchar, const float scale, const bool draw, const float px, const float py, VERTEXARRAY & v)
{
	float charwidth = 0;
	
	int c = curchar;
	if (c == 10)
		c = 32;
	c -= 32;
	if (c < 0 || c > 96)
	{
		stringstream err;
		err << "Character out of range: " << curchar << " (" << (int) curchar << ")" << endl;
		throw EXCEPTION(__FILE__, __LINE__, err.str());
		return 0.0;
	}
	int x = c % totalx;
	int y = c / totalx;
	
	float u1 = (float) x / (float) totalx;
	float v2 = (float) y / (float) totaly;
	
	float u2 = (float) (x+1) / (float) totalx;
	float v1 = (float) (y+1) / (float) totaly;
	
	float w = (1.0/(float)totalx)*tw; //w is the texture size in pixels
	float h = (1.0/(float)totaly)*th; //h is the texture size in pixels
	
	float realscale = 0.0008*scale; //scale things down to the 0.0 to 1.0 range
	//float realscale = 0.8*scale;
	//float hwratio = (float) game.graphics.GetH()/game.graphics.GetW();
	//float hwratio = (float) game.graphics.GetW()/game.graphics.GetH();
	float hwratio = 1.333333333;
	//cout << "*******" << hwratio << endl;
	
	if (c == 0)
	{
		//if (draw)
			//game.graphics.Translate(8.0*scale,0,0);
		//return 8.0*scale;
		
		//px += 8.0*realscale;
		return 8.0*realscale;
	}
	
	float offsetx = (charinfo[c].maxx+charinfo[c].minx)/2-w/2;
	float offsety = (charinfo[c].maxy+charinfo[c].miny)/2-h/2;
	
	float space = charinfo[c].maxx + 2.0;
	
	float x1, x2, y1, y2;
	x1 = px + offsetx*realscale;
	y1 = py + offsety*realscale*hwratio;
	x2 = x1+realscale*w;
	y2 = y1+realscale*h*hwratio;
	
	y1 -= 0.024;
	y2 -= 0.024;
	
	/*float vcorners[4][3];
	for (int i = 0; i < 4; i++)
		vcorners[i][2] = 0;
	vcorners[0][0] = x1; vcorners[0][1] = y1;
	vcorners[1][0] = x2; vcorners[1][1] = y1;
	vcorners[2][0] = x2; vcorners[2][1] = y2;
	vcorners[3][0] = x1; vcorners[3][1] = y2;*/
	
	float vcorners[12];
	vcorners[0] = x1;
	vcorners[1] = y1;
	vcorners[2] = 0;
	vcorners[3] = x2;
	vcorners[4] = y1;
	vcorners[5] = 0;
	vcorners[6] = x2;
	vcorners[7] = y2;
	vcorners[8] = 0;
	vcorners[9] = x1;
	vcorners[10] = y2;
	vcorners[11] = 0;
	
	/*for (int i = 0; i < 4; i++)
	{
		vcorners[i][0] = vcorners[i][0]*game.graphics.GetW();
		vcorners[i][1] = vcorners[i][1]*game.graphics.GetH();
	}*/
	
	/*float uvs[4][2];
	uvs[0][0] = u1; uvs[0][1] = v1;
	uvs[1][0] = u2; uvs[1][1] = v1;
	uvs[2][0] = u2; uvs[2][1] = v2;
	uvs[3][0] = u1; uvs[3][1] = v2;*/
	
	float uvs[8];
	uvs[0] = u1;
	uvs[1] = v1;
	uvs[2] = u2;
	uvs[3] = v1;
	uvs[4] = u2;
	uvs[5] = v2;
	uvs[6] = u1;
	uvs[7] = v2;
	
	int bfaces[6];
	bfaces[0] = 0;
	bfaces[1] = 1;
	bfaces[2] = 2;
	bfaces[3] = 0;
	bfaces[4] = 2;
	bfaces[5] = 3;
	
	/*for (int i = 0; i < 4; i++)
	{
		cout << i << ". " << vcorners[i][0] << "," << vcorners[i][1] << " -uv- " << uvs[i][0] << "," << uvs[i][1] << endl;
	}*/
	
	if (draw)
	{
		//game.graphics.DrawBox(vcorners, uvs);
		VERTEXARRAY thischar;
		thischar.SetFaces(bfaces, 6);
		thischar.SetVertices(vcorners, 12);
		thischar.SetTexCoordSets(1);
		thischar.SetTexCoords(0, uvs, 8);
		
		//VERTEXARRAY & v = *(varray.GetPtr());
		//v = v + thischar;
		
		v = v + thischar;
		
		/*const int * f;
		int nf;
		v.GetFaces(f, nf);
		cout << nf / 3 << endl;*/
	}
	
	//charwidth += offsetx*realscale+(space-offsetx)*realscale;
	charwidth += space*realscale;
	
	//px += space*realscale;
	
	return charwidth;
}

const float FONTMANAGER::GetAverageVerticalPosition(const string & name)
{
	return GetFont(name)->GetAverageCenter();
}

FONT * const FONTMANAGER::GetFont(const string & name)
{
	//for (GAMEDB::NODELIST_ITERATOR <FONT> i (fontlist); i.NotEnd(); i.Inc())
	ITERATE(FONT, fontlist, i)
	{
		if (i->GetFontName() == name)
		{
			return &(*i);
		}
	}
	
	throw EXCEPTION(__FILE__, __LINE__, "Font not found: "+name);
	
	return NULL;
}

const int FONTMANAGER::GetFontSize(const string & name)
{
	return GetFont(name)->GetFontSize();
}

/*float FONTMANAGER::GetNormalizedFontHeight(string & name, float scale, RENDERRECT & dest)
{
	FONT * f = GetFont(name);
	//float retheight = f->fontsize * 0.0008 * scale;
	//cout << retheight << endl;
	float retheight = (1.0/(float)f->totaly)*f->th*0.0008*scale*dest.GetRatio();
	//cout << retheight << "*" << endl;
	return retheight;
}*/

void FONTMANAGER::ClearAllText()
{
	ITERATE(FONT, fontlist, i)
	{
		i->ClearAllText();
	}
}

void FONT::ClearAllText()
{
	/*ITERMAPNOCONST(FONTCOLOR,BIPOINTER <DRAWABLE>,drawable_map,i)
	{
		VERTEXARRAY & v = i->second.GetPtr()->GetVertexArray();
		v.Clear();
	}*/
	
	/*drawable_map.clear();
	sceneparentnode->Clear();*/
	
	//cout << drawable_list.size() << endl;
	
	ITERATE(BIPOINTER <DRAWABLE>, drawable_list, i)
	{
		i->GetPtr()->SetDrawEnable(false);
		i->GetPtr()->GetVertexArray().Clear();
	}
}

DRAWABLE & FONT::GetDrawable(const FONTCOLOR & color)
{
	//DRAWABLE * drawable(NULL);	
	DRAWABLE * newdrawablecandidate(NULL);

	//for(list < BIPOINTER<DRAWABLE> >::iterator it = drawable_list.begin(); it != drawable_list.end() && !drawable; ++it)
	ITERATE(BIPOINTER<DRAWABLE>,drawable_list,it)
	{
		if (it->GetPtr()->GetDrawEnable())
		{
			FONTCOLOR drawablecolor;
			it->GetPtr()->GetColor(drawablecolor.r,drawablecolor.g,drawablecolor.b,drawablecolor.a);
			if (color == drawablecolor)
				return it->GetRef();
		}
		else
			newdrawablecandidate = it->GetPtr();
	}
	
	//couldn't find this font color already drawn, but there are empty drawables for use
	if (newdrawablecandidate)
	{
		newdrawablecandidate->SetColor(color.r, color.g, color.b, color.a);
		newdrawablecandidate->SetDrawEnable(true);
		return *newdrawablecandidate;
	}
	
	//our existing list of drawables is full
	drawable_list.push_back(BIPOINTER<DRAWABLE>());
	drawable_list.back().SetPtr(sceneparentnode->AddDrawable());
	newdrawablecandidate = drawable_list.back().GetPtr();
	newdrawablecandidate->SetToVertexArray();
	newdrawablecandidate->SetDiffuseMap(tinfo);
	newdrawablecandidate->Set2D(true);
	newdrawablecandidate->SetDrawOrderFont();
	newdrawablecandidate->SetLit(false);
	return *newdrawablecandidate;
	
	/*//try to insert the element
	pair <map < FONTCOLOR, BIPOINTER <DRAWABLE> >::iterator, bool> result = drawable_map.insert(pair < FONTCOLOR, BIPOINTER <DRAWABLE> > (color, BIPOINTER <DRAWABLE> ()));
	map < FONTCOLOR, BIPOINTER <DRAWABLE> >::iterator i = result.first;
	
	//if the insertion resulted in a new object being created, set it up
	if (result.second)
	{
		if (sceneparentnode == NULL)
			throw EXCEPTION(__FILE__, __LINE__, "Font has no parent scene node");
		DRAWABLE & drawable = sceneparentnode->AddDrawable();
		i->second.SetPtr(drawable);
		
		drawable.SetToVertexArray();
		drawable.SetDiffuseMap(tinfo);
		drawable.SetFont(true);
		drawable.SetLit(false);
		drawable.SetColor(color.r, color.g, color.b, color.a);
		
		//cout << "Created drawable, color " << color.r << "," << color.g << "," << color.b << "," << color.a << endl;
	}
	//else cout << "no need to create drawable, can use existing " << color.r << "," << color.g << "," << color.b << "," << color.a << endl;
	
	return i->second.GetRef();*/
}

void FONTMANAGER::Load()
{
	ClearFonts();
	
	//cout << "***Loading fonts..." << endl;
	if (!LoadFont("default"))
	{
		throw EXCEPTION(__FILE__,__LINE__,"Error loading font");
	}
	
	//cout << "********** loaded font default" << endl;
	
	if (!LoadFont("digital"))
	{
		throw EXCEPTION(__FILE__,__LINE__,"Error loading font");
	}
	
	/*if (!LoadFont("default"))
	{
		cerr << __FILE__ << "," << __LINE__ << ": " << "Error loading font" << endl;
	}*/
}

const string FONTMANAGER::GetBackwardCompatibilityFontName(const int set) const
{
	string fontname = "default";
	if (set == 0)
		fontname = "digital";
	return fontname;
}

const float FONTMANAGER::GetBackwardCompatibilityScale(const int size) const
{
	return (((float) size - 7.0f) * 0.25f) + 1.0f;
}

void FONTMANAGER::Print( float px, float py, const char *str, int set, int size, float r, float g, float b, float trans )
{
	string tempstr = str;
	string fontname = GetBackwardCompatibilityFontName(set);
	float scale = GetBackwardCompatibilityScale(size);
	Print(fontname, px, py, tempstr, scale, r, g, b, trans);
}

const float FONTMANAGER::Width(const char * str, int set, int size)
{
	string tempstr = str;
	string fontname = GetBackwardCompatibilityFontName(set);
	float scale = GetBackwardCompatibilityScale(size);
	return PrintWidth(fontname, tempstr, scale);
}

float FONTMANAGER::Height(const char * str, int set, int size)
{
	string fontname = GetBackwardCompatibilityFontName(set);
	float scale = GetBackwardCompatibilityScale(size);
	return GetFontHeight(fontname, scale);
}

const float FONTMANAGER::GetFontHeight(const string & name, const float scale)
{
	FONT * f = GetFont(name);
	if (!f) return 0.0;
	//float retheight = f->fontsize * 0.0008 * scale;
	//cout << retheight << endl;
	//float retheight = (1.0/(float)f->totaly)*f->th*0.0008*scale*dest.GetRatio();
	//float retheight = (1.0/(float)f->totaly)*f->th*0.0008*scale;
	//cout << retheight << "*" << endl;
	
	//return retheight;
	
	return f->GetFontSize()*0.0008;
}

#define FONTCOLOR_COMPARE_RETURN(var) if (var != other.var) return (var < other.var);

const bool FONTCOLOR::operator<(const FONTCOLOR& other) const
{
	FONTCOLOR_COMPARE_RETURN(r)
	FONTCOLOR_COMPARE_RETURN(g)
	FONTCOLOR_COMPARE_RETURN(b)
	FONTCOLOR_COMPARE_RETURN(a)
	return false;
	
	//return (memcmp(this,&other,sizeof(other))<0);
}
