#include "graphics.h"
#include "exception.h"
#include "game.h"
extern GAME game;
#include "common.h"

#ifdef _WIN32
PFNGLMULTITEXCOORD2FARBPROC     pglMultiTexCoord2f     = NULL;
PFNGLMULTITEXCOORD4FARBPROC     pglMultiTexCoord4f     = NULL;
PFNGLCLIENTACTIVETEXTUREARBPROC	pglActiveTexture       = NULL;
PFNGLACTIVETEXTUREARBPROC       pglClientActiveTexture = NULL;
#endif

void GRAPHICS::Init()
{
	#ifdef _WIN32
	pglMultiTexCoord4f = (PFNGLMULTITEXCOORD4FARBPROC) wglGetProcAddress("glMultiTexCoord4fARB");
	pglMultiTexCoord2f = (PFNGLMULTITEXCOORD2FARBPROC) wglGetProcAddress("glMultiTexCoord2fARB");
	pglActiveTexture = (PFNGLCLIENTACTIVETEXTUREARBPROC) wglGetProcAddress("glActiveTextureARB");
	pglClientActiveTexture = (PFNGLACTIVETEXTUREARBPROC) wglGetProcAddress("glClientActiveTextureARB");
	
	if (pglMultiTexCoord4f == NULL)
		cout << "main, WIN32: wglGetProcAddress(\"glMultiTexCoord4fARB\") failed." << endl;
	if (pglMultiTexCoord2f == NULL)
		cout << "main, WIN32: wglGetProcAddress(\"glMultiTexCoord2fARB\") failed." << endl;
	if (pglActiveTexture == NULL)
		cout << "main, WIN32: wglGetProcAddress(\"glActiveTextureARB\") failed." << endl;
	if (pglClientActiveTexture == NULL)
		cout << "main, WIN32: wglGetProcAddress(\"glClientActiveTextureARB\") failed." << endl;
		
		//cout << "utility init test" << endl;
	#endif

	game.settings.Get("display.shaders", shaders_enabled);

//#if defined(WIN32) || defined(__APPLE__)
	GLenum glew_err = glewInit();
	if( glew_err != GLEW_OK )
	{
		stringstream msg;
		msg << "GLEW failed to initialize: " << glewGetErrorString( glew_err );
		game.WriteDebuggingData(msg.str());
		shaders_enabled = false;
	}
	
	{
		stringstream msg;
		msg << "Status: Using GLEW " << glewGetString( GLEW_VERSION );
		game.WriteDebuggingData(msg.str());
	}
//#endif

	if (!game.gfxcard.GetCapability(CARDINFOTYPE::FRAGMENTSHADERS))
		shaders_enabled = false;
	
	if (glGetError() != GL_NO_ERROR)
	{
		//throw EXCEPTION(__FILE__, __LINE__, "OpenGL error during graphics init");
		cerr << "OpenGL error during graphics init" << endl;
	}
	
	if (shaders_enabled)
	{
		game.WriteDebuggingData("Fragment shaders enabled");
		shaders.Init();
		bool success = true;
		success = success && shaders.Load("simple");
		//success = success && shaders.Load("simplerect");
		success = success && shaders.Load("full");
		success = success && shaders.Load("full-noshadow");
		//success = success && shaders.Load("full-nospecular");
		//success = success && shaders.Load("full-nometallic");
		success = success && shaders.Load("blurpass");
		//success = success && shaders.Load("bloompass");
		//success = success && shaders.Load("hblur_rect");
		//success = success && shaders.Load("vblur_rect");
		success = success && shaders.Load("depthgen");
		//success = success && shaders.Load("shadowpass");
		shaders_enabled = success;
		if (!success)
			shaders.DeInit();
		
		if (glGetError() != GL_NO_ERROR)
		{
			//throw EXCEPTION(__FILE__, __LINE__, "OpenGL error after shader loading is complete");
			cerr << "OpenGL error after shader loading is complete" << endl;
		}
	}
	
	if (shaders_enabled)
	{
		/*TEXTUREINFO ti;
		ti.SetName(game.settings.GetFullDataPath("textures/effects/noise.png"));
		ti.SetMipMap(false);
		noise.Load(ti);*/
		
		game.settings.Get("display.motionblur", blur_enabled);
		
		blur_enabled = blur_enabled && game.gfxcard.GetCapability(CARDINFOTYPE::TEXTURERECTANGLE) && 
				game.gfxcard.GetCapability(CARDINFOTYPE::FRAMEBUFFER_OBJECTS);
		bloom_enabled = false && game.gfxcard.GetCapability(CARDINFOTYPE::TEXTURERECTANGLE) && 
				game.gfxcard.GetCapability(CARDINFOTYPE::FRAMEBUFFER_OBJECTS); //TODO: get value from settings class
		game.settings.Get("display.shadows", shadow_enabled);
		shadow_enabled = shadow_enabled && game.gfxcard.GetCapability(CARDINFOTYPE::SHADOW) 
				&& game.gfxcard.GetCapability(CARDINFOTYPE::DEPTHTEXTURE) 
				&& game.gfxcard.GetCapability(CARDINFOTYPE::FRAMEBUFFER_OBJECTS) 
				&& (game.gfxcard.GetCapability(CARDINFOTYPE::TEXTUREUNITS) >= 6);
		shadow_enabled = shadow_enabled;
		shadow_resolution = 2048; //TODO: get value from settings class
		
		if (blur_enabled)
			fbtexture.Init((int)GetW(),(int)GetH(),true,false);
		
		if (shadow_enabled)
		{
			for (int n = 0; n < NUM_CASCADED_SHADOW_MAP; n++)
				fbdepthtexture[n].Init(shadow_resolution, shadow_resolution, false, true);
			//fbshadowtexture.Init((int)(GetW()),(int)(GetH()),true,false);
		}
		
		int intsetting;
		/*game.settings.Get("display.lighting", intsetting);
		switch (intsetting)
		{
			case 1:
				lightmap_status = LIGHTMAP_STATIC;
				break;
			
			default:
				lightmap_status = LIGHTMAP_DISABLED;
		}*/
		
		game.settings.Get("display.reflections", intsetting);
		switch (intsetting)
		{
			case 1:
				reflection_status = REFLECTION_STATIC;
				break;
			
			default:
				reflection_status = REFLECTION_DISABLED;
		}
		if (game.gfxcard.GetCapability(CARDINFOTYPE::TEXTUREUNITS) <= 2)
			reflection_status = REFLECTION_DISABLED;
		
		/*if (lightmap_status == LIGHTMAP_STATIC)
		{
			TEXTUREINFO t;
			t.SetName("data/textures/weather/cubelighting.png");
			t.SetCube(true, true);
			t.SetMipMap(false);
			static_lightmap.Load(t);
		}
		if (game.gfxcard.GetCapability(CARDINFOTYPE::TEXTUREUNITS) <= 5)
			lightmap_status = LIGHTMAP_DISABLED;*/
		
		if (reflection_status == REFLECTION_STATIC)
		{
			TEXTUREINFO t;
			t.SetName("textures/weather/cubereflection-nosun.png");
			t.SetCube(true, true);
			t.SetMipMap(false);
			static_reflection.Load(t);
		}
		
		shaders.SetActiveShader("simple");
	}
	else
	{
		blur_enabled = false;
		bloom_enabled = false;
	}
	
	light_position.Rotate(3.141593*0.5*0.7,1,0,0);
	light_position.PreRotate(3.141593*0.4,0,1,0);
}

bool SceneDrawSort_Nearest (SCENEDRAW & first, SCENEDRAW & second)
{
	if (first.IsCollapsed() && second.IsCollapsed())
	{
		VERTEX firstpos(first.GetMatrix4()->TransformVertexOut(first.GetDraw()->GetObjectCenter()));
		VERTEX secondpos(second.GetMatrix4()->TransformVertexOut(second.GetDraw()->GetObjectCenter()));
		firstpos = firstpos + game.cam.position;
		secondpos = secondpos + game.cam.position;
		float firstdist2 = firstpos.x*firstpos.x + firstpos.y*firstpos.y + firstpos.z*firstpos.z;
		float seconddist2 = secondpos.x*secondpos.x + secondpos.y*secondpos.y + secondpos.z*secondpos.z;
		bool firstcloser = (firstdist2 > seconddist2);
		//cout << firstdist2 << "," << seconddist2 << endl;
		return firstcloser;
	}
	
	return false;
}

void GRAPHICS::DrawScene()
{
	map < DRAWABLE_FILTER *, list <SCENEDRAW> > drawlist_map;
	list <DRAWABLE_FILTER *> filter_list;
	DRAWABLE_FILTER no2d_noblend;
	no2d_noblend.SetFilter_is2d(true, false);
	no2d_noblend.SetFilter_partial_transparency(true, false);
	no2d_noblend.SetFilter_skybox(true,false);
	DRAWABLE_FILTER no2d_blend;
	no2d_blend.SetFilter_is2d(true, false);
	no2d_blend.SetFilter_partial_transparency(true, true);
	no2d_blend.SetFilter_skybox(true,false);
	DRAWABLE_FILTER only2d;
	only2d.SetFilter_is2d(true, true);
	only2d.SetFilter_skybox(true,false);
	
	DRAWABLE_FILTER no2d_noblend_blur;
	DRAWABLE_FILTER no2d_blend_blur;
	DRAWABLE_FILTER no2d_noblend_noblur;
	DRAWABLE_FILTER no2d_blend_noblur;
	
	DRAWABLE_FILTER skyboxes;
	skyboxes.SetFilter_skybox(true,true);
	
	if (blur_enabled)
	{
		//for the first pass, select only the things that we want blurred
		no2d_noblend_blur = no2d_noblend;
		no2d_blend_blur = no2d_blend;
		no2d_noblend_noblur = no2d_noblend;
		no2d_blend_noblur = no2d_blend;
		no2d_noblend_blur.SetFilter_blur(true, true);
		no2d_noblend_blur.SetFilter_skybox(true,false);
		no2d_blend_blur.SetFilter_blur(true, true);
		no2d_blend_blur.SetFilter_skybox(true,false);
		no2d_noblend_noblur.SetFilter_blur(true, false);
		no2d_noblend_noblur.SetFilter_skybox(true,false);
		no2d_blend_noblur.SetFilter_blur(true, false);
		no2d_blend_noblur.SetFilter_skybox(true,false);
		
		filter_list.push_back(&no2d_noblend);
		filter_list.push_back(&no2d_blend);
		filter_list.push_back(&only2d);
		filter_list.push_back(&no2d_noblend_blur);
		filter_list.push_back(&no2d_blend_blur);
		filter_list.push_back(&no2d_noblend_noblur);
		filter_list.push_back(&no2d_blend_noblur);
	}
	else
	{
		filter_list.push_back(&no2d_noblend);
		filter_list.push_back(&no2d_blend);
		filter_list.push_back(&only2d);
	}
	filter_list.push_back(&skyboxes);
	
	//scene.GetDrawList(filter_list, drawlist_map);
	scene.GetCollapsedDrawList(filter_list, drawlist_map);
	
	/*if (blur_enabled)
	{
		drawlist_map[&no2d_blend].sort(SceneDrawSort_Nearest);
		drawlist_map[&no2d_blend_noblur].sort(SceneDrawSort_Nearest);
		drawlist_map[&no2d_blend_blur].sort(SceneDrawSort_Nearest);
	}
	else
		drawlist_map[&no2d_blend].sort(SceneDrawSort_Nearest);*/
	
	/*cout << "normal objects: " << drawlist_map[&nofontsor2d].size() << endl;
	cout << "2d element objects: " << drawlist_map[&only2d].size() << endl;
	cout << "font objects: " << drawlist_map[&onlyfonts].size() << endl;
	cout << "cursor objects: " << drawlist_map[&onlycursor].size() << endl;*/
	
	if (shaders_enabled)
	{
		shaders.SetActiveShader("full-noshadow");
	}
	
	float lod_far;
	game.settings.Get( "display.view_distance", lod_far );
	
	if (shadow_enabled)
	{
		shaders.SetActiveShader("full");
		
		//do initial pass from the light's perspective
	
		//set up viewport
		//glViewport( 0, 0, shadowres, shadowres );
		
		MATRIX4 clipmat[NUM_CASCADED_SHADOW_MAP];
		
		//do multiple shadow passes
		for (int n = 0; n < NUM_CASCADED_SHADOW_MAP; n++)
		{
			fbdepthtexture[n].Begin();
			
			glClear (GL_DEPTH_BUFFER_BIT);
			
			//set up the modelview and projection matrices to be from the light's perspective
			glMatrixMode( GL_PROJECTION );
			glPushMatrix();
			glLoadIdentity();
			
			//float shadow_radius = (1 << n)*5.0; //5,10,20
			//float shadow_radius = (1<<n)*5.0+(n)*10.0; //5,20,40
			//float shadow_radius = (1 << n)*10.0-5.0; //5,15,35
			float closeshadow = 5.0;
			if (game.world.GetCameraMode() == CMInCar)
				closeshadow = 2.0;
			float shadow_radius = (1<<n)*closeshadow+(n)*20.0; //5,30
			VERTEX shadowbox;
			shadowbox.Set(1,1,1);
			shadowbox.Scale(shadow_radius*sqrt(2));
			VERTEX shadowoffset;
			shadowoffset.Set(0,0,1);
			shadowoffset.Scale(shadow_radius);
			shadowoffset = game.cam.dir.ReturnConjugate().RotateVec(shadowoffset);
			
			//glOrtho(-50.0,50.0,-50.0,50.0,-50.0,50.0);
			//glOrtho(-10.0,10.0,-10.0,10.0,0.0,10.0);
			glOrtho(-shadowbox.x,shadowbox.x,-shadowbox.y,shadowbox.y,-shadowbox.z-20.0,shadowbox.z);
			glMatrixMode( GL_MODELVIEW );
			glPushMatrix();
			GLdouble temp_matrix[16];
			light_position.GetMat(temp_matrix);
			glLoadMatrixd(temp_matrix);
			/*{
				glLoadIdentity();
				glRotatef(180.0*0.5*0.7,1,0,0);
				glRotatef(180.0*0.4,0,1,0);
			}*/
			glTranslatef(game.cam.GetPosition().x,game.cam.GetPosition().y,game.cam.GetPosition().z);
			glTranslatef(shadowoffset.x,shadowoffset.y,shadowoffset.z);
			game.cam.ExtractFrustum();
			
			shaders.SetActiveShader("depthgen");
			//shaders.UploadActiveShaderParameter1f("depthoffset",n*0.01);
			//shaders.UploadActiveShaderParameter1f("depthoffset",0.0);
			
			//render objects into depth buffer
			if (blur_enabled)
			{
				DrawList(drawlist_map[&no2d_noblend_blur], -1, shadowbox.z+20);
				DrawList(drawlist_map[&no2d_blend_blur], -1, shadowbox.z+20);			
				DrawList(drawlist_map[&no2d_noblend_noblur], -1, shadowbox.z+20);
				DrawList(drawlist_map[&no2d_blend_noblur], -1, shadowbox.z+20);
			}
			else
			{
				DrawList(drawlist_map[&no2d_noblend], -1, shadowbox.z+20);
				DrawList(drawlist_map[&no2d_blend], -1, shadowbox.z+20);
			}
			
			//store light matrices
			//float mv[16];ExtractMatricesTo(mv, light_matrix.data);
			float mv[16], mp[16];
			glGetFloatv( GL_PROJECTION_MATRIX, mp );
			glGetFloatv( GL_MODELVIEW_MATRIX, mv );
			
			/*MATRIX4 m4p,m4v;
			m4p.Set(mp);
			m4v.Set(mv);
			clipmat[n].Translate(VERTEX(0.5,0.5,0.5));
			clipmat[n].Scale(0.5);
			clipmat[n].Multiply(m4p);
			clipmat[n].Multiply(m4v);*/
			//glActiveTextureARB(GL_TEXTURE2_ARB);
			glMatrixMode( GL_TEXTURE );
			glPushMatrix();
			glLoadIdentity();
			glTranslatef (0.5, 0.5, 0.5);
			glScalef (0.5, 0.5, 0.5);
			glMultMatrixf(mp);
			glMultMatrixf(mv);
			//glTranslatef(-game.cam.GetPosition().x,-game.cam.GetPosition().y,-game.cam.GetPosition().z);
			//glMultMatrixd(temp_matrix);
			glGetFloatv(GL_TEXTURE_MATRIX, clipmat[n].data);
			glPopMatrix();
			//glActiveTextureARB(GL_TEXTURE0_ARB);
			
			//restore matrices
			glMatrixMode( GL_PROJECTION );
			glPopMatrix();
			glMatrixMode( GL_MODELVIEW );
			glPopMatrix();
			
			//save depth buffer
			fbdepthtexture[n].End();
		}
		
		shaders.SetActiveShader("full");
		for (int n = 0; n < NUM_CASCADED_SHADOW_MAP; n++)
		{
			glActiveTextureARB(GL_TEXTURE2_ARB+n);
			glMatrixMode(GL_TEXTURE);
			//fbdepthtexture[n].Activate();
			//glstate.Set(GL_TEXTURE_2D, true);
			//glstate.Set(GL_TEXTURE_RECTANGLE_ARB, false);
			/*stringstream matname;
			matname << "light_matrix_" << n;
			shaders.UploadActiveMat16(matname.str(), clipmat[n].data);*/
			glLoadMatrixf(clipmat[n].data);
			glMatrixMode(GL_MODELVIEW);
		}
		
		glActiveTextureARB(GL_TEXTURE4_ARB);
		fbdepthtexture[0].Activate();
		glstate.Set(GL_TEXTURE_2D, true);
		glstate.Set(GL_TEXTURE_RECTANGLE_ARB, false);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE_ARB, GL_COMPARE_R_TO_TEXTURE_ARB);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC_ARB, GL_LEQUAL);
		
		glActiveTextureARB(GL_TEXTURE5_ARB);
		fbdepthtexture[1].Activate();
		glstate.Set(GL_TEXTURE_2D, true);
		glstate.Set(GL_TEXTURE_RECTANGLE_ARB, false);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE_ARB, GL_COMPARE_R_TO_TEXTURE_ARB);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC_ARB, GL_LEQUAL);
		
		glActiveTextureARB(GL_TEXTURE0_ARB);
		
		game.cam.ExtractFrustum();
		
		/*fbshadowtexture.Begin();
		
		game.cam.ExtractFrustum();
		
		//clear buffers
		glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		
		//send light matrix
		//shaders.SetActiveShader("full");
		//shaders.UploadActiveMat16("light_modelviewproj", light_matrix.data);
		
		//select shadow textures
		//glActiveTextureARB(GL_TEXTURE2_ARB);
		shaders.SetActiveShader("shadowpass");
		for (int n = 0; n < NUM_CASCADED_SHADOW_MAP; n++)
		{
			glActiveTextureARB(GL_TEXTURE0_ARB+n);
			fbdepthtexture[n].Activate();
			glstate.Set(GL_TEXTURE_2D, true);
			glstate.Set(GL_TEXTURE_RECTANGLE_ARB, false);
			stringstream matname;
			matname << "light_matrix_" << n;
			shaders.UploadActiveMat16(matname.str(), clipmat[n].data);
		}
		
		//render shadows from camera perspective
		if (blur_enabled)
		{
			DrawList(drawlist_map[&no2d_noblend_blur], -2);
			DrawList(drawlist_map[&no2d_blend_blur], -2);			
			DrawList(drawlist_map[&no2d_noblend_noblur], -2);
			DrawList(drawlist_map[&no2d_blend_noblur], -2);
		}
		else
		{
			DrawList(drawlist_map[&no2d_noblend], -2);
			DrawList(drawlist_map[&no2d_blend], -2);
		}
		
		//capture shading in fbtexture
		fbshadowtexture.End();
		
		glstate.Set(GL_TEXTURE_2D, false);
		glstate.Set(GL_TEXTURE_RECTANGLE_ARB, true);*/
		
		/*//do a blur to get some soft shadows
		float shadowtexw(512.0);
		float shadowtexh(512.0);
		fbshadowtexture.Begin();
		//float bluramount = 0.0015;
		float bluramount = 0.002;
		shaders.SetActiveShader("hblur_rect");
		shaders.UploadActiveShaderParameter1f("blurfactorx",bluramount*shadowtexw);
		RenderQuad(1.0,0.5,0.5);
		fbshadowtexture.End();
		//horiz blurred texture is now active in TU0
		fbshadowtexture.Begin();
		shaders.SetActiveShader("vblur_rect");
		shaders.UploadActiveShaderParameter1f("blurfactory",bluramount*shadowtexh*(3.0/4.0));
		RenderQuad(1.0,0.5,0.5);
		fbshadowtexture.End();*/
		
		//glViewport( 0, 0, ( GLint )width, ( GLint )height );
		
		/*glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		
		glstate.Set(GL_TEXTURE_2D, true);
		glstate.Set(GL_TEXTURE_RECTANGLE_ARB, false);
		
		glActiveTextureARB(GL_TEXTURE2_ARB);
		glstate.Set(GL_TEXTURE_2D, false);
		glstate.Set(GL_TEXTURE_RECTANGLE_ARB, true);
		fbshadowtexture.Activate();
		
		glActiveTextureARB(GL_TEXTURE0_ARB);
		
		glstate.Set(GL_TEXTURE_2D, false);
		glstate.Set(GL_TEXTURE_RECTANGLE_ARB, true);
		fbshadowtexture.Activate();*/
	}
	
	//load global textures
	if (reflection_status == REFLECTION_STATIC)
	{
		glActiveTextureARB(GL_TEXTURE2_ARB);
		glstate.Set(GL_TEXTURE_CUBE_MAP_ARB, true);
		static_reflection.Activate();
		glActiveTextureARB(GL_TEXTURE0_ARB);
	}
	/*if (lightmap_status == LIGHTMAP_STATIC)
	{
		glActiveTextureARB(GL_TEXTURE3_ARB);
		glstate.Set(GL_TEXTURE_CUBE_MAP_ARB, true);
		static_lightmap.Activate();
		glActiveTextureARB(GL_TEXTURE0_ARB);
	}*/
	
	//draw most of the objects (first real pass)
	if (blur_enabled)
	{
		fbtexture.Begin();
		glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		
		//draw skyboxes
		DrawList(drawlist_map[&skyboxes], 0, lod_far);
		
		DrawList(drawlist_map[&no2d_noblend_blur], 0, lod_far);
		DrawList(drawlist_map[&no2d_blend_blur], 0, lod_far);
	}
	else
	{
		//draw skyboxes
		DrawList(drawlist_map[&skyboxes], 0, lod_far);
		
		DrawList(drawlist_map[&no2d_noblend], 0, lod_far);
		DrawList(drawlist_map[&no2d_blend], 0, lod_far);
	}
	
	if (blur_enabled)
	{
		fbtexture.End();
		glActiveTextureARB(GL_TEXTURE0_ARB);
		fbtexture.Activate();
		glstate.Set(GL_TEXTURE_2D, false);
		glstate.Set(GL_TEXTURE_RECTANGLE_ARB, true);
		//glActiveTextureARB(GL_TEXTURE1_ARB);
		//noise.Activate(true);
		glActiveTextureARB(GL_TEXTURE0_ARB);
		
		BeginScene();
		glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );
		glClearDepth( 1.0f );
		glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		
		/*//draw an extra quad far away that represents static geometry
		//a better way TODO this is by having the OBJECTS class always contain a skybox-like sphere
		glPushMatrix();
		glDepthMask(0);
		glLoadIdentity(); //draw in camera space
		shaders.SetActiveShader("blurpass");
		shaders.UploadActiveMat16("prev_modelviewproj", prev_modelviewproj);
		glDisable(GL_TEXTURE_2D);
		glBegin(GL_QUADS);
		glVertex3f( -900.0f,  -900.0f,  -100.0f);
		glVertex3f( 900.0f,  -900.0f,  -100.0f);
		glVertex3f( 900.0f,  900.0f,  -100.0f);
		glVertex3f( -900.0f,  900.0f,  -100.0f);
		glEnd();
		glEnable(GL_TEXTURE_2D);
		glDepthMask(1);
		glPopMatrix();*/
		
		//draw second pass (blurred) objects
		DrawList(drawlist_map[&skyboxes], 1, lod_far); //draw skyboxes
		
		/*//decrease the view distance when drawing second pass objects to improve performance
		glMatrixMode( GL_PROJECTION );
		glPushMatrix();
		glLoadIdentity();
		gluPerspective( fov, width/(float)height, 0.1f, 50.0f );
		glMatrixMode( GL_MODELVIEW );*/
		
		DrawList(drawlist_map[&no2d_noblend_blur], 1, lod_far);
		DrawList(drawlist_map[&no2d_blend_blur], 1, lod_far);
		
		//restore the view distance
		/*glMatrixMode( GL_PROJECTION );
		glPopMatrix();
		glMatrixMode( GL_MODELVIEW );*/
		
		//draw non-blurred objects
		DrawList(drawlist_map[&no2d_noblend_noblur], 0, lod_far);
		DrawList(drawlist_map[&no2d_blend_noblur], 0, lod_far);
	}
	
	/*if (bloom_enabled)
	{
		fbtexture.End();
		glstate.Set(GL_TEXTURE_2D, false);
		glstate.Set(GL_TEXTURE_RECTANGLE_ARB, true);
		glActiveTextureARB(GL_TEXTURE1_ARB);
		glstate.Set(GL_TEXTURE_2D, false);
		glstate.Set(GL_TEXTURE_RECTANGLE_ARB, true);
		glActiveTextureARB(GL_TEXTURE0_ARB);
		
		//shaders.SetActiveShader("bloompass");
		
		shaders.SetActiveShader("hblur_rect");
		RenderQuad();
		fbblurtexture.End();
		//horiz blurred texture is now active in TU0
		
		shaders.SetActiveShader("vblur_rect");
		RenderQuad();
		fbblurtexture.End();
		
		glActiveTextureARB(GL_TEXTURE1_ARB);
		fbblurtexture.Activate();
		glActiveTextureARB(GL_TEXTURE0_ARB);
		fbtexture.Activate();
		
		//original texture is in TU0, blurred texture is in TU1
		
		shaders.SetActiveShader("bloompass");
		
		RenderQuad();
		
		glActiveTextureARB(GL_TEXTURE1_ARB);
		glstate.Set(GL_TEXTURE_RECTANGLE_ARB, false);
		glActiveTextureARB(GL_TEXTURE0_ARB);
		glstate.Set(GL_TEXTURE_RECTANGLE_ARB, false);
	}*/

	/*shaders.SetActiveShader("simple");
	if (shadow_enabled)
	{
		glstate.Set(GL_TEXTURE_2D, true);
		fbdepthtexture[0].Activate();
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE_ARB, GL_NONE);
		RenderQuad(0.25);
	}*/
	
	/*shaders.SetActiveShader("simplerect");
	if (shadow_enabled)
	{
		glstate.Set(GL_TEXTURE_2D, false);
		glstate.Set(GL_TEXTURE_RECTANGLE_ARB, true);
		fbshadowtexture.Activate();
		RenderQuad(0.25, width*0.5,height*0.5);
		glstate.Set(GL_TEXTURE_RECTANGLE_ARB, false);
		glstate.Set(GL_TEXTURE_2D, true);
	}*/
	
	drawlist_map[&only2d].sort();
	DrawList(drawlist_map[&only2d], 0, lod_far);
}

void GRAPHICS::DrawList(list <SCENEDRAW> & drawlist, int passno, float lod_far)
{
	bool verbose = false;
	if (verbose)
	{
		cout << drawlist.size() << " scene elements" << endl;
		ITERATE(SCENEDRAW,drawlist,i)
		{
			i->DebugPrint();
		}
	}
	
	int cullcount = 0;
	
	currentshader = UNKNOWN;
	
	last_transform_valid = false;
	
	ITERATE(SCENEDRAW,drawlist,i)
	{
		if (i->IsDraw() && !(passno < 0 && i->GetDraw()->skybox))
		{
			//if (FrustumCull(*i, lod_far) && passno != -1)
			if (FrustumCull(*i, lod_far) && passno != 100) //pass 100 is our code for a spinning car!
				cullcount++;
			else if (passno >= 0 || !i->GetDraw()->issmoke)
			{
				bool blurring (false);
				if (shaders_enabled)
				{
					SelectAppropriateShader(*i, passno);
					blurring = (currentshader == SHADER_BLURPASS);
				}
				
				SelectFlags(*i, blurring, passno);
				
				SelectTexturing(*i, blurring, passno);
	
				bool need_pop = SelectTransformStart(*i, blurring);
				
				switch (i->GetDraw()->type)
				{
					case DRAWABLE::LIST:
						glCallList(i->GetDraw()->drawlist);
						//cout << i->GetDraw()->drawlist << endl;
						break;
					case DRAWABLE::VARRAY:
						glEnableClientState(GL_VERTEX_ARRAY);
					
						const float * verts;
						int vertcount;
						i->GetDraw()->varray.GetVertices(verts, vertcount);
						glVertexPointer(3, GL_FLOAT, 0, verts);
					
						const float * norms;
						int normcount;
						i->GetDraw()->varray.GetNormals(norms, normcount);
						glNormalPointer(GL_FLOAT, 0, norms);
						if (normcount > 0)
							glEnableClientState(GL_NORMAL_ARRAY);
					
						//const float * tc[i->GetDraw()->varray.GetTexCoordSets()];
						//int tccount[i->GetDraw()->varray.GetTexCoordSets()];
						const float * tc[1];
						int tccount[1];
						if (i->GetDraw()->varray.GetTexCoordSets() > 0)
						{
							i->GetDraw()->varray.GetTexCoords(0, tc[0], tccount[0]);
							glEnableClientState(GL_TEXTURE_COORD_ARRAY);
							glTexCoordPointer(2, GL_FLOAT, 0, tc[0]);
						}
					
						const int * faces;
						int facecount;
						i->GetDraw()->varray.GetFaces(faces, facecount);
					
						glDrawElements(GL_TRIANGLES, facecount, GL_UNSIGNED_INT, faces);
					
						glDisableClientState(GL_VERTEX_ARRAY);
						glDisableClientState(GL_NORMAL_ARRAY);
						glDisableClientState(GL_TEXTURE_COORD_ARRAY);
						
						//if (i->GetDraw()->font) cout << "Draw vertexarray: " << facecount/3 << " faces" << endl;
						break;
					case DRAWABLE::NONE:
						break;
					default:
						throw EXCEPTION(__FILE__, __LINE__, "Unimplemented drawnode type encountered in scenegraph");
						break;
				}
				
				SelectTransformEnd(*i, blurring, need_pop);
			}
		}
		else if (i->IsTransform())
		{
			//cout << "is transform" << endl;
			float mat[16];
			glPushMatrix();
			VERTEX translation;
			translation = i->GetTransform()->GetTranslation();
			glTranslatef(translation.x,translation.y,translation.z);
			i->GetTransform()->GetRotation().GetMat(mat);
			glMultMatrixf(mat);
			last_transform_valid = false;
		}
		else if (i->IsEmpty())
		{
			//cout << "is transform pop" << endl;
			glPopMatrix();
		}
	}
	
	if (last_transform_valid)
		glPopMatrix();
	
	//cout << "Culled " << cullcount << "/" << drawlist.size() << endl;
	
	//if (curlit)	EndTextureBasedLighting_Forest();
}

void GRAPHICS_GLSTATEMANAGER::Set(int stateid, bool newval)
{
	map <int,bool>::iterator i = state.find(stateid);
	if (i == state.end())
	{
		/*if (stateid == GL_TEXTURE_2D)
				cout << "GL_TEXTURE_2D: " << newval << endl;*/
		state[stateid] = newval;
		if (newval)
			glEnable(stateid);
		else
			glDisable(stateid);
	}
	else
	{
		if (i->second != newval)
		{
			i->second = newval;

			if (newval)
				glEnable(stateid);
			else
				glDisable(stateid);
		}
	}
}

// function to reset our viewport after a window resize
int GRAPHICS::resizeWindow(int w, int h, float f)
{
	width = w;
	height = h;
	fov = f;
	
	// Height / width ration
	GLfloat ratio;
 
	// Protect against a divide by zero
	if ( height == 0 )
		height = 1;

	ratio = ( GLfloat )width / ( GLfloat )height;

	// Setup our viewport.
	glViewport( 0, 0, ( GLint )width, ( GLint )height );

	// change to the projection matrix and set our viewing volume.
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity( );

	// Set our perspective
	gluPerspective( fov, ratio, 0.1f, 10000.0f );

	// Make sure we're changing the model view and not the projection
	glMatrixMode( GL_MODELVIEW );

	// Reset The View
	glLoadIdentity( );

	return( true );
}

// This function returns true if the extension is there.
bool GRAPHICS::isExtensionSupported(string extstring)
{
	char * temp = (char *) glGetString(GL_EXTENSIONS);
	if (temp == 0)
	{
		cout << "Error getting extensions.  Continuing anyway, errors may follow!" << endl;
		return true;
	}

	string s = temp;
	string::size_type temppos = s.find(extstring);

	bool hasext = (temppos <= s.length());

	if (!hasext)
	{
		cout << "Extension not supported: " << extstring << endl;
	}

	return hasext;
}

void GRAPHICS::BeginScene()
{
	glClearColor( 0.3750f, 0.4830f, 0.718750f, 0.0f );
	glClearDepth( 1.0f );
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	glShadeModel( GL_SMOOTH );
	glDepthFunc( GL_LEQUAL );
	glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
	glstate.Set(GL_TEXTURE_2D, true);
	glstate.Set(GL_DEPTH_TEST, true);
	glstate.Set(GL_LIGHTING, false);

	GLdouble temp_matrix[16];
	
	//this is done to be able to get real normals in the shaders
	if (shaders_enabled)
	{
		if( game.gfxcard.GetCapability( CARDINFOTYPE::MULTITEXTURE ) )
		{
			glActiveTextureARB(GL_TEXTURE1_ARB);
			glMatrixMode(GL_TEXTURE);
			game.cam.dir.ReturnConjugate().GetMat(temp_matrix);
			glLoadIdentity();
			glTranslatef(-game.cam.GetPosition().x,-game.cam.GetPosition().y,-game.cam.GetPosition().z);
			glMultMatrixd(temp_matrix);
			
			/*glActiveTextureARB(GL_TEXTURE2_ARB);
			glMatrixMode(GL_TEXTURE);
			game.cam.dir.GetMat(temp_matrix);
			glLoadIdentity();
			//glTranslatef(-game.cam.GetPosition().x,-game.cam.GetPosition().y,-game.cam.GetPosition().z);
			glMultMatrixd(temp_matrix);*/
		
			glActiveTextureARB(GL_TEXTURE0_ARB);
		}
	}
	
	const bool shadow_debug(false);
	
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	if (shadow_debug)
	{
		//glOrtho(-50.0,50.0,-50.0,50.0,0.0,100.0);
		glOrtho(-10.0,10.0,-10.0,10.0,0.0,10.0);
	}
	else
		gluPerspective( fov, width/(float)height, 0.1f, 10000.0f );
	
	glMatrixMode( GL_MODELVIEW );
	if (shadow_debug)
	{
		QUATERNION lightdir;
		lightdir.Rotate(3.141593*0.5,1,0,0);
		lightdir.GetMat(temp_matrix);
	}
	else
		game.cam.PutTransformInto(temp_matrix);
	glLoadMatrixd(temp_matrix);
	if (shadow_debug)
	{
		//glTranslatef(game.cam.GetPosition().x,game.cam.GetPosition().y-75,game.cam.GetPosition().z);
		glTranslatef(game.cam.GetPosition().x,game.cam.GetPosition().y-5,game.cam.GetPosition().z);
	}
	else
		glTranslatef(game.cam.GetPosition().x,game.cam.GetPosition().y,game.cam.GetPosition().z);
	game.cam.ExtractFrustum();
}

void GRAPHICS::EndScene()
{
	SDL_GL_SwapBuffers( );
	
	if (blur_enabled)
		ExtractMatrices();
	
	shaders.EndScene();
}

void FBTEXTURE_GL::Init(int sizex, int sizey, bool rect, bool newdepth)
{
	if (inited)
	{
		DeInit();
	}
	
	depth = newdepth;
	
	inited = true;
	
	sizew = sizex;
	sizeh = sizey;
	
	rectangle = rect;
	
	//initialize framebuffer object (FBO)
	glGenFramebuffersEXT(1, &framebuffer_object);
	
	//set texture info
	int texture_target(GL_TEXTURE_2D);
	if (rectangle)
		texture_target = GL_TEXTURE_RECTANGLE_ARB;
	int texture_format1(GL_RGB);
	int texture_format2(GL_RGB);
	int texture_format3(GL_UNSIGNED_BYTE);
	if (depth)
	{
		texture_format1 = GL_DEPTH_COMPONENT16;
		texture_format2 = GL_DEPTH_COMPONENT;
		texture_format3 = GL_UNSIGNED_INT;
	}
	
	//initialize the texture
	glGenTextures(1, &fbtexture);
	glBindTexture(texture_target, fbtexture);
	glTexImage2D(texture_target, 0, texture_format1, sizex, sizey, 0, texture_format2, texture_format3, NULL);
	//glTexParameteri(texture_target, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
	//glTexParameteri(texture_target, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);
	glTexParameteri(texture_target, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(texture_target, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(texture_target, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(texture_target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	if (depth)
	{
		//glTexParameteri(texture_target, GL_DEPTH_TEXTURE_MODE, GL_LUMINANCE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE_ARB, GL_COMPARE_R_TO_TEXTURE_ARB);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC_ARB, GL_LEQUAL);
	}
	glBindTexture(texture_target, 0); // don't leave the texture bound
	
	//bind the framebuffer
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, framebuffer_object);
	
	//initialize renderbuffer object that's used for our depth buffer when rendering to a texture
	if (!depth)
	{
		glGenRenderbuffersEXT(1, &renderbuffer_depth);
		glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, renderbuffer_depth);
		glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT, sizew, sizeh);
		
		//attach the render buffer to the FBO
		glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, renderbuffer_depth);
	}
	else
	{
		glDrawBuffer(GL_NONE); // no color buffer dest
		glReadBuffer(GL_NONE); // no color buffer src
	}
	
	//attach the texture to the FBO
	int texture_attachment(GL_COLOR_ATTACHMENT0_EXT);
	if (depth)
		texture_attachment = GL_DEPTH_ATTACHMENT_EXT;
	
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, texture_attachment, texture_target, fbtexture, 0);
	
	if (glGetError() != GL_NO_ERROR)
	{
		//throw EXCEPTION(__FILE__, __LINE__, "OpenGL error on FBO create");
		cerr << "OpenGL error on FBO create" << endl;
	}
	
	GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
	if (status == GL_FRAMEBUFFER_UNSUPPORTED_EXT)
	{
		cout << sizex << "," << sizey << "," << rect << "," << depth << endl;
		throw EXCEPTION(__FILE__, __LINE__, "Framebuffer object format not supported");
	}
	else if (status != GL_FRAMEBUFFER_COMPLETE_EXT)
	{
		cout << sizex << "," << sizey << "," << rect << "," << depth << endl;
		throw EXCEPTION(__FILE__, __LINE__, "Framebuffer object is not complete");
	}
	else
	{
		game.WriteDebuggingData("Framebuffer object complete");
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	}
}

void FBTEXTURE_GL::DeInit()
{
	if (inited)
	{
		glDeleteTextures(1, &fbtexture);
		glDeleteFramebuffersEXT(1, &framebuffer_object);
		if (!depth)
			glDeleteRenderbuffersEXT(1, &renderbuffer_depth);
	}
	
	loaded = false;
	inited = false;
}

void FBTEXTURE_GL::Begin(float viewscale)
{
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, framebuffer_object);
	glPushAttrib(GL_VIEWPORT_BIT);
	glViewport(0,0,(int)(sizew*viewscale),(int)(sizeh*viewscale));
}

void FBTEXTURE_GL::End()
{
	glPopAttrib();
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	
	if (glGetError() != GL_NO_ERROR)
	{
		//throw EXCEPTION(__FILE__, __LINE__, "OpenGL error on FBO end");
		cerr << "OpenGL error on FBO end" << endl;
	}
	
	/*if (rectangle)
	{
		if (!game.gfxcard.GetCapability(CARDINFOTYPE::TEXTURERECTANGLE))
			throw EXCEPTION(__FILE__, __LINE__, "Card does not support rectangular textures");
		
		glActiveTextureARB(GL_TEXTURE0_ARB);
		glBindTexture(GL_TEXTURE_RECTANGLE_ARB, fbtexture);
		//glCopyTexSubImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, 0, 0, 0, 0, sizew, sizeh);
		glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
		
		glTexParameteri(GL_TEXTURE_RECTANGLE_ARB,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
		glTexParameteri(GL_TEXTURE_RECTANGLE_ARB,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
	}
	else
	{
		glActiveTextureARB(GL_TEXTURE0_ARB);
		glBindTexture(GL_TEXTURE_2D, fbtexture);
		//glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0, sizew, sizeh);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		
		//glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
		//glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
		
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE_ARB, GL_COMPARE_R_TO_TEXTURE_ARB);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC_ARB, GL_LEQUAL);
		glTexParameteri(GL_TEXTURE_2D, GL_DEPTH_TEXTURE_MODE_ARB, GL_LUMINANCE);
		
		float border[4] = {0.0,0.0,0.0,0.0};
		glTexParameterfv(GL_TEXTURE_2D, GL_TEXTURE_BORDER_COLOR, border);
	}*/
}

void FBTEXTURE_GL::Activate()
{
	int texture_target(GL_TEXTURE_2D);
	if (rectangle)
		texture_target = GL_TEXTURE_RECTANGLE_ARB;

	glBindTexture(texture_target, fbtexture);
	//glEnable(texture_target);
}

//pull out the modelview and modelview * projection (clip-space) matrices
// these are used by the motion blur shader, and the camera frustum can also
// be extracted from these
void GRAPHICS::ExtractMatrices()
{
	ExtractMatricesTo(prev_modelview, prev_modelviewproj);
}

void GRAPHICS::ExtractMatricesTo(float * modelviewout, float * modelviewprojout)
{
	float   proj[16];
	float * modl = modelviewout;
	float * clip = modelviewprojout;
	
	glGetFloatv( GL_PROJECTION_MATRIX, proj );
	
	glGetFloatv( GL_MODELVIEW_MATRIX, modl );
	
	clip[ 0] = modl[ 0] * proj[ 0] + modl[ 1] * proj[ 4] + modl[ 2] * proj[ 8] + modl[ 3] * proj[12];
	clip[ 1] = modl[ 0] * proj[ 1] + modl[ 1] * proj[ 5] + modl[ 2] * proj[ 9] + modl[ 3] * proj[13];
	clip[ 2] = modl[ 0] * proj[ 2] + modl[ 1] * proj[ 6] + modl[ 2] * proj[10] + modl[ 3] * proj[14];
	clip[ 3] = modl[ 0] * proj[ 3] + modl[ 1] * proj[ 7] + modl[ 2] * proj[11] + modl[ 3] * proj[15];
	
	clip[ 4] = modl[ 4] * proj[ 0] + modl[ 5] * proj[ 4] + modl[ 6] * proj[ 8] + modl[ 7] * proj[12];
	clip[ 5] = modl[ 4] * proj[ 1] + modl[ 5] * proj[ 5] + modl[ 6] * proj[ 9] + modl[ 7] * proj[13];
	clip[ 6] = modl[ 4] * proj[ 2] + modl[ 5] * proj[ 6] + modl[ 6] * proj[10] + modl[ 7] * proj[14];
	clip[ 7] = modl[ 4] * proj[ 3] + modl[ 5] * proj[ 7] + modl[ 6] * proj[11] + modl[ 7] * proj[15];
	
	clip[ 8] = modl[ 8] * proj[ 0] + modl[ 9] * proj[ 4] + modl[10] * proj[ 8] + modl[11] * proj[12];
	clip[ 9] = modl[ 8] * proj[ 1] + modl[ 9] * proj[ 5] + modl[10] * proj[ 9] + modl[11] * proj[13];
	clip[10] = modl[ 8] * proj[ 2] + modl[ 9] * proj[ 6] + modl[10] * proj[10] + modl[11] * proj[14];
	clip[11] = modl[ 8] * proj[ 3] + modl[ 9] * proj[ 7] + modl[10] * proj[11] + modl[11] * proj[15];
	
	clip[12] = modl[12] * proj[ 0] + modl[13] * proj[ 4] + modl[14] * proj[ 8] + modl[15] * proj[12];
	clip[13] = modl[12] * proj[ 1] + modl[13] * proj[ 5] + modl[14] * proj[ 9] + modl[15] * proj[13];
	clip[14] = modl[12] * proj[ 2] + modl[13] * proj[ 6] + modl[14] * proj[10] + modl[15] * proj[14];
	clip[15] = modl[12] * proj[ 3] + modl[13] * proj[ 7] + modl[14] * proj[11] + modl[15] * proj[15];
}

void GRAPHICS::RecordFPSThisFrame(const float newfps)
{
	double coeff = 0.1;
	fps = fps*(1.0-coeff) + newfps*coeff;
	truefps = newfps;
}

//useful for shaders
void GRAPHICS::RenderQuad(float scale, float maxw, float maxh)
{
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho( 0, 1, 0, 1, -1, 1 );
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	
	glColor4f(1,1,1,1);
	glstate.Set(GL_BLEND, false);
	glstate.Set(GL_DEPTH_TEST, false);

	glScalef(scale,scale,0);
	//glTranslatef(1.0,1.0,0);
	
	glBegin(GL_QUADS);
	glTexCoord2f(0.0f, 0.0f); glVertex3f( 0.0f,  0.0f,  0.0f);
	glTexCoord2f(maxw, 0.0f); glVertex3f( 1.0f,  0.0f,  0.0f);
	glTexCoord2f(maxw, maxh); glVertex3f( 1.0f,  1.0f,  0.0f);
	glTexCoord2f(0.0f, maxh); glVertex3f( 0.0f,  1.0f,  0.0f);
	glEnd();
	
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	glstate.Set(GL_DEPTH_TEST, true);
}

bool GRAPHICS::FrustumCull(SCENEDRAW & tocull, float lod_far) const
{
	const SCENEDRAW * i (& tocull);
	const DRAWABLE * d (i->GetDraw());
	if (d->radius != 0.0 && d->parent != NULL && !d->skybox)
	{
		//do frustum culling
		VERTEX objpos(d->object_center);
		if (i->IsCollapsed())
			objpos = i->GetMatrix4()->TransformVertexOut(objpos);
		else
			objpos = i->GetDraw()->parent->TransformIntoWorldSpace(objpos);
		float dx=objpos.x+game.cam.position.x; float dy=objpos.y+game.cam.position.y; float dz=objpos.z+game.cam.position.z;
		float rc=dx*dx+dy*dy+dz*dz;
		float temp_lod_far = lod_far + d->radius;
		if (rc > temp_lod_far*temp_lod_far)
			return true;
		else
		{
			float bound, rd;
			bound = d->radius;
			for (int i=0; i<6; i++) 
			{
				rd=game.cam.frustum[i][0]*objpos.x+
				   game.cam.frustum[i][1]*objpos.y+
				   game.cam.frustum[i][2]*objpos.z+
				   game.cam.frustum[i][3];
				if (rd <= -bound)
				{
					return true;
				}
			}
		}
	}
	
	return false;
}

void GRAPHICS::SelectAppropriateShader(SCENEDRAW & forme, int passno)
{
	SCENEDRAW * i(&forme);
	if (passno == -1)
	{
		if (currentshader != SHADER_DEPTH_GEN)
		{
			shaders.SetActiveShader("depthgen");
			currentshader = SHADER_DEPTH_GEN;
		}
	}
	else if (passno == -2)
	{
		if (currentshader != SHADER_SHADOW_PASS)
		{
			shaders.SetActiveShader("shadowpass");
			currentshader = SHADER_SHADOW_PASS;
		}
	}
	else if (passno != 1 || (i->GetDraw()->is2d || !i->GetDraw()->blur || i->GetDraw()->decal))
	{
		if (i->GetDraw()->is2d || i->GetDraw()->decal || !i->GetDraw()->lit || i->GetDraw()->issmoke)
		{
			if (currentshader != SHADER_SIMPLE)
			{
				shaders.SetActiveShader("simple");
				currentshader = SHADER_SIMPLE;
			}
		}
		else
		{
			if (currentshader != SHADER_FULL)
			{
				if (shadow_enabled)
					shaders.SetActiveShader("full");
				else
					shaders.SetActiveShader("full-noshadow");
				currentshader = SHADER_FULL;
			}
			
			/*if (i->GetDraw()->misc_map1.PtrValid())
			{
				if (i->GetDraw()->metallic)
				{
					if (currentshader != SHADER_FULL)
					{
						shaders.SetActiveShader("full");
						currentshader = SHADER_FULL;
					}
				}
				else
				{
					if (currentshader != SHADER_FULL_NOMETALLIC)
					{
						shaders.SetActiveShader("full-nometallic");
						currentshader = SHADER_FULL_NOMETALLIC;
					}
				}
			}
			else
			{
				if (currentshader != SHADER_FULL_NOSPECULAR)
				{
					shaders.SetActiveShader("full-nospecular");
					currentshader = SHADER_FULL_NOSPECULAR;
				}
			}*/
		}
	}
	else
	{
		if (currentshader != SHADER_BLURPASS)
		{
			shaders.SetActiveShader("blurpass");
			currentshader = SHADER_BLURPASS;
			
			shaders.UploadActiveMat16("prev_modelviewproj", current_matrix.data);
		}
		
		/*if (currentshader != SHADER_SIMPLE)
		{
			shaders.SetActiveShader("simple");
			currentshader = SHADER_SIMPLE;
		}*/
	}
}

void GRAPHICS::SelectFlags(SCENEDRAW & forme, bool blurring, int passno)
{
	SCENEDRAW * i(&forme);
	if (passno == -1)
	{
		//glPolygonOffset(1.0,1.0);
		//glPolygonOffset(0.2,0.2);
		//glstate.Set(GL_POLYGON_OFFSET_FILL, true);
		glstate.Set(GL_POLYGON_OFFSET_FILL, false);
	}
	else if ((i->GetDraw()->decal || i->GetDraw()->partial_transparency) && !blurring)
	{
		glPolygonOffset(-1.0,-1.0);
		glstate.Set(GL_POLYGON_OFFSET_FILL, true);
	}
	else
	{
		glstate.Set(GL_POLYGON_OFFSET_FILL, false);
	}
	
	if (passno == -1)
	{
		glstate.Set(GL_CULL_FACE, true);
		glCullFace(GL_BACK);
		glstate.Set(GL_CULL_FACE, false);
	}
	else
	{
		glstate.Set(GL_CULL_FACE, i->GetDraw()->cull);
		if (i->GetDraw()->cull)
		{
			if (i->GetDraw()->cull_front)
				glCullFace(GL_FRONT);
			else
				glCullFace(GL_BACK);
		}
		//glstate.Set(GL_CULL_FACE, false);
	}
	
	if ((i->GetDraw()->decal || i->GetDraw()->is2d || i->GetDraw()->partial_transparency) && passno >= 0)
	{
		glstate.Set(GL_ALPHA_TEST, false);
		glstate.Set(GL_BLEND, true);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		//glBlendFunc(GL_SRC_ALPHA,GL_DST_ALPHA);
		
		if (i->GetDraw()->issmoke)
		{
			glBlendFunc(GL_SRC_ALPHA,GL_ONE);
			//glBlendFunc(GL_ONE,GL_ONE);
		}
	}
	else if (passno < 0)
	{
		glstate.Set(GL_BLEND, false);
		glAlphaFunc(GL_GREATER, 0.50f);
	        glstate.Set(GL_ALPHA_TEST, true);
	}
	else
	{
		/*glstate.Set(GL_ALPHA_TEST, false);
		glstate.Set(GL_BLEND, true);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);*/
		//glEnable(GL_SAMPLE_ALPHA_TO_COVERAGE_ARB);
		glstate.Set(GL_BLEND, false);
		glAlphaFunc(GL_GREATER, 0.25f);
        	glstate.Set(GL_ALPHA_TEST, true);
	}
	
	if (i->GetDraw()->issmoke || i->GetDraw()->skybox)
	{
		glDepthMask(0);
	}
	else
	{
		glDepthMask(1);
	}
	
	if (i->GetDraw()->issmoke || i->GetDraw()->is2d)
		glstate.Set(GL_DEPTH_TEST,false);
	else
		glstate.Set(GL_DEPTH_TEST,true);
	
	glColor4f(i->GetDraw()->r,i->GetDraw()->g,i->GetDraw()->b,i->GetDraw()->a);
}

void GRAPHICS::SelectTexturing(SCENEDRAW & forme, bool blurring, int passno)
{
	SCENEDRAW * i(&forme);
	if (i->GetDraw()->diffuse_map.PtrValid() && !blurring)
	{
		if (i->GetDraw()->diffuse_map.GetPtr()->Loaded())
		{
			if (current_diffusemap != &(i->GetDraw()->diffuse_map.GetPtr()->GetTexture()))
			{
				/*if (passno == -2)
					glActiveTextureARB(GL_TEXTURE3_ARB);
				else
					glActiveTextureARB(GL_TEXTURE0_ARB);*/
				
				glstate.Set(GL_TEXTURE_2D, true);
				
				i->GetDraw()->diffuse_map.GetPtr()->Activate();
				//cout << "Activating " << i->GetDraw()->diffuse_map.GetPtr()->GetTexture().GetTextureInfo().GetName() << endl;
				current_diffusemap = &(i->GetDraw()->diffuse_map.GetPtr()->GetTexture());
				
				if (shaders_enabled && passno >= 0)
				{
					if (i->GetDraw()->misc_map1.PtrValid())
					{
						glActiveTextureARB(GL_TEXTURE1_ARB);
						i->GetDraw()->misc_map1.GetPtr()->Activate();
						glActiveTextureARB(GL_TEXTURE0_ARB);
					}
					
					//if (game.gfxcard.GetCapability(CARDINFOTYPE::TEXTUREUNITS) >= 6 && shaders_enabled)
					//{
						if (i->GetDraw()->additive_map1.PtrValid())
						//if (i->GetDraw()->additive_map1.PtrValid())
						{
							glActiveTextureARB(GL_TEXTURE6_ARB);
							if (i->GetDraw()->self_illumination)
							{
								//shaders.UploadActiveShaderParameter1f("self_illumination",1.0f);
								i->GetDraw()->additive_map1.GetPtr()->Activate();
							}
							else
							{
								glBindTexture(GL_TEXTURE_2D,0);
								//shaders.UploadActiveShaderParameter1f("self_illumination",0.0f);
							}
							glActiveTextureARB(GL_TEXTURE0_ARB);
						}
						else
						{
							glActiveTextureARB(GL_TEXTURE6_ARB);
							glBindTexture(GL_TEXTURE_2D,0);
							glActiveTextureARB(GL_TEXTURE0_ARB);
						}
					//}
					
					if (!i->GetDraw()->misc_map1.PtrValid())
					{
						glActiveTextureARB(GL_TEXTURE1_ARB);
						glBindTexture(GL_TEXTURE_2D, 0);
						glActiveTextureARB(GL_TEXTURE0_ARB);
					}
				}
			}
		}
		else
		{
			glstate.Set(GL_TEXTURE_2D, false);
			current_diffusemap = NULL;
		}
	}
}

///returns true if the matrix was pushed
bool GRAPHICS::SelectTransformStart(SCENEDRAW & forme, bool blurring)
{
	bool need_a_pop = true;
	
	SCENEDRAW * i(&forme);
	if (i->GetDraw()->is2d)
	{
		if (last_transform_valid)
			glPopMatrix();
		last_transform_valid = false;
		
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho( 0, 1, 0, 1, -1, 1 );
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		if (i->IsCollapsed())
		{
			glMultMatrixf(i->GetMatrix4()->data);
		}
	}
	else
	{
		glstate.Set(GL_DEPTH_TEST, true);

		if (i->IsCollapsed())
		{
			if (i->GetDraw()->skybox)
			{
				if (last_transform_valid)
					glPopMatrix();
				last_transform_valid = false;
				
				glPushMatrix();
				double temp_matrix[16];
				game.cam.PutTransformInto(temp_matrix);
				glLoadMatrixd(temp_matrix);
				if (game.track.GetVerticalTrackingSkyboxes())
				{
					VERTEX objpos(i->GetDraw()->object_center);
					objpos = i->GetDraw()->parent->TransformIntoWorldSpace(objpos); 
					//glTranslatef(-objpos.x,-objpos.y,-objpos.z);
					//glTranslatef(0,game.cam.position.y,0);
					glTranslatef(0.0,-objpos.y,0.0);
				}
				glMultMatrixf(i->GetMatrix4()->data);
			}
			else
			{
				if (!last_transform.Equals(*i->GetMatrix4()) || !last_transform_valid)
				//if (1)
				{
					if (last_transform_valid)
						glPopMatrix();
					
					glPushMatrix();
					glMultMatrixf(i->GetMatrix4()->data);
					last_transform = *i->GetMatrix4();
					last_transform_valid = true;
					
					need_a_pop = false;
				}
				else need_a_pop = false;
			}
			
			//if (blurring) //TODO:  move this so it's done even for culled objects
			if (blurring)
			{
				if (!current_matrix.Equals(i->GetDraw()->last_modelviewproj))
				{
					current_matrix.Set(i->GetDraw()->last_modelviewproj);
					shaders.UploadActiveMat16("prev_modelviewproj", current_matrix.data);
						//shaders.UploadActiveMat16("prev_modelviewproj", (const_cast <DRAWABLE *> (i->GetDraw()))->last_modelviewproj.data);
				}
					
				float this_modelview[16];
				ExtractMatricesTo(this_modelview,(const_cast <DRAWABLE *> (i->GetDraw()))->last_modelviewproj.data);
			}
		}
	}
	
	return need_a_pop;
}

void GRAPHICS::SelectTransformEnd(SCENEDRAW & forme, bool blurring, bool need_pop)
{
	SCENEDRAW * i(&forme);
	if (i->GetDraw()->is2d && need_pop)
	{
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
	}
	else if (i->IsCollapsed())
	{
		if (need_pop)
		{
			glPopMatrix();
			//cout << "popping" << endl;
		}
		//else cout << "not popping" << endl;
	}
}
