#define GL_GLEXT_PROTOTYPES

#include "model.h"
#include "common.h"
#include "game.h"
extern GAME game;

JOEMODEL::JOEMODEL() :
	pObject (NULL),
	loadedfile (false),
	radius (0.0),
	radiusxz (0.0)
{
	
}

int JOEMODEL::BinaryRead(void * buffer, unsigned int size, unsigned int count, FILE * f, JOEPACK * pack)
{
	if (pack == NULL)
	{
		return fread(buffer, size, count, f);
	}
	else
	{
		return pack->Pack_fread(buffer, size, count);
	}
}

bool JOEMODEL::Load(string strFileName, JOEPACK * pack, bool adjustnormals)
{
	modelpath = strFileName;
	
	if (loadedfile)
	{
		loadedfile = false;
		
		if (pObject != NULL)
		{	
			for (int i = 0; i < pObject->info.num_frames; i++)
			{
				delete [] pObject->frames[i].faces;
				delete [] pObject->frames[i].verts;
				delete [] pObject->frames[i].normals;
				delete [] pObject->frames[i].texcoords;
			}
			
			delete [] pObject->frames;
			
			delete pObject;
		}
		
		pObject = NULL;
		
		glDeleteLists(static_list, 1);
		static_list = 0;
	}

	string filename = strFileName;

	FILE * m_FilePointer = NULL;
	
    //open file
	bool fileinpack = false;
	if (pack == NULL)
	{
    	m_FilePointer = fopen(filename.c_str(), "rb");
	}
	else
		fileinpack = pack->Pack_fopen(filename);

    // Make sure we have a valid file pointer (we found the file)
    if ((pack == NULL && !m_FilePointer) || (pack != NULL && !fileinpack))
    {
        // Display an error message and don't load anything if no file was found
		loadedfile = false;
		throw EXCEPTION(__FILE__, __LINE__, "Unable to find file "+filename);
    }

	bool val = LoadFromHandle(m_FilePointer, pack, adjustnormals);
	
	// Clean up after everything
    CleanUp(m_FilePointer, pack);
	
	return val;
}

bool JOEMODEL::LoadFromHandle(FILE * m_FilePointer, JOEPACK * pack, bool adjustnormals)
{	
	//create a new object
	pObject = new JOEObject;

    // Read the header data and store it in our variable
    BinaryRead(&pObject->info, sizeof(JOEHeader), 1, m_FilePointer, pack);
	
	pObject->info.magic = ENDIAN_SWAP_32(pObject->info.magic);
	pObject->info.version = ENDIAN_SWAP_32(pObject->info.version);
	pObject->info.num_faces = ENDIAN_SWAP_32(pObject->info.num_faces);
	pObject->info.num_frames = ENDIAN_SWAP_32(pObject->info.num_frames);
	
    // Make sure the version is what we expect or else it's a bad egg
    if(pObject->info.version != JOE_VERSION)
    {
        // Display an error message for bad file format, then stop loading
        //sprintf(strMessage, "Invalid file format (Version is %d not %d): %s!", pObject->info.version, JOE_VERSION, modelpath.c_str());
		loadedfile = false;
        //cerr << strMessage << endl;
		delete [] pObject;
		pObject = NULL;
		stringstream err;
		err << "Invalid file format (Version is " << pObject->info.version << " not " << JOE_VERSION << ": " << modelpath;
		throw EXCEPTION(__FILE__, __LINE__, err.str());
		return false;
    }
	
	if (pObject->info.num_faces > JOE_MAX_FACES)
	{
		stringstream err;
		err << modelpath << " has " << pObject->info.num_faces << " faces (max " << JOE_MAX_FACES << ".";
		throw EXCEPTION(__FILE__, __LINE__, err.str());
	}
	
	if (pObject->info.num_frames > JOE_MAX_FRAMES)
	{
		stringstream err;
		err << modelpath << " has " << pObject->info.num_frames << " frames (max " << JOE_MAX_FRAMES << ".";
		throw EXCEPTION(__FILE__, __LINE__, err.str());
	}
	
	loadedfile = true;

	//cout << pObject->info.version << "," << pObject->info.num_faces << endl;
	
    // Read in the model and animation data
	ReadData(m_FilePointer, pack, adjustnormals);

	//optimize frame zero into a static display list
	static_list = glGenLists(1);
	glNewList (static_list, GL_COMPILE);
	Draw(0,0,0);
	glEndList ();
	
    // Return a success
    return true;
}

void JOEMODEL::ReadData(FILE *m_FilePointer, JOEPACK * pack, bool adjustnormals)
{
	int num_frames = pObject->info.num_frames;
	int num_faces = pObject->info.num_faces;
	
	pObject->frames = new JOEFrame [num_frames];
	
	int i;
	for (i = 0; i < num_frames; i++)
	//for (i = 0; i < 1; i++)
	{
		pObject->frames[i].faces = new JOEFace [num_faces];
		
		BinaryRead(pObject->frames[i].faces, sizeof(JOEFace), num_faces, m_FilePointer, pack);
		CorrectEndian(pObject->frames[i].faces, num_faces);
		
		BinaryRead(&pObject->frames[i].num_verts, sizeof(int), 1, m_FilePointer, pack);
		pObject->frames[i].num_verts = ENDIAN_SWAP_32(pObject->frames[i].num_verts);
		BinaryRead(&pObject->frames[i].num_texcoords, sizeof(int), 1, m_FilePointer, pack);
		pObject->frames[i].num_texcoords = ENDIAN_SWAP_32(pObject->frames[i].num_texcoords);
		BinaryRead(&pObject->frames[i].num_normals, sizeof(int), 1, m_FilePointer, pack);
		pObject->frames[i].num_normals = ENDIAN_SWAP_32(pObject->frames[i].num_normals);
		//cout << pObject->frames[i].num_verts << "," << pObject->frames[i].num_texcoords << "," << pObject->frames[i].num_normals << endl;
		
		pObject->frames[i].verts = new JOEVertex [pObject->frames[i].num_verts];
		pObject->frames[i].normals = new JOEVertex [pObject->frames[i].num_normals];
		pObject->frames[i].texcoords = new JOETexCoord [pObject->frames[i].num_texcoords];
		
		BinaryRead(pObject->frames[i].verts, sizeof(JOEVertex), pObject->frames[i].num_verts, m_FilePointer, pack);
		CorrectEndian(pObject->frames[i].verts, pObject->frames[i].num_verts);
		BinaryRead(pObject->frames[i].normals, sizeof(JOEVertex), pObject->frames[i].num_normals, m_FilePointer, pack);
		CorrectEndian(pObject->frames[i].normals, pObject->frames[i].num_normals);
		BinaryRead(pObject->frames[i].texcoords, sizeof(JOETexCoord), pObject->frames[i].num_texcoords, m_FilePointer, pack);
		CorrectEndian(pObject->frames[i].texcoords, pObject->frames[i].num_texcoords);
		
		//cout << pObject->frames[i].texcoords[0].u << "," << pObject->frames[i].texcoords[0].v << endl;
	}
	
	float maxv[3];
	float minv[3];
	bool havevals[6];
	int n = 0;
	for (n = 0; n < 6; n++)
		havevals[n] = false;
	
	//go do scaling
	for (i = 0; i < num_frames; i++)
	{
		int v;
		
		for (v = 0; v < pObject->frames[i].num_verts; v++)
		{
			VERTEX temp;
			
			temp.Set(pObject->frames[i].verts[v].vertex);
			temp.Scale(MODEL_SCALE);
			
			//cache for bbox stuff
			for (n = 0; n < 3; n++)
			{
				if (temp.v3()[n] > maxv[n] || !havevals[n])
				{
					maxv[n] = temp.v3()[n];
					havevals[n] = true;
				}
				if (temp.v3()[n] < minv[n] || !havevals[n+3])
				{
					minv[n] = temp.v3()[n];
					havevals[n+3] = true;
				}
			}
			
			float r = temp.len();
			float rxz = sqrt(temp.x*temp.x+temp.z*temp.z);
			if (r > radius)
				radius = r;
			if (rxz > radiusxz)
				radiusxz = rxz;
			/*pObject->frames[i].verts[v].vertex[0] = temp.y;
			pObject->frames[i].verts[v].vertex[1] = temp.z;
			pObject->frames[i].verts[v].vertex[2] = -temp.x;*/
			
			pObject->frames[i].verts[v].vertex[0] = temp.x;
			pObject->frames[i].verts[v].vertex[1] = temp.y;
			pObject->frames[i].verts[v].vertex[2] = temp.z;
		}
		
		if (adjustnormals)
		{
			for (int ni = 0; ni < pObject->frames[i].num_normals; ni++)
			{
				bool isnan = false;
				for (int v = 0; v < 3; v++)
					if (!(pObject->frames[i].normals[ni].vertex[v] < 0 || pObject->frames[i].normals[ni].vertex[v] >= 0))
						isnan = true;
				if (isnan)
				{
					for (int v = 0; v < 3; v++)
						pObject->frames[i].normals[ni].vertex[v] = 0;
				}
				VERTEX temp;
				temp.Set(pObject->frames[i].normals[ni].vertex);
				pObject->frames[i].normals[ni].vertex[0] = temp.x;
				pObject->frames[i].normals[ni].vertex[1] = -temp.z;
				pObject->frames[i].normals[ni].vertex[2] = temp.y;
			}
		}
	}
	
	//make sure vertex ordering is consistent with normals
	/*for (i = 0; i < pObject->info.num_faces; i++)
	{
		short vi[3];
		VERTEX tri[3];
		VERTEX norms[3];
		for (unsigned int v = 0; v < 3; v++)
		{
			vi[v] = GetFace(i)[v];
			tri[v].Set(GetVert(vi[v]));
			norms[v].Set(GetNorm(GetNormIdx(i)[v]));
		}
		VERTEX norm;
		for (unsigned int v = 0; v < 3; v++)
			norm = norm + norms[v];
		norm = norm.normalize();
		VERTEX tnorm = (tri[2] - tri[0]).cross(tri[1] - tri[0]);
		if (norm.dot(tnorm) > 0)
		{
			short tvi = pObject->frames[0].faces[i].vertexIndex[1];
			pObject->frames[0].faces[i].vertexIndex[1] = pObject->frames[0].faces[i].vertexIndex[2];
			pObject->frames[0].faces[i].vertexIndex[2] = tvi;
			
			tvi = pObject->frames[0].faces[i].normalIndex[1];
			pObject->frames[0].faces[i].normalIndex[1] = pObject->frames[0].faces[i].normalIndex[2];
			pObject->frames[0].faces[i].normalIndex[2] = tvi;
			
			tvi = pObject->frames[0].faces[i].textureIndex[1];
			pObject->frames[0].faces[i].textureIndex[1] = pObject->frames[0].faces[i].textureIndex[2];
			pObject->frames[0].faces[i].textureIndex[2] = tvi;
		}
	}*/
	
	VERTEX minvals, maxvals;
	minvals.Set(minv);
	maxvals.Set(maxv);
	bbox.SetFromCorners(minvals, maxvals);
	
	VERTEX center;
	center = bbox.GetCenter();
	radius = (minvals - center).len();
	VERTEX mvnoy;
	mvnoy = minvals;
	mvnoy.y = 0;
	center.y = 0;
	radiusxz = (mvnoy - center).len();
}

void JOEMODEL::CleanUp(FILE *m_FilePointer, JOEPACK * pack)
{
	if (pack == NULL)
    	fclose(m_FilePointer);                      // Close the current file pointer
	else
		pack->Pack_fclose();
}

JOEMODEL::~JOEMODEL()
{
	int i;
	
	// Free the faces, normals, vertices, and texture coordinates.
	if (pObject != NULL)
	{	
		for (i = 0; i < pObject->info.num_frames; i++)
		{
			delete [] pObject->frames[i].faces;
			delete [] pObject->frames[i].verts;
			delete [] pObject->frames[i].normals;
			delete [] pObject->frames[i].texcoords;
		}
		
		delete [] pObject->frames;
		
		delete pObject;
	}
	
	glDeleteLists(static_list, 1);
}

void JOEMODEL::Draw(int currentFrame, int nextframe, float t)
{
	//utility_SelectTU(0);
	//glEnable(GL_TEXTURE_2D);
	
	int num_frames = pObject->info.num_frames;
	int num_faces = pObject->info.num_faces;
	JOEFrame * pFrame, * pNextFrame;
	
	if (currentFrame >= num_frames)
	{
		currentFrame = num_frames-1;
	}
	
	if (nextframe >= num_frames)
		nextframe = num_frames-1;
	
	//cout << currentFrame << "," << nextframe << endl;
	
	pFrame = &pObject->frames[currentFrame];
	pNextFrame = &pObject->frames[nextframe];
	
	// Start rendering triangles
    glBegin(GL_TRIANGLES);

        // Go through all of the faces (polygons) of the current frame and draw them
        for(int j = 0; j < num_faces; j++)
        {
		
			//calculate normals for flat shading
			/*VERTEX a, b, c;
			a.Set(pFrame->verts[pFrame->faces[j].vertexIndex[0]].vertex);
			b.Set(pFrame->verts[pFrame->faces[j].vertexIndex[1]].vertex);
			c.Set(pFrame->verts[pFrame->faces[j].vertexIndex[2]].vertex);
			
			float v0[3], v1[3], v2[3];
			v0[0] = a.x; v0[1] = a.y; v0[2] = a.z;
			v1[0] = b.x; v1[1] = b.y; v1[2] = b.z;
			v2[0] = c.x; v2[1] = c.y; v2[2] = c.z;
			
			float vb[3], va[3];
			int i;
			for (i = 0; i < 3; i++)
			{
				va[i] = v0[i]-v1[i];	
				vb[i] = v2[i]-v1[i];
			}
			float norm[3];
			norm[0] = va[1]*vb[2]-vb[1]*va[2];
			norm[1] = va[2]*vb[0]-vb[2]*va[0];
			norm[2] = va[0]*vb[1]-vb[0]*va[1];
			float normmag = sqrt(norm[0]*norm[0]+norm[1]*norm[1]+norm[2]*norm[2]);
			for (i = 0; i < 3; i++)
			{
				norm[i]=norm[i]/normmag;
			}
			
			//glNormal3fv(norm);*/
			
			
            // Go through each corner of the triangle and draw it.
            for(int whichVertex = 0; whichVertex < 3; whichVertex++)
            {
				
                // Get the index for each point of the face
                int vertIndex = pFrame->faces[j].vertexIndex[whichVertex];
				int nextvertIndex = pNextFrame->faces[j].vertexIndex[whichVertex];

                // Get the index for each texture coordinate for this face
                int texIndex  = pFrame->faces[j].textureIndex[whichVertex];				
				
				int normIndex = pFrame->faces[j].normalIndex[whichVertex];
				int nextnormIndex = pNextFrame->faces[j].normalIndex[whichVertex];
				                     
				// Pass in the texture coordinate for this vertex
				glTexCoord2f(pFrame->texcoords[ texIndex ].u, pFrame->texcoords[ texIndex ].v);
				/*utility_TexCoord2d2f(0, pFrame->texcoords[ texIndex ].u, pFrame->texcoords[ texIndex ].v);
				utility_TexCoord2d2f(1, pFrame->texcoords[ texIndex ].u, pFrame->texcoords[ texIndex ].v);
				utility_TexCoord2d2f(2, pFrame->texcoords[ texIndex ].u, pFrame->texcoords[ texIndex ].v);
				utility_TexCoord2d2f(3, pFrame->texcoords[ texIndex ].u, pFrame->texcoords[ texIndex ].v);*/
				
				//JGL_2MTEXCOORD(GL_TEXTURE0_ARB, pFrame->texcoords[ texIndex ].u, pFrame->texcoords[ texIndex ].v);
				//JGL_2MTEXCOORD(GL_TEXTURE1_ARB, pFrame->texcoords[ texIndex ].u, pFrame->texcoords[ texIndex ].v);
				//JGL_2MTEXCOORD(GL_TEXTURE2_ARB, pFrame->texcoords[ texIndex ].u, pFrame->texcoords[ texIndex ].v);
				//JGL_2MTEXCOORD(GL_TEXTURE3_ARB, pFrame->texcoords[ texIndex ].u, pFrame->texcoords[ texIndex ].v);

                // Now we get to the interpolation part! (*Bites his nails*)
                // Below, we first store the vertex we are working on for the current
                // frame and the frame we are interpolating too.  Next, we use the
                // linear interpolation equation to smoothly transition from one
                // key frame to the next.
                
				VERTEX vPoint1, vPoint2;
				vPoint1.Set(pFrame->normals[ normIndex ].vertex);
                vPoint2.Set(pNextFrame->normals[ nextnormIndex ].vertex);
				vPoint1 = vPoint1.normalize();
				vPoint2 = vPoint2.normalize();
				//glNormal3f(vPoint1.x + t * (vPoint2.x - vPoint1.x), // Find the interpolated X
                //           vPoint1.y + t * (vPoint2.y - vPoint1.y), // Find the interpolated Y
                //           vPoint1.z + t * (vPoint2.z - vPoint1.z));// Find the interpolated Z
			
				VERTEX n;
				n.x = vPoint1.x + t * (vPoint2.x - vPoint1.x);
				n.y = vPoint1.y + t * (vPoint2.y - vPoint1.y);
				n.z = vPoint1.z + t * (vPoint2.z - vPoint1.z);
				//n.z = vPoint1.y + t * (vPoint2.y - vPoint1.y);
				//n.y = vPoint1.z + t * (vPoint2.z - vPoint1.z);
				
				//n.Set(norm[0], norm[1], norm[2]);
				n = n.normalize();
				//n = rotation.RotateVec(n);
				
				//if (pass == 1)
				/*if (1)
					TexCoordFromNormal(n, lightdir);
				else
				{
					if (pass == 1)
						ColorFromNormal(n, lightdir);
					else
						TexCoordFromNormal(n, lightdir);
				}*/

				/*QUATERNION goofyfoot;
				goofyfoot.Rotate(-3.141593/2.0, 1,0,0);
				n = goofyfoot.ReturnConjugate().RotateVec(n);*/
				glNormal3fv(n.v3());
				//glNormal3f(n.y,-n.x,n.z);
				
				//if (pass == 2)
				
				
                // Store the current and next frame's vertex
                vPoint1.Set(pFrame->verts[ vertIndex ].vertex);
                vPoint2.Set(pNextFrame->verts[ nextvertIndex ].vertex);
				
				//CVector3 vNorm1 = pFrame->pNormals[ vertIndex ];
                //CVector3 vNorm2 = pNextFrame->pNormals[ vertIndex ];
				
				//glNormal3f(vNorm1.x + t * (vNorm2.x - vNorm1.x), // Find the interpolated X
                //           vNorm1.y + t * (vNorm2.y - vNorm1.y), // Find the interpolated Y
                //           vNorm1.z + t * (vNorm2.z - vNorm1.z));// Find the interpolated Z

				//cout << vNorm1.x << "," << vNorm1.y << "," << vNorm1.z << endl;

				//glNormal3f(pFirstFrame->pNormals[vertIndex].x,pFirstFrame->pNormals[vertIndex].y,pFirstFrame->pNormals[vertIndex].z);
				
                // By using the equation: p(t) = p0 + t(p1 - p0), with a time t
                // passed in, we create a new vertex that is closer to the next key frame.
				VERTEX finv;
				finv.x = vPoint1.x + t * (vPoint2.x - vPoint1.x);
				finv.y = vPoint1.y + t * (vPoint2.y - vPoint1.y);
				finv.z = vPoint1.z + t * (vPoint2.z - vPoint1.z);
				
				//tree shadows
				//utility.TexCoord2d2f(3, (finv.x-terrain.GetOffset().x)/terrain.GetSize().x, (finv.y-terrain.GetOffset().z)/terrain.GetSize().z);
				//utility.TexCoord2d2f(3, pFrame->texcoords[ texIndex ].u, pFrame->texcoords[ texIndex ].v);
                glVertex3f(finv.x, finv.y, finv.z); //required for compatibility with goofy vamos z=up
						   
				
            }
        }

    // Stop rendering the triangles
    glEnd();
}

/*void JOEMODEL::DrawStatic() //draw frame 0, optimized for speed
{
	int i;
	
	for (i = 0; i < MAX_TEXTURE_UNITS && i < utility_numTUs(); i++)
	{
		utility_SelectTU(i);
		glDisable(GL_TEXTURE_2D);
	}
	
	for (i = 0; i < MAX_TEXTURE_UNITS && i < utility_numTUs(); i++)
	{
		if (texturemode[i] != TEXTUREMODE_NOTEX)
		{	
			utility_SelectTU(i);
			//glBindTexture(GL_TEXTURE_2D, textureid[i]);
			textureid[i].Activate();
			if (tuenable[i])
			{
				glEnable(GL_TEXTURE_2D);
				
				//glColor4f(0,1,0,0.5);
				
				//default texture environment:  modulate w/ previous.				
				glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
				
				glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, GL_PREVIOUS_ARB);
				glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB,GL_SRC_COLOR);
				glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB, GL_TEXTURE);
				glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB,GL_SRC_COLOR);
				glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_MODULATE);
				//glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_REPLACE);
				
				glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_ALPHA_ARB, GL_PREVIOUS_ARB);
				glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_ALPHA_ARB,GL_SRC_ALPHA);
				glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_ALPHA_ARB, GL_TEXTURE);
				glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_ALPHA_ARB,GL_SRC_ALPHA);
				glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_ALPHA_ARB, GL_MODULATE);
				//glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_ALPHA_ARB, GL_REPLACE);
				
				glTexEnvi(GL_TEXTURE_ENV, GL_RGB_SCALE_ARB, 1);
				glTexEnvi(GL_TEXTURE_ENV, GL_ALPHA_SCALE, 1);
				
				if (texturemode[i] == TEXTUREMODE_ADD)
				{
					glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
					glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB, GL_PREVIOUS_ARB);
					glTexEnvi(GL_TEXTURE_ENV,GL_OPERAND1_RGB_ARB,GL_SRC_COLOR);
					glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, GL_TEXTURE);
					glTexEnvi(GL_TEXTURE_ENV,GL_OPERAND0_RGB_ARB,GL_SRC_COLOR);
					//if (i > 0)
						glTexEnvi(GL_TEXTURE_ENV,GL_COMBINE_RGB_ARB, GL_ADD);
					//else	glTexEnvi(GL_TEXTURE_ENV,GL_COMBINE_RGB_ARB, GL_MODULATE);
					
					glTexEnvi(GL_TEXTURE_ENV, GL_RGB_SCALE_ARB, 1);
				}
				
				if (texturemode[i] == TEXTUREMODE_REFLECTION)
				{
					glTexGenf(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP); 
					glTexGenf(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP); 
					glEnable(GL_TEXTURE_GEN_S);
					glEnable(GL_TEXTURE_GEN_T);
					glEnable(GL_TEXTURE_2D);
					float sub[4];
					//sub[0] = sub[1] = sub[2] = 0.8f;
					sub[0] = sub[1] = sub[2] = 0.7f;
					//sub[0] = sub[1] = sub[2] = 0.0f;
					sub[3] = 0.0f;
					glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, sub);
					glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
					glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB, GL_PREVIOUS_ARB);
					glTexEnvi(GL_TEXTURE_ENV,GL_OPERAND1_RGB_ARB,GL_SRC_COLOR);
					glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, GL_TEXTURE);
					glTexEnvi(GL_TEXTURE_ENV,GL_OPERAND0_RGB_ARB,GL_SRC_COLOR);
					//glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE2_RGB_ARB, GL_CONSTANT_ARB);
					//glTexEnvi(GL_TEXTURE_ENV,GL_OPERAND2_RGB_ARB,GL_SRC_COLOR);
					//glTexEnvi(GL_TEXTURE_ENV,GL_COMBINE_RGB_ARB, GL_INTERPOLATE);
					glTexEnvi(GL_TEXTURE_ENV, GL_RGB_SCALE_ARB, 1);
					
					//if (i > 0)
						glTexEnvi(GL_TEXTURE_ENV,GL_COMBINE_RGB_ARB, GL_ADD);
						
						
						//glTexEnvi(GL_TEXTURE_ENV,GL_COMBINE_RGB_ARB, GL_ADD_SIGNED_ARB);
						//glTexEnvi(GL_TEXTURE_ENV,GL_COMBINE_RGB_ARB, GL_REPLACE);
					//else	glTexEnvi(GL_TEXTURE_ENV,GL_COMBINE_RGB_ARB, GL_MODULATE);
					
				}
			}
			else
				glDisable(GL_TEXTURE_2D);
		}
		
		//cout << modelpath << ": " << i << ": " << texturemode[i] << endl;
	}
	
	//glBindTexture(GL_TEXTURE_2D, texid[texnum]);
	//utility.SelectTU(0);
	
	if (!loadedfile)
		return;
	
	//tex unit 1 is used by the reflections
	
	//glAlphaFunc(GL_GREATER, 0.9f);
	//glEnable(GL_ALPHA_TEST);
	//glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	//glBlendFunc(GL_SRC_ALPHA, GL_ZERO);
	//glEnable(GL_BLEND);
	
    glCallList(static_list);
	
	//glDisable(GL_ALPHA_TEST);
	
	for (i = 0; i < MAX_TEXTURE_UNITS && i < utility_numTUs(); i++)
	{
		if (texturemode[i] != TEXTUREMODE_NOTEX && i > 0)
		{
			utility_SelectTU(i);
			glDisable(GL_TEXTURE_2D);
			
			if (texturemode[i] == TEXTUREMODE_REFLECTION)
			{
				glDisable(GL_TEXTURE_GEN_S);
				glDisable(GL_TEXTURE_GEN_T);
				
			}
		}
	}
}*/

/*GLuint JOEMODEL::GetTex(int tnum)
{
	return texid[tnum];
}*/

/*void JOEMODEL::SetBaseTexture(int tid)
{
	//cout << tid << ",";
	
	if (tid < 0)
	{
		tid = -tid;
		tid = tid % numtex;
		tid = numtex - tid;
	}
	//active_base = tid % numtex;
	tid = tid % numtex;
	
	if (loadedfile && has_paint && numtex > 1)
	{
		glDeleteTextures(1, &texid);
		char fn[256];
		sprintf(fn, "%stextures/%s/body%02d.png", modelpath.c_str(), settings.GetTexSize().c_str(), tid);
		texid = utility.TexLoad(fn, true);
	}
	
	//cout << tid << endl;
}*/

void JOEMODEL::CorrectEndian(struct JOEFace * p, int num)
{
	int i;
	
	for (i = 0; i < num; i++)
	{
		int d;
		for (d = 0; d < 3; d++)
		{
			p[i].vertexIndex[d] = ENDIAN_SWAP_16(p[i].vertexIndex[d]);
			p[i].normalIndex[d] = ENDIAN_SWAP_16(p[i].normalIndex[d]);
			p[i].textureIndex[d] = ENDIAN_SWAP_16(p[i].textureIndex[d]);
		}
	}
}

void JOEMODEL::CorrectEndian(struct JOEVertex *p, int num)
{
	int i;
	
	for (i = 0; i < num; i++)
	{
		int d;
		for (d = 0; d < 3; d++)
		{
			p[i].vertex[d] = ENDIAN_SWAP_FLOAT(p[i].vertex[d]);
		}
	}
}

void JOEMODEL::CorrectEndian(struct JOETexCoord *p, int num)
{
	int i;
	
	for (i = 0; i < num; i++)
	{
		p[i].u = ENDIAN_SWAP_FLOAT(p[i].u);
		p[i].v = ENDIAN_SWAP_FLOAT(p[i].v);
	}
}
