#include "objects.h"
#include "3dmath.h"
#include "scenegraph.h"
#include "vamosworld.h" //needed for texture based lighting
#include "singleton.h"
#include "game.h"
extern GAME game;

OBJECTS::OBJECTS()
{
	error_log.open((game.settings.GetSettingsDir() + "/logs/objects.log").c_str());
//	UpdateSettings();
	
	//sphere_reflection_loaded = false;
	
	//object_list = NULL;
	model_list = NULL;
}

OBJECTS::~OBJECTS()
{
	error_log.close();
	
	//deallocate objects and models
	DeleteAll();
}

void OBJECTS::DeleteAll()
{
	object_list.clear();
	
	while (model_list != NULL)
		delmodel();
	
	if (parentnode.PtrValid())
	{
		game.world.GetVamosNode()->Delete(parentnode.GetPtr());
		parentnode.Invalidate();
	}
}

void OBJECTS::delmodel()
{
	if (model_list == NULL)
		return;
	
	//save old tree
	OBJECTMODEL * old = model_list;
	model_list = model_list->next;
	delete old;
}

void OBJECTS::UpdateSettings()
{
	game.settings.Get( "display.view_distance", lod_far );
	game.settings.Get( "display.width", display_x );
	game.settings.Get( "display.height", display_y );
	game.settings.Get( "display.FOV", field_of_view );
}

OBJECTNODE * OBJECTS::Add(VERTEX pos, float rotation, string modelname, string texname, bool mip, bool fullbright, bool skybox, bool drv, bool col, bool blend, JOEPACK * pack, float f1, float f2, float bl, float bm, float rr, float rd, SCENENODE * parent)
{
	{
		OBJECTNODE newnode;
		object_list.push_back(newnode);
	}
	
	object_list.back().pos.x = pos.x;
	//object_list->pos.y = pos.y;
	object_list.back().pos.z = pos.z;
	
	object_list.back().driveable = drv;
	object_list.back().cancollide = col;
	
	object_list.back().friction1 = f1;
	object_list.back().friction2 = f2;
	
	object_list.back().bumplength = bl;
	object_list.back().bumpmag = bm;
	
	object_list.back().rolling_resistance_factor = rr;
	object_list.back().rolling_drag = rd;
	
	//if (pos.y == -1337)
	{
//		object_list->pos.y = terrain.GetHeight(object_list->pos.x, object_list->pos.z) + pos.y;
	}
	
	//apply rotation (no longer used)
	//object_list->dir.Rotate(rotation, 0, 1, 0);
	
	OBJECTMODEL * mplist = model_list;
	
	bool found = false;
	int count = 0;
	while (mplist != NULL && !found)
	{
		//cout << mplist->name << "," << modelname << endl;
		if (mplist->name == modelname)
		{
			found = true;
			object_list.back().model = mplist;
		}
		mplist = mplist -> next;
		count++;
	}
	
	if (!found)
	{
		object_list.back().model = AddModel(modelname, texname, mip, fullbright, skybox, blend, pack);
		object_list.back().texture = texname;
	}
	
	if (!object_list.back().drawable.PtrValid())
		object_list.back().drawable.SetPtr(parent->AddDrawable());
	
	DRAWABLE & d = object_list.back().drawable.GetRef();
	
	d.SetToList(object_list.back().model->jmodel.GetDrawList());
	d.SetLit(!fullbright);
	d.SetPartialTransparency(blend);
	//d.SetDiffuseMap(object_list.back().model->jmodel.GetTextureHandle(0).GetTexture().GetTextureInfo());
	TEXTUREINFO texinfo;
	texinfo.SetName(path + "/" + texname);
	int anisotropy = 0;
	game.settings.Get( "display.anisotropic", anisotropy );
	if (game.gfxcard.GetCapability(CARDINFOTYPE::ANISOTROPY) && anisotropy > 0)
		texinfo.SetMipMap(true);
	else
		texinfo.SetMipMap(mip);
	d.SetDiffuseMap(texinfo);
	string miscpath = path + "/" + texname.substr(0,texname.length()-4) + "-misc1.png";
	if (game.utility.FileExists(miscpath))
	{
		//cout << "*********" << miscpath << endl;
		texinfo.SetName(miscpath);
		d.SetMiscMap1(texinfo);
	}
	//d.SetCull(game.track.GetCullFaces(), true);
	d.SetCull(game.track.GetCullFaces(), false);
	//cout << "**************" << game.track.GetCullFaces() << endl;
	d.SetRadius(object_list.back().model->jmodel.GetRadius());
	VERTEX objcenter = object_list.back().model->jmodel.GetBBOX().GetCenter();
	d.SetObjectCenter(objcenter);
	d.SetSkybox(skybox);
	
	//if (object_list.back().cancollide || object_list.back().driveable)
	if (object_list.back().cancollide)
	{
		object_list.back().LoadModelIntoCollisionData();
		object_list.back().RegisterWithCollisionSystem();
	}
	
	return &(object_list.back());
}

OBJECTNODE::OBJECTNODE() : vertices(NULL),faces(NULL),vcount(0),fcount(0)
{
	dir.LoadMultIdent();
}

OBJECTMODEL * OBJECTS::AddModel(string modelname, string texname, bool mip, bool fullbright, bool skybox, bool blend, JOEPACK * pack)
{
	OBJECTMODEL * oldfirst = model_list;
	model_list = new OBJECTMODEL;
	model_list->next = oldfirst;
	
	model_list->name = modelname;
	model_list->jmodel.Load(path + "/" + modelname, pack, true);
	
	model_list->fullbright = fullbright;
	model_list->blend = blend;
	model_list->skybox = skybox;
	
	//search for the texture in our texture db
	/*map <string, TEXTURE_HANDLE>::iterator tslot = texture_db.find(texname);
	if (tslot == texture_db.end())
	{
		TEXTURE_HANDLE newtexid;
		newtexid.Load(path + "/" + texname, mip);
		texture_db[texname] = newtexid;
		//cout << texname << ": " << newtexid << endl;
		model_list->jmodel.TextureID(&newtexid, 0);
	}
	else
	{
		model_list->jmodel.TextureID(&(tslot->second), 0);
	}*/
	
	//model_list->jmodel.ReflectionTextureID(sphere_reflection, 1);
	
	//cout << "Addmodel: " << modelname << endl;
	return model_list;
}

extern void LoadingScreen(string loadtext);

void OBJECTS::LoadObjectsFromFolder(string objectpath)
{
	DeleteAll();
	
	if (!parentnode.PtrValid())
	{
		//parentnode.SetPtr(game.graphics.GetScene().GetRoot().AddNode());
		parentnode.SetPtr(game.world.GetVamosNode()->AddNode());
		QUATERNION goofyfoot;
		//goofyfoot.Rotate(-3.141593/2.0, 1,0,0);
		//goofyfoot.Rotate(-3.141593/2.0, 0,1,0);
		goofyfoot.Rotate(-3.141593/2.0, 0,0,1);
		parentnode.GetRef().GetTransform().SetRotation(goofyfoot);
	}
	
	//sphere_reflection = textures.Load("weather/trackreflect.png", true);
	//sphere_reflection_loaded = true;
	
	JOEPACK pack;
	JOEPACK * packptr = NULL;
	
	if (pack.LoadPack(objectpath+"/objects.jpk"))
	{
		packptr = &pack;
		//cout << "Using object pack" << endl;
	}
	
	OBJECTNODE * added = NULL;
	
	ifstream o;
	
	o.open((objectpath+"/list.txt").c_str());
	
	path = objectpath;
	
	if (o)
	{
		string m;
		string t;
		string extra;
		bool mip;
		bool fb, sb;
		bool blend;
		VERTEX p;
		float r;
		bool c, d;
		float f1, f2;
		float bl, bm;
		float rr, rd;
		
		int count = 0;
		
		const int expectedparams = 14;
		int numparams = expectedparams;
		numparams = game.utility.iGetParam(o);
		
		if (numparams != expectedparams)
		{
			cerr << "Parameters per entry number incorrect, error in object list: " << objectpath+"/list.txt" << endl;
			cerr << "Expected " << expectedparams << " parameters, got " << numparams << endl;
			cerr << "Check the format.txt file in the VDrift-trackeditor/listedit/format.txt folder." << endl;
			return;
		}
		
		while (!o.eof())
		{
			if (count % 100 == 0)
			{
				//char buffer[256];
				//sprintf(buffer, "Loading...\nLoading scenery objects\n%i", count);
				//LoadingScreen(buffer);
				string msg = "Loading...\nLoading scenery objects";
				for (int sn = 0; sn < count/100; sn++)
					msg = msg + ".";
				#ifndef WRAPPER
				game.LoadingScreen(msg);
				#endif
			}
			
			m = game.utility.sGetParam(o);
			t = game.utility.sGetParam(o);
			mip = game.utility.bGetParam(o);
			fb = game.utility.bGetParam(o);
			sb = game.utility.bGetParam(o);
			//p.x = game.utility.fGetParam(o);
			blend = game.utility.bGetParam(o);
			//p.y = game.utility.fGetParam(o);
			//p.z = game.utility.fGetParam(o);
			bl = game.utility.fGetParam(o);
			bm = game.utility.fGetParam(o);
			//r = game.utility.fGetParam(o);
			p.zero();
			r = 0;
			d = game.utility.bGetParam(o);
			c = game.utility.bGetParam(o);
			f1 = game.utility.fGetParam(o);
			f2 = game.utility.fGetParam(o);
			rr = game.utility.fGetParam(o);
			rd = game.utility.fGetParam(o);
			
			for (int i = 0; i < numparams - expectedparams; i++)
			{
				extra = game.utility.sGetParam(o);
			}
			
			added = NULL;
			
			if (m != "" && m != game.utility.GetEOFString())
			{
				added = Add(p, r, m, t, mip, fb, sb, d, c, blend, packptr, f1, f2, bl, bm, rr, rd, parentnode.GetPtr());
				//collision.AddCollision(added);
				count++;
			}
		}
		
		o.close();

		//collision.GenerateCollisionTree();
	}
	else
	{
		error_log << "Couldn't open Object List: " << objectpath << "/list.txt" << endl;
	}
	
	if (packptr != NULL)
		packptr->ClosePack();
	
	if (parentnode.PtrValid())
	{
		//sort drawables by state for a free speedup
		parentnode.GetRef().SortDrawablesByRenderState();
	}
}

void OBJECTNODE::LoadModelIntoCollisionData()
{
	DestroyCollisionData();
	
	if (!model->jmodel.Loaded())
	{
		return;
	}
	
	vcount = model->jmodel.GetNumVerts();
	fcount = model->jmodel.GetFaces();
	vertices = new PHYSICSOBJECTVERTEX[vcount];
	faces = new PHYSICSOBJECTFACE[fcount];
	
	//copy normals
	/*for (int i = 0; i < fcount; i++)
	{
		VERTEX vertnorm[3];
		VERTEX facenorm;
		short * normidx = model->jmodel.GetNormIdx(i);
		for (int ni = 0; ni < 3; ni++)
		{
			vertnorm[ni].Set(model->jmodel.GetNorm(normidx[ni]));
			facenorm = facenorm + vertnorm[ni];
		}
		//VERTEX zerovert; if (facenorm.equals(zerovert)) facenorm.y = 1; //hack to prevent faces with zero normals
		facenorm = facenorm.normalize();
		
		for (int vi = 0; vi < 3; vi++)
		{
			normals[i].vertex[vi] = facenorm.v3()[vi];
		}
	}*/
	
	//copy verts
	//cout << "verts" << endl;
	for (int i = 0; i < vcount; i++)
	{
		for (int vi = 0; vi < 3; vi++)
		{
			vertices[i].vertex[vi] = model->jmodel.GetVert(i)[vi];
			//cout << model.GetVert(i)[vi] << endl;
		}
	}
	
	//copy faces
	//cout << "faces" << endl;
	for (int i = 0; i < fcount; i++)
	{
		/*for (int fi = 0; fi < 3; fi++)
		{
			faces[i].index[fi] = model->jmodel.GetFace(i)[fi];
			//cout << model.GetFace(i)[fi] << endl;
		}*/
		
		faces[i].index[0] = model->jmodel.GetFace(i)[0];
		faces[i].index[1] = model->jmodel.GetFace(i)[2];
		faces[i].index[2] = model->jmodel.GetFace(i)[1];
	}
}

void OBJECTNODE::DestroyCollisionData()
{
	if (vertices != NULL)
		delete [] vertices;
	if (faces != NULL)
		delete [] faces;
	//if (normals != NULL) delete [] normals;
	fcount = 0;
	vcount = 0;
	vertices = NULL;
	//normals = NULL;
	faces = NULL;
	if (physicsobject.PtrValid())
		game.physics.DeletePhysicsObject(physicsobject.GetPtr());
}

void OBJECTNODE::RegisterWithCollisionSystem()
{
	physicsobject.SetPtr(game.physics.NewPhysicsObject());
	
	PHYSICSOBJECTSETTINGS settings;
	settings.SetMapObject(true);
	//settings.SetObjectID(physicsobject.GetPtr());
	settings.SetObjectID(this);
	//physicsobject.GetPtr()->InitTrimesh((void *) vertices, 3*sizeof(float), vcount, (void *) faces, fcount*3, 3*sizeof(int), (void *)normals, settings);
	physicsobject.GetPtr()->InitTrimesh((void *) vertices, 3*sizeof(float), vcount, (void *) faces, fcount*3, 3*sizeof(int), NULL, settings);
	
	QUATERNION goofyfoot;
	goofyfoot.Rotate(-3.141593/2.0, 1,0,0);
	goofyfoot.Rotate(-3.141593/2.0, 0,1,0);
	physicsobject.GetPtr()->SetQuaternion(goofyfoot);
}

void OBJECTNODE::DestroySceneData()
{
	if (drawable.PtrValid())
		drawable.GetRef().GetParent()->Delete(drawable.GetPtr());
}
