#include "sound.h"
#include "settings.h"
#include "camera.h"
#include "utility.h"
#include "common.h"
#include "singleton.h"
#include "game.h"
extern GAME game;
#include "exception.h"
#include <vorbis/vorbisfile.h>

#define VDRIFT

#ifdef VDRIFT
#define PATHMANAGER_FUNCTION_CALL game.settings.GetFullDataPath
#define FILEEXISTS_FUNCTION_CALL game.utility.FileExists
#define GLOBAL_SOUND_OBJECT game.sound
extern bool verbose_output;
#define UNRECOVERABLE_ERROR_FUNCTION(file,line,msg) throw EXCEPTION(msg)
#else
#define PATHMANAGER_FUNCTION_CALL game.pathmanager.GetSoundPath
#define FILEEXISTS_FUNCTION_CALL game.pathmanager.FileExists
#define GLOBAL_SOUND_OBJECT game.sound
bool verbose_output = false;
#define UNRECOVERABLE_ERROR_FUNCTION(file,line,msg) throw EXCEPTION(file,line,msg)
#endif

const TESTER & SOUND::Test()
{
	int points = 0;
	
	//test loading a wave file
	SOUNDBUFFER testbuffer;
	SOUNDINFO sinfo;
	sinfo.Set(4096, 44100, 2, 2);
	
	/*testbuffer.Load(game.pathmanager.GetSoundPath("test/11k_u8_c1.wav"), sinfo);	
	points = 0;
	if (testbuffer.GetSoundInfo().GetFrequency() == 44100) points++;
	if (testbuffer.GetSoundInfo().GetBytesPerSample() == 2) points++;
	if (testbuffer.GetSoundInfo().GetChannels() == 1) points++;
	if (testbuffer.GetOriginalInfo().GetFrequency() == 11025) points++;
	if (testbuffer.GetOriginalInfo().GetBytesPerSample() == 1) points++;
	if (testbuffer.GetOriginalInfo().GetSamples() == 41040) points++;
	if (testbuffer.GetOriginalInfo().GetChannels() == 1) points++;
	mytest.SubTestComplete("Wave loading 11k u8 c1", points, 7);*/
	
	string testogg = "test/oggtest";
	testbuffer.Load(PATHMANAGER_FUNCTION_CALL(testogg+".ogg"), sinfo);	
	points = 0;
	if (testbuffer.GetSoundInfo().GetFrequency() == 44100) points++;
	if (testbuffer.GetSoundInfo().GetBytesPerSample() == 2) points++;
	if (testbuffer.GetSoundInfo().GetChannels() == 2) points++;
	if (testbuffer.GetSoundInfo().GetSamples() == 1412399) points++;
	if (testbuffer.GetSample16bit(1,500000) == -4490) points++;
	//cout << testbuffer.GetSample16bit(1,500000) << endl;
	/*if (testbuffer.GetOriginalInfo().GetFrequency() == 44100) points++;
	if (testbuffer.GetOriginalInfo().GetBytesPerSample() == 2) points++;
	if (testbuffer.GetOriginalInfo().GetSamples() == 35952) points++;
	if (testbuffer.GetOriginalInfo().GetChannels() == 2) points++;*/
	mytest.SubTestComplete("OGG Vorbis loading", points, 5);
	
	string testwav = "test/44k_s16_c2";
	testbuffer.Load(PATHMANAGER_FUNCTION_CALL(testwav+".wav"), sinfo);	
	points = 0;
	if (testbuffer.GetSoundInfo().GetFrequency() == 44100) points++;
	if (testbuffer.GetSoundInfo().GetBytesPerSample() == 2) points++;
	if (testbuffer.GetSoundInfo().GetChannels() == 2) points++;
	if (testbuffer.GetSoundInfo().GetSamples() == 35952) points++;
	if (testbuffer.GetSample16bit(1,1024) == 722) points++;
	/*if (testbuffer.GetOriginalInfo().GetFrequency() == 44100) points++;
	if (testbuffer.GetOriginalInfo().GetBytesPerSample() == 2) points++;
	if (testbuffer.GetOriginalInfo().GetSamples() == 35952) points++;
	if (testbuffer.GetOriginalInfo().GetChannels() == 2) points++;*/
	mytest.SubTestComplete("Wave loading 44k s16 c2", points, 5);
	
	SOUNDSOURCE testsrc;
	testsrc.SetBuffer(testbuffer);
	points = 0;
	testsrc.SeekToSample(1024);
	int out1, out2;
	testsrc.SampleAndAdvance16bit(&out1, &out2, 1);
	//cout << out1 << "," << out2 << endl;
	if (out1 == 0 && out2 == 0) points++;// else cout << "fail1" << endl;
	testsrc.Play();
	testsrc.SeekToSample(1014);
	testsrc.Sample16bit(10, out1, out2);
	if (out1 == 722 && out2 == -224) points++;// else cout << "fail2" << endl;
	testsrc.Advance(10);
	testsrc.SampleAndAdvance16bit(&out1, &out2, 1);
	if (out1 == 722 && out2 == -224) points++;// else cout << "fail3" << endl;
	//cout << out1 << "," << out2 << endl;
	mytest.SubTestComplete("Sound source testing", points, 3);
	
	SDL_Init(0);
	{
		points = 0;
		testsrc.SeekToSample(0);
		testsrc.SetLoop(true);
		unsigned int numsources = 32;
		unsigned int startticks = SDL_GetTicks();
		for (unsigned int i = 0; i < 44100; i++)
		{
			int accum1 = 0;
			int accum2 = 0;
			for (unsigned int s = 0; s < numsources; s++)
			{
				//testsrc.SampleAndAdvance16bit(out1, out2);
				testsrc.Sample16bit(i, out1, out2);
				accum1 += out1;
				accum2 += out2;
			}
		}
		unsigned int endticks = SDL_GetTicks();
		stringstream msg;
		msg << "Performance testing Sample16bit (" << endticks - startticks << "ms for 1 second of 44.1khz stereo audio with 32 sources)";
		if (endticks - startticks < 500)
			points++;
		mytest.SubTestComplete(msg.str(), points, 1);
	}
	{
		points = 0;
		testsrc.SeekToSample(0);
		//testsrc.SetLoop(false);
		//testsrc.SetLoop(true);
		unsigned int numsources = 32;
		unsigned int startticks = SDL_GetTicks();
		short stream[44100*2];
		int buffer1[44100];
		int buffer2[44100];
		for (unsigned int s = 0; s < numsources; s++)
		{
			testsrc.SampleAndAdvance16bit(buffer1, buffer2, 44100);
			if (s == 0)
			{
				for (int i = 0; i < 44100; i++)
				{
					int pos = i*2;
					((short *) stream)[pos] = buffer1[i];
					((short *) stream)[pos+1] = buffer2[i];
				}
			}
		}
		/*for (unsigned int i = 0; i < 44100; i++)
		{
			int accum1 = 0;
			int accum2 = 0;
			for (unsigned int s = 0; s < numsources; s++)
			{
				testsrc.SampleAndAdvance16bit(&out1, &out2, 1);
				//testsrc.Sample16bit(i, out1, out2);
				accum1 += out1;
				accum2 += out2;
			}
		}*/
		unsigned int endticks = SDL_GetTicks();
		stringstream msg;
		msg << "Performance testing SampleAndAdvance16bit (" << endticks - startticks << "ms for 1 second of 44.1khz stereo audio with 32 sources)";
		if (endticks - startticks < 500)
			points++;
		mytest.SubTestComplete(msg.str(), points, 1);
		testsrc.SetLoop(true);
		//cout << "Done" << endl;
	}
	{
		points = 0;
		testsrc.SeekToSample(0);
		testsrc.SetLoop(true);
		testsrc.SetPitch(0.7897878);
		//testsrc.SetPitch(0.1);
		unsigned int numsources = 32;
		unsigned int startticks = SDL_GetTicks();
		short stream[44100*2];
		int buffer1[44100];
		int buffer2[44100];
		for (unsigned int s = 0; s < numsources; s++)
		{
			testsrc.SampleAndAdvanceWithPitch16bit(buffer1, buffer2, 44100);
			if (s == 0)
			{
				for (int i = 0; i < 44100; i++)
				{
					int pos = i*2;
					((short *) stream)[pos] = buffer1[i];
					((short *) stream)[pos+1] = buffer2[i];
				}
			}
		}
		/*for (unsigned int i = 0; i < 44100; i++)
		{
			int accum1 = 0;
			int accum2 = 0;
			for (unsigned int s = 0; s < numsources; s++)
			{
				testsrc.SampleAndAdvanceWithPitch16bit(&out1, &out2, 1);
				//testsrc.Sample16bit(i, out1, out2);
				accum1 += out1;
				accum2 += out2;
			}
		}*/
		unsigned int endticks = SDL_GetTicks();
		stringstream msg;
		msg << "Performance testing SampleAndAdvanceWithPitch16bit (" << endticks - startticks << "ms for 1 second of 44.1khz stereo audio with 32 sources)";
		if (endticks - startticks < 500)
			points++;
		mytest.SubTestComplete(msg.str(), points, 1);
	}
	{
		points = 0;
		testsrc.SeekToSample(0);
		testsrc.SetLoop(true);
		testsrc.SetPitch(0.7897878);
		SOUNDFILTER filter;
		filter.SetFilterOrder1(0.1,0,0.9);
		unsigned int numsources = 32;
		unsigned int startticks = SDL_GetTicks();
		short stream[44100*2];
		int buffer1[44100];
		int buffer2[44100];
		for (unsigned int s = 0; s < numsources; s++)
		{
			testsrc.SampleAndAdvanceWithPitch16bit(buffer1, buffer2, 44100);
			filter.Filter(buffer1, buffer2, 44100);
			if (s == 0)
			{
				for (int i = 0; i < 44100; i++)
				{
					int pos = i*2;
					((short *) stream)[pos] = buffer1[i];
					((short *) stream)[pos+1] = buffer2[i];
				}
			}
		}
		/*for (unsigned int i = 0; i < 44100; i++)
		{
			int accum1 = 0;
			int accum2 = 0;
			for (unsigned int s = 0; s < numsources; s++)
			{
				testsrc.SampleAndAdvanceWithPitch16bit(&out1, &out2, 1);
				//testsrc.Sample16bit(i, out1, out2);
				accum1 += out1;
				accum2 += out2;
			}
		}*/
		unsigned int endticks = SDL_GetTicks();
		stringstream msg;
		msg << "Performance testing SampleAndAdvanceWithPitch16bit with 1st order filter (" << endticks - startticks << "ms for 1 second of 44.1khz stereo audio with 32 sources)";
		if (endticks - startticks < 1000)
			points++;
		mytest.SubTestComplete(msg.str(), points, 1);
	}
	SDL_Quit();
	
	testsrc.SeekToSample(1014);
	testsrc.Sample16bit(10, out1, out2);
	deviceinfo = sinfo;
	LoadBuffer(testwav);
	SOUNDSOURCE & testsrc2 = NewSource(testwav);
	testsrc2.SetPitch(0.7);
	testsrc2.SetLoop(true);
	testsrc2.Play();
	{
		int numsamp = 512;
		Uint8 stream[numsamp];
		loaded = true;
		initdone = true;
		testsrc.Play();
		Callback16bitstereo(NULL, stream, numsamp);
		testsrc.Stop();
		initdone = false;
		loaded = false;
		ofstream testdump;
		testdump.open(PATHMANAGER_FUNCTION_CALL("test/testdump.txt").c_str());
		ifstream testgoal;
		testgoal.open(PATHMANAGER_FUNCTION_CALL("test/testgoal.txt").c_str());
		//points = 256*2;
		points = 0;
		for (int i = 0; i < numsamp; i++)
		{
			short goal;
			testgoal >> goal;
			testdump << dec << (short)stream[i] << endl;
			if (stream[i] - goal < 2 && goal - stream[i] < 2)
				points++;
		//	else
		//		cout << (short)goal << " != " << (short)stream[i] << endl;
		}
		testdump.close();
		mytest.SubTestComplete("Sound callback testing", points, numsamp);
	}
	points = 0;
	testsrc2.SeekToSample(1024);
	int out3, out4;
	testsrc2.SampleAndAdvance16bit(&out3, &out4, 1);
	//cout << out1 << "," << out2 << endl;
	if (out1 == out3 && out2 == out4) points++;
	SOUNDSOURCE & testsrc3 = NewSource(testwav);
	testsrc3.Play();
	list <SOUNDSOURCE *> active_sourcelist;
	list <SOUNDSOURCE *> inactive_sourcelist;
	DetermineActiveSources(active_sourcelist, inactive_sourcelist);
	if (active_sourcelist.size() == 2) points++;
	testsrc3.Stop();
	DetermineActiveSources(active_sourcelist, inactive_sourcelist);
	if (active_sourcelist.size() == 1 && inactive_sourcelist.size() == 1) points++;
	testsrc2.Stop();
	DetermineActiveSources(active_sourcelist, inactive_sourcelist);
	if (active_sourcelist.size() == 0 && inactive_sourcelist.size() == 2) points++;
	DeleteSource(&testsrc3);
	if (sourcelist.size() == 1) points++;
	DeleteSource(&testsrc2);
	if (sourcelist.size() == 0) points++;
	mytest.SubTestComplete("Sound management testing", points, 6);
	
	points = 0;
	SOUNDSOURCE & testsrc4 = NewSource(testwav);
	active_sourcelist.clear();
	active_sourcelist.push_back(&testsrc4);
	testsrc4.Set3DEffects(false);
	testsrc4.SetGain(0.5);
	VERTEX spos;
	spos.Set(300,0,0);
	testsrc4.SetPosition(spos);
	VERTEX lpos;
	QUATERNION lrot;
	Compute3DEffects(active_sourcelist, lpos, lrot);
	if (testsrc4.ComputedGain(1) == 0.5 && testsrc4.ComputedGain(2) == 0.5) points++;
	testsrc4.Set3DEffects(true);
	Compute3DEffects(active_sourcelist, lpos, lrot);
	if (testsrc4.ComputedGain(2) == 0.0) points++;
	if (testsrc4.ComputedGain(1) < 0.1) points++;
	mytest.SubTestComplete("Sound effects testing", points, 3);
	
	{
		points = 0;
		SOUNDFILTER filt1;
		int size = 10;
		int data1[size];
		int data2[size];
		for (int i = 0; i < size; i++)
		{
			if (i < 3)
			{
				data1[i] = data2[i] = 0;
			}
			else
			{
				data1[i] = 1000;
				data2[i] = 0;
			}
		}
		float xc[2];
		float yc[2];
		xc[0] = 0.1; //simple low-pass filter
		xc[1] = 0;
		yc[0] = 0;
		yc[1] = 0.9;
		filt1.SetFilter(1, xc, yc);
		filt1.Filter(data1,data2,5);
		if (data1[3] == 100 && data2[3] == 0) points++;
		filt1.Filter(&(data1[5]),&(data2[5]),5);
		if (data1[6] == 342 && data1[9] == 519 && data2[9] == 0) points++;
		//cout << data1[6] << endl;
		//for (int i = 0; i < 10; i++) cout << i << ". " << data1[i] << endl;
		
		mytest.SubTestComplete("Sound filter testing", points, 2);
	}
	
	return mytest;
}

void SOUND_CallbackWrapper(void *unused, Uint8 *stream, int len)
{
	GLOBAL_SOUND_OBJECT.Callback16bitstereo(unused, stream, len);
}

extern bool verbose_output;

void SOUND::Init()
{
	if (disable || initdone)
		return;
	
	SDL_AudioSpec desired, obtained;
	
	desired.freq = 44100;
	desired.format = AUDIO_S16;
	desired.samples = 512;
	desired.callback = SOUND_CallbackWrapper;
	desired.userdata = NULL;
	desired.channels = 2;
	
	if (SDL_OpenAudio(&desired, &obtained) < 0)
	{
		//string error = SDL_GetError();
		UNRECOVERABLE_ERROR_FUNCTION(__FILE__,__LINE__,"Error opening audio device.");
		//throw EXCEPTION(__FILE__, __LINE__, "Error opening audio device: " + error);
		DisableAllSound();
		return;
	}
	
	int frequency = obtained.freq;
	int channels = obtained.channels;
	int samples = obtained.samples;
	int bytespersample = 2;
	if (obtained.format == AUDIO_U8 || obtained.format == AUDIO_S8)
	{
		bytespersample = 1;
	}
	
	stringstream dout;
	dout << "Obtained audio device:" << endl;
	dout << "Frequency: " << frequency << endl;
	dout << "Format: " << obtained.format << endl;
	dout << "Bits per sample: " << bytespersample * 8 << endl;
	dout << "Channels: " << channels << endl;
	dout << "Silence: " << (int) obtained.silence << endl;
	dout << "Samples: " << samples << endl;
	dout << "Size: " << (int) obtained.size << endl;
	game.WriteDebuggingData(dout.str());
	
	if (bytespersample != 2 || obtained.channels != desired.channels || obtained.freq != desired.freq)
	{
		//throw EXCEPTION(__FILE__, __LINE__, "Sound did not create a 44.1kHz, 16 bit, stereo device as requested.");
		//cerr << __FILE__ << "," << __LINE__ << ": Sound interface did not create a 44.1kHz, 16 bit, stereo device as requested.  Disabling game.sound." << endl;
		UNRECOVERABLE_ERROR_FUNCTION(__FILE__,__LINE__,"Sound interface did not create a 44.1kHz, 16 bit, stereo device as requested.  Disabling game.sound.");
		DisableAllSound();
		return;
	}

	deviceinfo.Set(samples, frequency, channels, bytespersample);

	UpdateSettings();

	initdone = true;
}

void SOUND::UpdateSettings()
{
	float master_volume = 0.5;
	game.settings.Get( "sound.volume", master_volume );
	SetMasterVolume( master_volume );
}

void SOUND::Pause(const bool pause_on)
{
	paused = pause_on;
	if (pause_on)
	{
		//cout << "sound pause on" << endl;
		SDL_PauseAudio(1);
	}
	else
	{
		//cout << "sound pause off" << endl;
		SDL_PauseAudio(0);
	}
}

void SOUND::Callback16bitstereo(void *unused, Uint8 *stream, int len)
{
	if (!loaded || !initdone)
		return;
	
	list <SOUNDSOURCE *> active_sourcelist;
	list <SOUNDSOURCE *> inactive_sourcelist;
	DetermineActiveSources(active_sourcelist, inactive_sourcelist);
	Compute3DEffects(active_sourcelist);//, cam.GetPosition().ScaleR(-1), cam.GetRotation());

	//increment inactive sources
	ITERATE(SOUNDSOURCE*, inactive_sourcelist, s)
	{
		SOUNDSOURCE * src = *s;
		src->IncrementWithPitch(len/4);
	}
	
	int buffer1[len/4];
	int buffer2[len/4];
	ITERATE(SOUNDSOURCE*, active_sourcelist, s)
	{
		SOUNDSOURCE * src = *s;
		src->SampleAndAdvanceWithPitch16bit(buffer1, buffer2, len/4);
		for (int f = 0; f < src->NumFilters(); f++)
		{
			src->GetFilter(f).Filter(buffer1, buffer2, len/4);
		}
		volume_filter.Filter(buffer1, buffer2, len/4);
		if (s == active_sourcelist.begin())
		{
			for (int i = 0; i < len/4; i++)
			{
				int pos = i*2;
				((short *) stream)[pos] = ENDIAN_SWAP_16(buffer1[i]);
				((short *) stream)[pos+1] = ENDIAN_SWAP_16(buffer2[i]);
			}
		}
		else
		{
			for (int i = 0; i < len/4; i++)
			{
				int pos = i*2;
				((short *) stream)[pos] += ENDIAN_SWAP_16(buffer1[i]);
				((short *) stream)[pos+1] += ENDIAN_SWAP_16(buffer2[i]);
			}
		}
	}
	
	//cout << active_sourcelist.size() << "," << inactive_sourcelist.size() << endl;
	
	if (active_sourcelist.empty())
	{
		for (int i = 0; i < len/4; i++)
		{
			int pos = i*2;
			((short *) stream)[pos] = ((short *) stream)[pos+1] = 0;
		}
	}
	
	/*for (int i = 0; i < len/4; i++)
	{
		int out1 = 0;
		int out2 = 0;
		
		ITERATE(SOUNDSOURCE*, active_sourcelist, s)
		{
			int sample1, sample2;
			SOUNDSOURCE * src = *s;
			//src->SampleAndAdvance16bit(sample1, sample2);
			src->SampleAndAdvanceWithPitch16bit(&sample1, &sample2, 1);
			out1 += sample1;
			out2 += sample2;
		}
		
		const int maxval = 32768;
		const int minval = -32767;
		
		if (out1 > maxval) out1 = maxval;
		if (out2 > maxval) out2 = maxval;
		if (out1 < minval) out1 = minval;
		if (out2 < minval) out2 = minval;
		
		((short *) stream)[i*2] = out1;
		((short *) stream)[i*2+1] = out2;
	}*/
	
	//cout << "Callback: " << len << endl;
}

SOUND::~SOUND()
{
	if (loaded)
	{
		SDL_CloseAudio();
	}
}

void SOUND::Load()
{
	if (disable)
		return;
	
	/*//load all car-specific sounds
	list<string> carlist;
	string folderpath = game.settings.GetFullDataPath("/cars");
	game.utility.GetFolderIndex(folderpath, carlist);
	for( list<string>::iterator i = carlist.begin(); i != carlist.end(); i++ )
	{
		string car_about_filename = folderpath + "/" + *i + "/about.txt";
		if( game.utility.FileExists( car_about_filename ) )
		{
			LoadBuffer("cars/"+*i+"/engine");
		}
	}
	
	//load other (generic) sounds
	LoadBuffer("sounds/tire_squeal");*/
	
	loaded = true;
}

void SOUNDINFO::DebugPrint() const
{
	cout << "Samples: " << samples << endl;
	cout << "Frequency: " << frequency << endl;
	cout << "Channels: " << channels << endl;
	cout << "Bits per sample: " << bytespersample*8 << endl;
}

void SOUNDBUFFER::Load(const string & filename, const SOUNDINFO & sound_device_info)
{
	if (filename.find(".wav") < filename.length())
		LoadWAV(filename, sound_device_info);
	else if (filename.find(".ogg") < filename.length())
		LoadOGG(filename, sound_device_info);
	else
		throw EXCEPTION(__FILE__, __LINE__, "Can't determine sound file format based on extension: " + filename);
}

void SOUNDBUFFER::LoadWAV(const string & filename, const SOUNDINFO & sound_device_info)
{
	if (loaded)
		Unload();
	
	name = filename;
	
	FILE *fp;
	
	unsigned int size;
	
	fp = fopen(filename.c_str(), "rb");
	if (fp)
	{
		char id[5]; //four bytes to hold 'RIFF'
		
		fread(id,sizeof(char),4,fp); //read in first four bytes
		id[4] = '\0';
		if (!strcmp(id,"RIFF"))
		{ //we had 'RIFF' let's continue
			fread(&size,sizeof(unsigned int),1,fp); //read in 32bit size value
			size = ENDIAN_SWAP_32(size);
			fread(id,sizeof(char),4,fp); //read in 4 byte string now
			if (!strcmp(id,"WAVE"))
			{ //this is probably a wave file since it contained "WAVE"
				fread(id,sizeof(char),4,fp); //read in 4 bytes "fmt ";
				if (!strcmp(id,"fmt "))
				{
					unsigned int format_length, sample_rate, avg_bytes_sec;
					short format_tag, channels, block_align, bits_per_sample;
					
					fread(&format_length, sizeof(unsigned int),1,fp);
					format_length = ENDIAN_SWAP_32(format_length);					
					fread(&format_tag, sizeof(short), 1, fp);
					format_tag = ENDIAN_SWAP_16(format_tag);										
					fread(&channels, sizeof(short),1,fp);
					channels = ENDIAN_SWAP_16(channels);															
					fread(&sample_rate, sizeof(unsigned int), 1, fp);
					sample_rate = ENDIAN_SWAP_32(sample_rate);										
					fread(&avg_bytes_sec, sizeof(unsigned int), 1, fp);
					avg_bytes_sec = ENDIAN_SWAP_32(avg_bytes_sec);															
					fread(&block_align, sizeof(short), 1, fp);
					block_align = ENDIAN_SWAP_16(block_align);															
					fread(&bits_per_sample, sizeof(short), 1, fp);
					bits_per_sample = ENDIAN_SWAP_16(bits_per_sample);															
	
					
					//new wave seeking code
					//find data chunk
					bool found_data_chunk = false;
					long filepos = format_length + 4 + 4 + 4 + 4 + 4;
					int chunknum = 0;
					while (!found_data_chunk && chunknum < 10)
					{
						fseek(fp, filepos, SEEK_SET); //seek to the next chunk
						fread(id, sizeof(char), 4, fp); //read in 'data'
						fread(&size, sizeof(unsigned int), 1, fp); //how many bytes of sound data we have
						size = ENDIAN_SWAP_32(size);						
						if (!strcmp(id,"data"))
						{
							found_data_chunk = true;
						}
						else
						{
							filepos += size + 4 + 4;
						}
						
						chunknum++;
					}
					
					if (chunknum >= 10)
					{
						//cerr << __FILE__ << "," << __LINE__ << ": Sound file contains more than 10 chunks before the data chunk: " + filename << endl;
						return;
					}

					sound_buffer = new char[size];
					
					fread(sound_buffer, sizeof(char), size, fp); //read in our whole sound data chunk
					
					/* THIS IS JUST AN BIG ENDIAN EXPIRIMENT SHOULD BE DELETED IF IT DOESN'T WORK OUT
					#if SDL_BYTEORDER == SDL_BIG_ENDIAN
					if (bits_per_sample == 16)
					{
						for (int i = 0; i < size; i+=2)
						{
							cout << "preswap i: " << sound_buffer[i] << "preswap i+1: " << sound_buffer[i+1] << endl;
							((short *)sound_buffer)[i] = ENDIAN_SWAP_16(((short *)sound_buffer)[i]);
							cout << "postswap i: " << sound_buffer[i] << "postswap i+1: " << sound_buffer[i+1] << endl;							
						}
					}
					else if (bits_per_sample != 8)
						UNRECOVERABLE_ERROR_FUNCTION(__FILE__,__LINE__,"Sound file with x bits per sample not supported on BIG ENDIAN! x = " + bits_per_sample);
					#endif*/
					
					info.Set(size/((bits_per_sample/8)*channels), sample_rate, channels, bits_per_sample/8);
					SOUNDINFO original_info;
					original_info.Set(size/((bits_per_sample/8)*channels), sample_rate, channels, bits_per_sample/8);
					
					loaded = true;
					SOUNDINFO desired_info;
					desired_info.Set(original_info.GetSamples(), sound_device_info.GetFrequency(), original_info.GetChannels(), sound_device_info.GetBytesPerSample());
					//ConvertTo(desired_info);
					if (desired_info == original_info)
					{
						
					}
					else
					{
						cout << "SOUND FORMAT:" << endl;
						original_info.DebugPrint();
						cout << "DESIRED FORMAT:" << endl;
						desired_info.DebugPrint();
						//throw EXCEPTION(__FILE__, __LINE__, "Sound file isn't in desired format: " + filename);
						//cerr << __FILE__ << "," << __LINE__ << ": Sound file isn't in desired format: " + filename << endl;
						UNRECOVERABLE_ERROR_FUNCTION(__FILE__,__LINE__,"Sound file isn't in desired format: "+filename);
						return;
					}
					
					
					/*fread(id, sizeof(char), 4, fp); //read in 'data'
					fread(&size, sizeof(unsigned int), 1, fp); //how many bytes of sound data we have
					//sound_buffer = (BYTE *) malloc (sizeof(BYTE) * data_size); //set aside sound buffer space
					sound_buffer = new char[size];
					fread(sound_buffer, sizeof(char), size, fp); //read in our whole sound data chunk 
					
					//original_info.Set(size/((bits_per_sample/8)*channels), sample_rate, channels, bits_per_sample/8);
					info.Set(size/((bits_per_sample/8)*channels), sample_rate, channels, bits_per_sample/8);
					SOUNDINFO original_info;
					original_info.Set(size/((bits_per_sample/8)*channels), sample_rate, channels, bits_per_sample/8);
					//original_info.Set(size, sample_rate, channels, bits_per_sample/8);
					//original_info.DebugPrint();
					
					loaded = true;
					SOUNDINFO desired_info;
					desired_info.Set(original_info.GetSamples(), sound_device_info.GetFrequency(), original_info.GetChannels(), sound_device_info.GetBytesPerSample());
					//ConvertTo(desired_info);
					if (desired_info == original_info)
					{
						
					}
					else
					{
						cout << "SOUND FORMAT:" << endl;
						original_info.DebugPrint();
						cout << "DESIRED FORMAT:" << endl;
						desired_info.DebugPrint();
						//throw EXCEPTION(__FILE__, __LINE__, "Sound file isn't in desired format: " + filename);
						cerr << __FILE__ << "," << __LINE__ << ": Sound file isn't in desired format: " + filename << endl;
						return;
					}*/
				}
				else
				{
					//throw EXCEPTION(__FILE__, __LINE__, "Sound file doesn't have \"fmt \" header: " + filename);
					//cerr << __FILE__ << "," << __LINE__ << ": Sound file doesn't have \"fmt \" header: " + filename << endl;
					UNRECOVERABLE_ERROR_FUNCTION(__FILE__,__LINE__,"Sound file doesn't have \"fmt\" header: "+filename);
					return;
				}
			}
			else
			{
				//throw EXCEPTION(__FILE__, __LINE__, "Sound file doesn't have WAVE header: " + filename);
				//cerr << __FILE__ << "," << __LINE__ << ": Sound file doesn't have WAVE header: " + filename << endl;
				UNRECOVERABLE_ERROR_FUNCTION(__FILE__,__LINE__,"Sound file doesn't have WAVE header: "+filename);
				return;
			}
		}
		else
		{
			//throw EXCEPTION(__FILE__, __LINE__, "Sound file doesn't have RIFF header: " + filename);
			//cerr << __FILE__ << "," << __LINE__ << ": Sound file doesn't have WAVE header: " + filename << endl;
			UNRECOVERABLE_ERROR_FUNCTION(__FILE__,__LINE__,"Sound file doesn't have RIFF header: "+filename);
			return;
		}
	}
	else
	{
		//throw EXCEPTION(__FILE__, __LINE__, "Can't open sound file: " + filename);
		//cerr << __FILE__ << "," << __LINE__ << ": Can't open sound file: " + filename << endl;
		UNRECOVERABLE_ERROR_FUNCTION(__FILE__,__LINE__,"Can't open sound file: "+filename);
		return;
	}
	
	//cout << size << endl;
}

void SOUNDBUFFER::LoadOGG(const string & filename, const SOUNDINFO & sound_device_info)
{
	if (loaded)
		Unload();
	
	name = filename;
	
	FILE *fp;
	
	unsigned int samples;
	
	fp = fopen(filename.c_str(), "rb");
	if (fp)
	{
		vorbis_info *pInfo;
		OggVorbis_File oggFile;

		ov_open(fp, &oggFile, NULL, 0);
		
		pInfo = ov_info(&oggFile, -1);
		
		//I assume ogg is always 16-bit (2 bytes per sample) -Venzon
		samples = ov_pcm_total(&oggFile,-1);
		info.Set(samples, pInfo->rate, pInfo->channels, 2);
		
		SOUNDINFO desired_info;
		desired_info.Set(info.GetSamples(), sound_device_info.GetFrequency(), sound_device_info.GetChannels(), sound_device_info.GetBytesPerSample());
		
		if (!(desired_info == info))
		{
			cout << "SOUND FORMAT:" << endl;
			info.DebugPrint();
			cout << "DESIRED FORMAT:" << endl;
			desired_info.DebugPrint();
			
			UNRECOVERABLE_ERROR_FUNCTION(__FILE__,__LINE__,"Sound file isn't in desired format: "+filename);
			ov_clear(&oggFile);
			return;
		}
		
		//allocate space
		unsigned int size = info.GetSamples()*info.GetChannels()*info.GetBytesPerSample();
		sound_buffer = new char[size];
		int bitstream;
		int endian = 0; //0 for Little-Endian, 1 for Big-Endian
		int wordsize = 2; //again, assuming ogg is always 16-bits
		int issigned = 1; //use signed data
		
		int bytes = 1;
		unsigned int bufpos = 0;
		while (bytes > 0)
		{
			bytes = ov_read(&oggFile, sound_buffer+bufpos, size-bufpos, endian, wordsize, issigned, &bitstream);
			bufpos += bytes;
			//cout << bytes << "...";
		}
		
		loaded = true;
		
		//note: no need to call fclose(); ov_clear does it for us
		ov_clear(&oggFile);
	}
	else
	{
		UNRECOVERABLE_ERROR_FUNCTION(__FILE__,__LINE__,"Can't open sound file: "+filename);
		return;
	}
}

/*void SOUNDBUFFER::ConvertTo(const SOUNDINFO & to_info)
{
	if (!loaded || sound_buffer == NULL)
	{
		throw EXCEPTION(__FILE__, __LINE__, "Can't convert unloaded sound buffer.");
		return;
	}
	
	char * newbuffer;
	cout << "CONVERTING...";
	
	if (original_info.GetFrequency() < to_info.GetFrequency())
	{
		//up-sample
		
		if (to_info.GetFrequency() % original_info.GetFrequency() != 0)
		{
			throw EXCEPTION(__FILE__, __LINE__, "Non-integer up-sample not yet supported.");
			return;
		}
		
		int sampleratio = to_info.GetFrequency() / original_info.GetFrequency();
		
		unsigned int new_size = original_info.GetSamples()*to_info.GetBytesPerSample()*to_info.GetChannels()*sampleratio;
		newbuffer = new char[new_size];
		unsigned int old_size = original_info.GetSamples()*original_info.GetBytesPerSample()*original_info.GetChannels();
		
		char sample_u8[2];
		short sample_s16[2];
		
		for (unsigned int i = 0; i < old_size; i++)
		{
			//get old sample
			if (original_info.GetBytesPerSample() == 1)
			{
				sample_u8[0] = sound_buffer[i*original_info.GetChannels()]
				if (original_info.GetChannels() == 2)
				{
					sample_u8[1] = sound_buffer[i*original_info.GetChannels()+1];
				}
			}
			else if (original_info.GetBytesPerSample() == 2)
			{
				sample_u8[0] = sound_buffer[i*original_info.GetChannels()]
				if (original_info.GetChannels() == 2)
				{
					sample_u8[1] = sound_buffer[i*original_info.GetChannels()+1];
				}
			}
		}
		
		char * oldbuf = sound_buffer;
		sound_buffer = newbuffer;
		delete [] oldbuf;
	}
	else
	{
		//down-sample
		
		throw EXCEPTION(__FILE__, __LINE__, "Down-sample not yet supported.");
		return;
	}
	
	cout << "done";
}*/

inline const int SOUNDBUFFER::GetSample16bit(const unsigned int channel, const unsigned int position)
{
	//line below is used for performance testing to find bottlenecks
	//return 123;
	
	#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	short result = ((short *)sound_buffer)[position*info.channels+(channel-1)*(info.channels-1)];
	short swapped_result = ENDIAN_SWAP_16(result);
	return swapped_result;
	#else
	return ((short *)sound_buffer)[position*info.channels+(channel-1)*(info.channels-1)];
	#endif 
}

inline const int SOUNDINFO::GetChannels() const
{
	return channels;
}

#define PREFER_BRANCHING_OVER_MULTIPLY
#define FASTER_SOUNDBUFFER_ACCESS
#define FAST_SOUNDBUFFER_ACCESS

inline void SOUNDSOURCE::SampleAndAdvance16bit(int * chan1, int * chan2, int len)
{
	#ifdef FAST_SOUNDBUFFER_ACCESS
	//if (buffer->info.channels == 2)
	{
		unsigned int samples = buffer->info.samples;
		
		#ifdef FASTER_SOUNDBUFFER_ACCESS
		unsigned int remaining_len = len;
		unsigned int this_start = 0;
		while (remaining_len > 0)
		{
			unsigned int this_run = remaining_len;
			unsigned int this_end = this_start+this_run;
			
			if (playing)
			{
				if (sample_pos + remaining_len > samples)
				{
					this_run = samples - sample_pos;
					this_end = this_start+this_run;
				}
				
				if (buffer->info.channels == 2)
				{
					for (unsigned int i = this_start, n = sample_pos*2; i < this_end; i++,n+=2)
					{
						#if SDL_BYTEORDER == SDL_BIG_ENDIAN
						short result = ((short *)buffer->sound_buffer)[n];
						short swapped_result = ENDIAN_SWAP_16(result);
						chan1[i] = swapped_result;
						result = ((short *)buffer->sound_buffer)[n+1];
						swapped_result = ENDIAN_SWAP_16(result);
						chan2[i] = swapped_result;
						#else
						chan1[i] = ((short *)buffer->sound_buffer)[n];
						chan2[i] = ((short *)buffer->sound_buffer)[n+1];
						#endif
					}
				}
				else //it is assumed that this means channels == 1
				{
					for (unsigned int i = this_start, n = sample_pos; i < this_end; i++, n++)
					{
						#if SDL_BYTEORDER == SDL_BIG_ENDIAN
						short result = ((short *)buffer->sound_buffer)[n];
						short swapped_result = ENDIAN_SWAP_16(result);
						chan1[i] = swapped_result;
						result = ((short *)buffer->sound_buffer)[n+1];
						swapped_result = ENDIAN_SWAP_16(result);
						chan2[i] = swapped_result;
						#else
						chan1[i] = chan2[i] = ((short *)buffer->sound_buffer)[n];
						#endif
					}
				}
			}
			else
			{
				//cout << "Finishing run of " << this_run << endl;
				for (unsigned int i = this_start; i < this_end; i++)
				{
					chan1[i] = chan2[i] = 0;
				}
			}
			
			unsigned int newpos = sample_pos + this_run;
			if (newpos >= samples && !loop)
			{
				playing = 0;
				//cout << "Stopping play" << endl;
			}
			else
			{
				sample_pos = (newpos) % samples;
				//cout << newpos << "<" << samples << " loop: " << loop << endl;
			}
			
			//cout << remaining_len << " - " << this_run << endl;
			remaining_len -= this_run;
			this_start += this_run;
		}
		
		//cout << remaining_len << endl;
		#else
		if (buffer->info.channels == 2)
		{
			for (int i = 0; i < len; i++)
			{
				if (playing)
				{
					#if SDL_BYTEORDER == SDL_BIG_ENDIAN
					chan1[i] = (int) (buffer->GetSample16bit(1, sample_pos)*computed_gain1);
					chan2[i] = (int) (buffer->GetSample16bit(2, sample_pos)*computed_gain2);
					#else
					int sample_idx = sample_pos*2;
					chan1[i] = ((short *)buffer->sound_buffer)[sample_idx];
					chan2[i] = ((short *)buffer->sound_buffer)[sample_idx+1];
					#endif
					
					unsigned int newpos = sample_pos + 1;
					if (newpos >= samples && !loop)
						playing = 0;
					else
						sample_pos = (newpos) % samples;
				}
				else
				{
					chan1[i] = chan2[i] = 0;
				}
			}
		}
		else //channels == 1
		{
			for (int i = 0; i < len; i++)
			{
				if (playing)
				{
					#if SDL_BYTEORDER == SDL_BIG_ENDIAN
					chan1[i] = (int) (buffer->GetSample16bit(1, sample_pos)*computed_gain1);
					chan2[i] = (int) (buffer->GetSample16bit(2, sample_pos)*computed_gain2);
					#else
					chan1[i] = ((short *)buffer->sound_buffer)[sample_pos];
					chan2[i] = ((short *)buffer->sound_buffer)[sample_pos];
					#endif
					
					int newpos = sample_pos + 1;
					if (!loop && newpos >= samples && playing)
						playing = 0;
					else
						sample_pos = (newpos) % samples;
				}
				else
				{
					chan1[i] = chan2[i] = 0;
				}
			}
		}
		#endif
	}
	/*else //it's assumed this means channels == 1
	{
		unsigned int samples = buffer->info.samples;
	
		#ifdef FASTER_SOUNDBUFFER_ACCESS
		unsigned int remaining_len = len;
		while (remaining_len > 0)
		{
			unsigned int this_run = remaining_len;
			
			if (playing)
			{			
				if (sample_pos + remaining_len > samples)
				{
					this_run = samples - sample_pos;
				}
				
				for (unsigned int i = 0, n = sample_pos; i < this_run; i++, n++)
				{
					#if SDL_BYTEORDER == SDL_BIG_ENDIAN
					short result = ((short *)buffer->sound_buffer)[n];
					short swapped_result = ENDIAN_SWAP_16(result);
					chan1[i] = swapped_result;
					result = ((short *)buffer->sound_buffer)[n+1];
					swapped_result = ENDIAN_SWAP_16(result);
					chan2[i] = swapped_result;
					#else
					chan1[i] = ((short *)buffer->sound_buffer)[n];
					chan2[i] = ((short *)buffer->sound_buffer)[n];
					#endif
				}
			}
			else
			{
				cout << "Finishing up run of " << this_run << endl;
				for (unsigned int i = 0; i < this_run; i++)
				{
					chan1[i] = chan2[i] = 0;
				}
			}
			
			unsigned int newpos = sample_pos + this_run;
			if (newpos >= samples && !loop && playing)
				playing = 0;
			else
				sample_pos = (newpos) % samples;
			
			remaining_len -= this_run;
		}
		#else
		for (int i = 0; i < len; i++)
		{
			if (playing)
			{
				#if SDL_BYTEORDER == SDL_BIG_ENDIAN
				chan1[i] = (int) (buffer->GetSample16bit(1, sample_pos)*computed_gain1);
				chan2[i] = (int) (buffer->GetSample16bit(2, sample_pos)*computed_gain2);
				#else
				chan1[i] = ((short *)buffer->sound_buffer)[sample_pos];
				chan2[i] = ((short *)buffer->sound_buffer)[sample_pos];
				#endif
				
				int newpos = sample_pos + 1;
				if (newpos >= samples && !loop)
					playing = 0;
				else
					sample_pos = (newpos) % samples;
			}
			else
			{
				chan1[i] = chan2[i] = 0;
			}
		}
		#endif
	}*/
	#else
	for (int i = 0; i < len; i++)
	{
		#ifdef PREFER_BRANCHING_OVER_MULTIPLY
		if (playing)
		{
			chan1[i] = (int) (buffer->GetSample16bit(1, sample_pos)*computed_gain1);
			chan2[i] = (int) (buffer->GetSample16bit(2, sample_pos)*computed_gain2);
			
			Increment(1);
		}
		else
		{
			chan1[i] = chan2[i] = 0;
		}
		#else
		chan1[i] = (int) (buffer->GetSample16bit(1, sample_pos)*playing*computed_gain1);
		chan2[i] = (int) (buffer->GetSample16bit(2, sample_pos)*playing*computed_gain2);
		
		//stop playing if at the end of loop and not looping
		//this line is a complicated piece of engineering to avoid branching...
		//consider that (sample_pos + 1) / buffer.GetPtr()->GetSoundInfo().GetSamples() is 1 if we're at the end, 0 if we're not
		//let e = (sample_pos + 1) / buffer.GetPtr()->GetSoundInfo().GetSamples()
		//let l = loop
		//let p = playing
		//X means "don't care"
		//we want:
		//e=0 && l=X => p=1
		//e=1 && l=1 => p=1
		//e=1 && l=0 => p=0
		playing = 1 - ((sample_pos + 1) / samples)*(1-loop);
		
		sample_pos = (sample_pos + 1*playing) % samples;
		#endif
	}
	#endif
}

inline void SOUNDSOURCE::Increment(int num)
{
	int samples = buffer->GetSoundInfo().GetSamples();
	
	int newpos = sample_pos + num;
	if (newpos >= samples && !loop)
		playing = 0;
	else
		sample_pos = (newpos) % samples;
}

//#define RESAMPLE_QUALITY_LOWEST
//#define RESAMPLE_QUALITY_MEDIUM
//#define RESAMPLE_QUALITY_FAST_MEDIUM
#define RESAMPLE_QUALITY_FASTER_MEDIUM
//#define RESAMPLE_QUALITY_HIGH

void SOUNDSOURCE::SampleAndAdvanceWithPitch16bit(int * chan1, int * chan2, int len)
{
	/*if (pitch == 1.0f)
	{
		SampleAndAdvance16bit(chan1, chan2, len);
		return;
	}*/
	
	unsigned int samples = buffer->info.samples;
	int samp1[2];
	int samp2[2];
	unsigned int remaining_len = len;
	unsigned int this_start = 0;
	
	//float thispitch = pitch;
	
	while (remaining_len > 0)
	{
		unsigned int this_run = remaining_len;
		unsigned int this_end = this_start+this_run;
		
		float n_remain = sample_pos_remainder;
		unsigned int n = sample_pos;
		
		float samplecount = 0;
		
		if (playing)
		{
			if (sample_pos + sample_pos_remainder + remaining_len*pitch > samples)
			{
				this_run = (unsigned int)( (samples - (sample_pos+sample_pos_remainder))*(1.0/pitch) )+1;
				this_end = this_start+this_run;
			}
			
			const float maxrate = 1.0/(44100.0*0.01);
			const float negmaxrate = -maxrate;
			const float maxpitchrate = maxrate;
			const float negmaxpitchrate = -maxpitchrate;
			const int chan = buffer->info.channels;
			short * buf = (short *)buffer->sound_buffer;
			const int chaninc = chan - 1;
			
			unsigned int idx = n*chan;
			
			for (unsigned int i = this_start; i < this_end-1; i++)
			{
				float gaindiff1 = computed_gain1 - last_computed_gain1;
				float gaindiff2 = computed_gain2 - last_computed_gain2;
				if (gaindiff1 > maxrate)
					gaindiff1 = maxrate;
				else if (gaindiff1 < negmaxrate)
					gaindiff1 = negmaxrate;
				if (gaindiff2 > maxrate)
					gaindiff2 = maxrate;
				else if (gaindiff2 < negmaxrate)
					gaindiff2 = negmaxrate;
				last_computed_gain1 = last_computed_gain1 + gaindiff1;
				last_computed_gain2 = last_computed_gain2 + gaindiff2;
				
				float pitchdiff = pitch - last_pitch;
				if (pitchdiff > maxpitchrate)
					pitchdiff = maxpitchrate;
				else if (pitchdiff < negmaxpitchrate)
					pitchdiff = negmaxpitchrate;
				last_pitch = last_pitch + pitchdiff;
				
				//unsigned int idx = n*chan;
				
				samp1[0] = buf[idx];
				samp1[1] = buf[idx+chaninc];
				
				samp2[0] = buf[idx+chan];
				samp2[1] = buf[idx+chan+chaninc];
				
				chan1[i] = (int) ((n_remain*(samp2[0] - samp1[0]) + samp1[0])*last_computed_gain1);
				chan2[i] = (int) ((n_remain*(samp2[1] - samp1[1]) + samp1[1])*last_computed_gain2);
				
				n_remain += last_pitch;
				idx += (unsigned int) n_remain*chan;
				n_remain -= (unsigned int) n_remain;
				
				samplecount += last_pitch;
			}
			
			n = idx / chan;

			//last sample is a special case due to wrap-around
			{
				int i = this_end - 1;
				
				float maxrate = 1.0/(44100.0*0.01);
				float gaindiff1 = computed_gain1 - last_computed_gain1;
				if (gaindiff1 > maxrate)
					gaindiff1 = maxrate;
				if (gaindiff1 < -maxrate)
					gaindiff1 = -maxrate;
				float gaindiff2 = computed_gain2 - last_computed_gain2;
				if (gaindiff2 > maxrate)
					gaindiff2 = maxrate;
				if (gaindiff2 < -maxrate)
					gaindiff2 = -maxrate;
				float this_gain_1 = last_computed_gain1 + gaindiff1;
				float this_gain_2 = last_computed_gain2 + gaindiff2;
				last_computed_gain1 = this_gain_1;
				last_computed_gain2 = this_gain_2;
				
				float maxpitchrate = maxrate;
				float pitchdiff = pitch - last_pitch;
				pitchdiff = (pitchdiff > maxpitchrate) ? maxpitchrate : pitchdiff;
				pitchdiff = (pitchdiff < -maxpitchrate) ? -maxpitchrate : pitchdiff;
				last_pitch = last_pitch + pitchdiff;
				
				unsigned int idx1 = n*chan;
				samp1[0] = buf[idx1];
				samp1[1] = buf[idx1+chan-1];
				if (n+1 >= samples)
				{
					//cout << "blending to " << ((n+1)%samples) << endl;
					unsigned int idx2 = ((n+1)%samples)*chan;
					samp2[0] = buf[idx2];
					samp2[1] = buf[idx2+chan-1];
				}
				else
				{
					unsigned int idx2 = idx1+chan;
					samp2[0] = buf[idx2];
					samp2[1] = buf[idx2+chan-1];
				}
				chan1[i] = (int) ((n_remain*(samp2[0] - samp1[0]) + samp1[0])*this_gain_1);
				chan2[i] = (int) ((n_remain*(samp2[1] - samp1[1]) + samp1[1])*this_gain_2);
				
				samplecount += last_pitch;
			}
		}
		else
		{
			for (unsigned int i = this_start; i < this_end; i++)
			{
				chan1[i] = chan2[i] = 0;
			}
		}
		
		double newpos = sample_pos + sample_pos_remainder + samplecount;
		
		if (newpos >= samples && !loop)
			playing = 0;
		else
		{
			while (newpos >= samples)
				newpos -= samples;
			sample_pos = (unsigned int) newpos;
			sample_pos_remainder = newpos - sample_pos;
		}
		
		remaining_len -= this_run;
		this_start += this_run;
	}
}

inline void SOUNDSOURCE::IncrementWithPitch(int num)
{
	/*int samples = buffer->GetSoundInfo().GetSamples();
	sample_pos_remainder += pitch*num;
	while (sample_pos_remainder >= 1.0 && playing)
	{
		sample_pos_remainder -= 1.0;
		int newpos = sample_pos + 1;
		if (newpos >= samples && !loop)
			playing = 0;
		else
			sample_pos = (newpos) % samples;
	}*/
	
	int samples = buffer->GetSoundInfo().GetSamples();
	double newpos = sample_pos + sample_pos_remainder + (num)*pitch;
	if (newpos >= samples && !loop)
		playing = 0;
	else
	{
		while (newpos >= samples)
			newpos -= samples;
		sample_pos = (unsigned int) newpos;
		sample_pos_remainder = newpos - sample_pos;
	}
}

inline void SOUNDSOURCE::Sample16bit(unsigned int peekoffset, int & chan1, int & chan2)
{
	//lines below are used for performance testing to find bottlenecks
	//chan1 = 123;
	//chan2 = 123;
	//chan1 = buffer->GetSample16bit(1, peekoffset);
	//chan2 = buffer->GetSample16bit(2, peekoffset);
	
	int samples = buffer->GetSoundInfo().GetSamples();
	
	#ifdef PREFER_BRANCHING_OVER_MULTIPLY
	if (playing)
	{
		int newsamplepos = (sample_pos + peekoffset);
		if (newsamplepos < samples)
		{
			chan1 = (int) (buffer->GetSample16bit(1, newsamplepos)*computed_gain1);
			chan2 = (int) (buffer->GetSample16bit(2, newsamplepos)*computed_gain2);
		}
		else if (loop)
		{
			int peekpos = (sample_pos + peekoffset) % samples;
			chan1 = (int) (buffer->GetSample16bit(1, peekpos)*computed_gain1);
			chan2 = (int) (buffer->GetSample16bit(2, peekpos)*computed_gain2);
		}
		else
		{
			chan1 = chan2 = 0;
		}
	}
	else
	{
		chan1 = chan2 = 0;
	}
	#else
	int peekplaying = 1 - ((sample_pos + peekoffset) / samples)*(1-loop);
	peekplaying *= playing;
	int peekpos = (sample_pos + peekoffset*peekplaying) % samples;
	chan1 = (int) (buffer->GetSample16bit(1, peekpos)*computed_gain1*peekplaying);
	chan2 = (int) (buffer->GetSample16bit(2, peekpos)*computed_gain2*peekplaying);
	#endif
}

inline void SOUNDSOURCE::Advance(unsigned int offset)
{
	playing = 1 - ((sample_pos + offset) / buffer->GetSoundInfo().GetSamples())*(1-loop);
	
	sample_pos = (sample_pos + offset*playing) % buffer->GetSoundInfo().GetSamples();
}

/*inline const SOUNDINFO & SOUNDBUFFER::GetSoundInfo() const
{
	return info;
}*/

SOUNDSOURCE & SOUND::NewSource(const string & buffername)
{
	if (disable)
	{
		{
			SOUNDSOURCE newsrc;
			sourcelist.push_back(newsrc);
		}
		sourcelist.back().Reset();
		sourcelist.back().SetLoop(false);
		sourcelist.back().Stop();
		return sourcelist.back();
	}
	
	map <string, SOUNDBUFFER>::iterator i = buffermap.find(buffername);
	if (i == buffermap.end())
	{
		//throw EXCEPTION(__FILE__, __LINE__, "Can't create new source with invalid buffer \"" + buffername + "\"");
		//cerr << __FILE__ << "," << __LINE__ << ": Can't create new source with invalid buffer \"" + buffername + "\"" << endl;
		LoadBuffer(buffername);
		return NewSource(buffername);
	}
	else
	{
		{
			SOUNDSOURCE newsrc;
			sourcelist.push_back(newsrc);
		}
		sourcelist.back().SetBuffer(i->second);
		sourcelist.back().Reset();
		sourcelist.back().SetLoop(false);
		sourcelist.back().Stop();
		return sourcelist.back();
	}
}

void SOUND::LoadBuffer(const string & buffername)
{
	//prefer ogg
	if (FILEEXISTS_FUNCTION_CALL(PATHMANAGER_FUNCTION_CALL(buffername+".ogg")))
		buffermap[buffername].Load(PATHMANAGER_FUNCTION_CALL(buffername+".ogg"), deviceinfo);
	else
	{
		buffermap[buffername].Load(PATHMANAGER_FUNCTION_CALL(buffername+".wav"), deviceinfo);
	}
}

void SOUND::Tick()
{
	if (disable)
		return;
	
	list <SOUNDSOURCE *> todel;
	ITERATE(SOUNDSOURCE, sourcelist, i)
	{
		if (!i->Audible() && i->GetAutoDelete())
		{
			todel.push_back(&(*i));
		}
	}
	
	ITERATE(SOUNDSOURCE *, todel, i)
	{
		DeleteSource(*i);
	}
	
	//cout << sourcelist.size() << endl;
}

void SOUND::DetermineActiveSources(list <SOUNDSOURCE *> & active_sourcelist, list <SOUNDSOURCE *> & inaudible_sourcelist)
{
	active_sourcelist.clear();
	inaudible_sourcelist.clear();
	//int sourcenum = 0;
	ITERATE(SOUNDSOURCE, sourcelist, i)
	{
		if (i->Audible())
		{
			active_sourcelist.push_back(&(*i));
			//cout << "Tick: " << &(*i) << endl;
			//cout << "Source is audible: " << i->GetName() << ", " << i->GetGain() << ":" << i->ComputedGain(1) << "," << i->ComputedGain(2) << endl;
			//cout << "Source " << sourcenum << " is audible: " << i->GetName() << endl;
		}
		else
		{
			inaudible_sourcelist.push_back(&(*i));
		}
		//sourcenum++;
	}
	//cout << "sounds active: " << active_sourcelist.size() << ", sounds inactive: " << inaudible_sourcelist.size() << endl;
}

void SOUND::DeleteSource(SOUNDSOURCE * todel)
{
	if (disable)
		return;
	
	list <SOUNDSOURCE>::iterator delit = sourcelist.end();
	ITERATE(SOUNDSOURCE, sourcelist, i)
	{
		if (&(*i) == todel)
			delit = i;
	}
	
	if (delit == sourcelist.end())
	{
		//cerr << __FILE__ << "," << __LINE__ << "Can't find source to delete" << endl;
		//throw EXCEPTION(__FILE__, __LINE__, "Can't find source to delete");
		UNRECOVERABLE_ERROR_FUNCTION(__FILE__,__LINE__,"Can't find source to delete");
		return;
	}
	else
		sourcelist.erase(delit);
}

void SOUND::NewSourcePlayOnce(const string & buffername)
{
	if (disable)
		return;
	
	SOUNDSOURCE & src = NewSource(buffername);
	src.SetAutoDelete(true);
	src.Set3DEffects(false);
	src.Play();
}

void SOUND::Compute3DEffects(list <SOUNDSOURCE *> & sources, const VERTEX & listener_pos, const QUATERNION & listener_rot)
{
	ITERATE(SOUNDSOURCE *, sources, i)
	{
		if ((*i)->Get3DEffects())
		{
			VERTEX relvec;
			relvec = (*i)->GetPosition() - listener_pos;
			//cout << "sound pos: ";(*i)->GetPosition().DebugPrint();
			//cout << "listener pos: ";listener_pos.DebugPrint();
			//cout << "listener pos: ";listener_pos.DebugPrint();
			//cout << "camera pos: ";cam.GetPosition().ScaleR(-1.0).DebugPrint();
			float len = relvec.len();
			relvec = listener_rot.RotateVec(relvec);
			float cgain = log(1000.0/pow((double)len,1.3))/log(100);
			if (cgain > 1.0)
				cgain = 1.0;
			if (cgain < 0.0)
				cgain = 0.0;
			float xcoord = relvec.normalize().x;
			float pgain1 = -xcoord;
			if (pgain1 < 0)
				pgain1 = 0;
			float pgain2 = xcoord;
			if (pgain2 < 0)
				pgain2 = 0;
			//cout << cgain << endl;
			//cout << xcoord << endl;
			(*i)->SetComputationResults(cgain*(*i)->GetGain()*(1.0-pgain1), cgain*(*i)->GetGain()*(1.0-pgain2));
		}
		else
		{
			(*i)->SetComputationResults((*i)->GetGain(), (*i)->GetGain());
		}
	}
}

const bool SOUNDSOURCE::Audible() const
{
	bool canhear = (playing == 1) && (GetGain() > 0);
	/*if (Get3DEffects())
	{
		canhear = canhear && (ComputedGain(1) > 0 || ComputedGain(2) > 0);
	}*/
	
	return canhear;
}

void SOUNDSOURCE::SetGainSmooth(const float newgain, const float dt)
{
	//float coeff = dt*40.0;
	if (dt <= 0)
		return;
	
	//low pass filter
	/*float coeff = 0.98*(0.004/dt);
	if (coeff > 1.0)
		coeff = 1.0;
	if (coeff < 0.0)
		coeff = 0.0;
	gain = gain*(coeff)+newgain*(1.0-coeff);*/
	
	//rate limit
	float ndt = dt * 4.0;
	float delta = newgain - gain;
	if (delta > ndt)
		delta = ndt;
	if (delta < -ndt)
		delta = -ndt;
	gain = gain + delta;
}

void SOUNDSOURCE::SetPitchSmooth(const float newpitch, const float dt)
{
	//float coeff = dt*40.0;
	if (dt > 0)
	{
		//low pass filter
		/*float coeff = 0.9*(0.004/dt);
		if (coeff > 1.0)
			coeff = 1.0;
		if (coeff < 0.0)
			coeff = 0.0;
		pitch = pitch*(coeff)+newpitch*(1.0-coeff);*/
		
		//rate limit
		float ndt = dt * 4.0;
		float delta = newpitch - pitch;
		if (delta > ndt)
			delta = ndt;
		if (delta < -ndt)
			delta = -ndt;
		pitch = pitch + delta;
	}
}

//the coefficients are arrays of size xcoeff[neworder+1]
void SOUNDFILTER::SetFilter(const int neworder, const float * xcoeff, const float * ycoeff)
{
	if (neworder > MAX_FILTER_ORDER || neworder < 0)
		//cerr << __FILE__ << "," << __LINE__ << "Filter order is larger than maximum order" << endl;
		UNRECOVERABLE_ERROR_FUNCTION(__FILE__,__LINE__,"Filter order is larger than maximum order");
	else
	{
		order = neworder;
		for (int i = 0; i < neworder+1; i++)
		{
			xc[i] = xcoeff[i];
			yc[i] = ycoeff[i];
		}
	}
}

void SOUNDFILTER::Filter(int * chan1, int * chan2, const int len)
{
	for (int i = 0; i < len; i++)
	{
		//store old state
		for (int s = 1; s <= order; s++)
		{
			statex[0][s] = statex[0][s-1];
			statex[1][s] = statex[1][s-1];
			
			statey[0][s] = statey[0][s-1];
			statey[1][s] = statey[1][s-1];
		}
		
		//set the sample state for now to the current input
		statex[0][0] = chan1[i];
		statex[1][0] = chan2[i];
		
		/*//zero the output
		chan1[i] = 0;
		chan2[i] = 0;
		
		//cout << "filter: " << i << ": " << statey[0][1] << ", " << yc[1] << endl;
		
		for (int o = 0; o <= order; o++)
		{
			//do the filtering
			chan1[i] += (int) (statex[0][o]*xc[o]+statey[0][o]*yc[o]);
			chan2[i] += (int) (statex[1][o]*xc[o]+statey[1][o]*yc[o]);
		}*/
		
		switch (order)
		{
			case 1:
				chan1[i] = (int) (statex[0][0]*xc[0]+statex[0][1]*xc[1]+statey[0][1]*yc[1]);
				chan2[i] = (int) (statex[1][0]*xc[0]+statex[1][1]*xc[1]+statey[1][1]*yc[1]);
				break;
			case 0:
				chan1[i] = (int) (statex[0][0]*xc[0]);
				chan2[i] = (int) (statex[1][0]*xc[0]);
				break;
			default:
				break;
		}
		
		//store the state of the output
		statey[0][0] = chan1[i];
		statey[1][0] = chan2[i];
	}
}

SOUNDFILTER & SOUNDSOURCE::GetFilter(int num)
{
	int curnum = 0;
	ITERATE(SOUNDFILTER, filters, i)
	{
		if (num == curnum)
			return *i;
	}
	
	//cerr << __FILE__ << "," << __LINE__ << "Asked for a non-existant filter" << endl;
	UNRECOVERABLE_ERROR_FUNCTION(__FILE__,__LINE__,"Asked for a non-existant filter");
	SOUNDFILTER * nullfilt = NULL;
	return *nullfilt;
}

void SOUNDFILTER::SetFilterOrder1(float xc0, float xc1, float yc1)
{
	order = 1;
	xc[0] = xc0;
	xc[1] = xc1;
	yc[1] = yc1;
}

void SOUNDFILTER::SetFilterOrder0(float xc0)
{
	order = 0;
	xc[0] = xc0;
}
