#include "vertexarray.h"
//#include "game.h"
//extern GAME game;
#include "common.h"

const TESTER VERTEXARRAY::Test()
{
	TESTER mytest;
	
	int points = 0;
	
	const float * ptr;
	int ptrnum;
	float somevec[3];
	somevec[0] = somevec[2] = 0;
	somevec[1] = 1000.0;
	SetNormals(somevec, 3);
	GetNormals(ptr, ptrnum);
	if (ptr[1] == 1000.0) points++;
	if (ptrnum == 3) points++;
	SetNormals(NULL, 0);
	GetNormals(ptr, ptrnum);
	if (ptrnum == 0 && ptr == NULL) points++;
	mytest.SubTestComplete("Normal assignment/get", points, 3);
	
	//by similarity, the vertex and face assignment functions are OK if the above normal test is OK
	
	points = 0;
	SetTexCoordSets(2);
	if (GetTexCoordSets() == 2) points++;
	GetTexCoords(0, ptr, ptrnum);
	if (ptrnum == 0 && ptr == NULL && texcoords[1].empty()) points++;
	SetTexCoords(1, somevec, 2);
	GetTexCoords(1, ptr, ptrnum);
	if (texcoords[0].empty() && texcoords[1].size() == 2) points++;
	if (ptrnum == 2 && ptr[1] == 1000.0) points++;
	mytest.SubTestComplete("Texture coordinate assignment/get", points, 4);
	
	points = 0;
	Clear();
	SetNormals(somevec, 3);
	VERTEXARRAY otherarray(*this);
	otherarray.GetNormals(ptr, ptrnum);
	if (ptr[1] == 1000.0) points++;
	if (ptrnum == 3) points++;
	otherarray.Clear();
	otherarray.GetNormals(ptr, ptrnum);
	if (ptrnum == 0) points++;
	otherarray = *this;
	otherarray.GetNormals(ptr, ptrnum);
	if (ptr[1] == 1000.0) points++;
	if (ptrnum == 3) points++;
	mytest.SubTestComplete("Normal copy", points, 5);
	
	points = 0;
	VERTEXARRAY addarray;
	otherarray.Clear();
	Clear();
	SetNormals(somevec, 3);
	otherarray.SetNormals(somevec, 3);
	addarray = *this + otherarray;
	addarray.GetNormals(ptr, ptrnum);
	if (ptrnum == 6) points++;
	if (ptr[1] == 1000.0) points++;
	if (ptr[4] == 1000.0) points++;
	//for (int i = 0; i < ptrnum; i++) cout << i << ". " << ptr[i] << endl;
	mytest.SubTestComplete("Normal add", points, 3);
	
	points = 0;
	Clear();
	VERTEXARRAY facearray1, facearray2;
	int someint[3];
	someint[0] = 0;
	someint[1] = 1;
	someint[2] = 2;
	facearray1.SetFaces(someint, 3);
	facearray2.SetFaces(someint, 3);
	*this = facearray1 + facearray2;
	const int * ptri;
	GetFaces(ptri, ptrnum);
	if (ptrnum == 6) points++;
	if (ptri[1] == 1) points++;
	if (ptri[4] == 1) points++;
	Clear();
	facearray1.SetVertices(somevec, 3);
	facearray2.SetVertices(somevec, 3);
	*this = facearray1 + facearray2;
	GetFaces(ptri, ptrnum);
	if (ptrnum == 6) points++;
	if (ptri[1] == 1) points++;
	if (ptri[4] == 2) points++;
	//for (int i = 0; i < ptrnum; i++) cout << i << ". " << ptri[i] << endl;
	mytest.SubTestComplete("Face add", points, 6);
	
	return mytest;
}

void VERTEXARRAY::SetNormals(float * newarray, unsigned int newarraycount)
{
	if (newarraycount % 3)
		throw EXCEPTION(__FILE__, __LINE__, "Tried to assign values that aren't in sets of 3");
	
	if (newarraycount != normals.size())
	{
		normals.resize(newarraycount);
	}
	
	float * myarray = &(normals[0]);
	for (unsigned int i = 0; i < newarraycount; ++i)
	{
		myarray[i] = newarray[i];
	}
	
	/*for (int i = 0; i < newarraycount; i++)
	{
		normals[i] = newarray[i];
	}*/
}

void VERTEXARRAY::SetVertices(float * newarray, unsigned int newarraycount)
{
	if (newarraycount % 3)
		throw EXCEPTION(__FILE__, __LINE__, "Tried to assign values that aren't in sets of 3");
	
	if (newarraycount != vertices.size())
	{
		//cout << "New size: " << newarraycount << " (was " << vertices.size() << ")" << endl;
		vertices.resize(newarraycount);
	}
	
	float * myarray = &(vertices[0]);
	for (unsigned int i = 0; i < newarraycount; ++i)
	{
		myarray[i] = newarray[i];
	}
	
	/*int i = 0;
	ITERVEC(float, vertices, it)
	{
		*it = newarray[i];
		++i;
	}*/
}

void VERTEXARRAY::SetFaces(int * newarray, unsigned int newarraycount)
{
	if (newarraycount % 3)
		throw EXCEPTION(__FILE__, __LINE__, "Tried to assign values that aren't in sets of 3");
	
	if (newarraycount != faces.size())
		faces.resize(newarraycount);
	
	
	int * myarray = &(faces[0]);
	for (unsigned int i = 0; i < newarraycount; ++i)
	{
		myarray[i] = newarray[i];
	}
	
	/*for (int i = 0; i < newarraycount; i++)
	{
		faces[i] = newarray[i];
	}*/
}

void VERTEXARRAY::SetTexCoordSets(int newtcsets)
{
	texcoords.clear();
	texcoords.resize(newtcsets);
}

//set is zero indexed
void VERTEXARRAY::SetTexCoords(unsigned int set, float * newarray, unsigned int newarraycount)
{
	if (set >= texcoords.size())
		throw EXCEPTION(__FILE__, __LINE__, "Tried to assign a tex coord set beyond the allocated number of sets");
	
	if (newarraycount % 2)
		throw EXCEPTION(__FILE__, __LINE__, "Tried to assign values that aren't in sets of 2");
	
	if (texcoords[set].size() != newarraycount)
		texcoords[set].resize(newarraycount);
	
	float * myarray = &(texcoords[set][0]);
	for (unsigned int i = 0; i < newarraycount; ++i)
	{
		myarray[i] = newarray[i];
	}
	
	/*for (int i = 0; i < newarraycount; i++)
	{
		texcoords[set][i] = newarray[i];
	}*/
}

void VERTEXARRAY::GetNormals(const float * & output_array_pointer, int & output_array_num) const
{
	output_array_num = normals.size();
	output_array_pointer = normals.empty() ? NULL : &normals[0];
}

void VERTEXARRAY::GetVertices(const float * & output_array_pointer, int & output_array_num) const
{
	output_array_num = vertices.size();
	output_array_pointer = vertices.empty() ? NULL : &vertices[0];
}

void VERTEXARRAY::GetFaces(const int * & output_array_pointer, int & output_array_num) const
{
	output_array_num = faces.size();
	output_array_pointer = faces.empty() ? NULL : &faces[0];
}

void VERTEXARRAY::GetTexCoords(unsigned int set, const float * & output_array_pointer, int & output_array_num) const
{
	if (set >= texcoords.size())
		throw EXCEPTION(__FILE__, __LINE__, "Tried to get a tex coord set beyond the allocated number of sets");
	
	output_array_num = texcoords[set].size();
	output_array_pointer = texcoords[set].empty() ? NULL : &texcoords[set][0];
}

#define COMBINEVECTORS(vname) {out.vname.reserve(vname.size() + v.vname.size());out.vname.insert(out.vname.end(), vname.begin(), vname.end());out.vname.insert(out.vname.end(), v.vname.begin(), v.vname.end());}

VERTEXARRAY VERTEXARRAY::operator+ (const VERTEXARRAY & v) const
{
	VERTEXARRAY out;
	
	//out.normals.reserve(normals.size() + v.normals.size());
	//out.normals.insert(out.normals.end(), normals.begin(), normals.end());
	//out.normals.insert(out.normals.end(), v.normals.begin(), v.normals.end());
	
	int idxoffset = vertices.size()/3;
	
	COMBINEVECTORS(normals)
	COMBINEVECTORS(vertices)
	
	//COMBINEVECTORS(faces)
	out.faces.reserve(faces.size() + v.faces.size());
	out.faces.insert(out.faces.end(), faces.begin(), faces.end());
	//out.faces.insert(out.faces.end(), v.faces.begin(), v.faces.end());
	//int startat = out.faces.size();
	for (unsigned int i = 0; i < v.faces.size(); i++)
	{
		//out.faces[faces.size() + i] = v.faces[i] + idxoffset;
		//out.faces[startat + i] = v.faces[i];
		out.faces.push_back(v.faces[i]+idxoffset);
	}
	
	int maxtcsets = GetTexCoordSets();
	if (v.GetTexCoordSets() > maxtcsets)
		maxtcsets = v.GetTexCoordSets();
	int tcsets1 = GetTexCoordSets();
	int tcsets2 = v.GetTexCoordSets();
	out.SetTexCoordSets(maxtcsets);
	//ITERVEC(vector <float>,texcoords,i)
	for (int i = 0; i < maxtcsets; i++)
	{
		if (i >= tcsets1 && i < tcsets2)
		{
			out.texcoords[i] = v.texcoords[i];
		}
		else if (i < tcsets1 && i >= tcsets2)
		{
			out.texcoords[i] = texcoords[i];
		}
		else if (i < tcsets1 && i < tcsets2)
		{
			COMBINEVECTORS(texcoords[i])
		}
	}
	
	return out;
}

void VERTEXARRAY::SetToBillboard(float scale)
{
	/*glNormal3f( 0.0f, 0.0f, 1.0f);
	glTexCoord2f(0.0f, 0.0f); glVertex3f( 0, -scale,  -scale);
	glTexCoord2f(1.0f, 0.0f); glVertex3f( 0, -scale,  scale);
	glTexCoord2f(1.0f, 1.0f); glVertex3f( 0,  scale,  scale);
	glTexCoord2f(0.0f, 1.0f); glVertex3f( 0,  scale,  -scale);*/
	
	int bfaces[6];
	bfaces[0] = 0;
	bfaces[1] = 1;
	bfaces[2] = 2;
	bfaces[3] = 0;
	bfaces[4] = 2;
	bfaces[5] = 3;
	SetFaces(bfaces, 6);
	
	float normals[12];
	for (int i = 0; i < 12; i+=3)
	{
		normals[i] = 0;
		normals[i+1] = 0;
		normals[i+2] = 1;
	}
	SetNormals(normals, 12);
	
	float verts[12];
	/*verts[0] = verts[3] = verts[6] = verts[9] = 0.0;
	verts[1] = verts[2] = verts[4] = verts[11] = -scale;
	verts[5] = verts[7] = verts[8] = verts[10] = scale;*/
	verts[2] = verts[5] = verts[8] = verts[11] = 0.0;
	verts[1] = verts[0] = verts[4] = verts[9] = -scale;
	verts[3] = verts[7] = verts[6] = verts[10] = scale;
	SetVertices(verts, 12);
	
	float tc[8];
	tc[0] = tc[1] = tc[3] = tc[6] = 0.0;
	tc[2] = tc[4] = tc[5] = tc[7] = 1.0;
	SetTexCoordSets(1);
	SetTexCoords(0, tc, 8);
}
